
// ****************************************************************************
//
// Routine:      acReadGlobal
// Author:       T.Oest
//
// Description:  read global alignment constants from database
//          
// Called by:    acread_      
//
// Modification:
// transformation routine for GEDE table called                     16/09/99 TO
// access by key or data                                            01/12/99 TO
// include return code                                              07/02/00 TO
// GEDE transformation moved to acread.C                            06/03/00 TO
// corrected include path bug                                       16/08/00 AG
// added ARTE_mark()                                                23/11/00 AG
//
// 08/12/00 (SKAG) -reading from file added
// 12/12/00 (SKAG) -added checks for the old ACGA tables for backward 
//                  compatibility
// 23/04/01 (SKAG) -program doesn't crash if there is no alignment file;
//                  it only issues an error.
//
// ****************************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <algo.h>
#include <limits.h>

// ARTE includes
#include "arte/arte.hh"
#include "farm/farm.hh"

// ONLINE includes
#include "table.h"
#include "esp.h"

// prototypes
int  acReadGlobalByTime( int keystatus, char *dbserver );
int  acReadGlobalByKey ( int keystatus, char *dbserver );
void ackuexec          ( char * KuipCommand );
 
// globals
static int first              = 0;
int        acReadGlobalSwitch = 1;

//*****************************************************************
// function
//*****************************************************************

int
acReadGlobal( int KTstatus )
{
  // show up
  ARTE_mark( "acReadGlobal 0", 0 );
  
  int flag       = 0;
  int ACGAupdate = 0;
  
  int KeyStatus = KTstatus;
  if( KTstatus == 0 ) KeyStatus = 2;
  
  //==================================================================
  // no action if switched off
  //==================================================================
  if( acReadGlobalSwitch != 1 ){
    cout << " acReadGlobal switched off: no global alignment constants read in " << endl;
    ARTE_mark( "acReadGlobal 1", 0 );
    return -1;
  }
  
  //==================================================================
  // no action for MC
  //==================================================================
  if (ArteTable<MTRA>::size() > 0 ){
    ARTE_mark( "acReadGlobal 1", 0 );
    return 0;
  }
  
  //==================================================================
  // read constants from file
  //==================================================================
  char *glalfilename ;
  if (( glalfilename=getenv( "READ_GLAL_FROM_FILE" ))) {
    FILE *glalfile;
    if (( glalfile = fopen ( glalfilename, "r" ))) {
      ACGA acga;
      ArteTable<ACGA>::clear();

      int npar ;
      while((npar=fscanf(glalfile,
		       "0x%x %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f\n",
		       &acga.cmp, &acga.dx, &acga.dy, &acga.dz,
		       &acga.dtx, &acga.dty, &acga.fz, &acga.z,
		       &acga.rot[0], &acga.rot[1], &acga.rot[2],
		       &acga.rot[3], &acga.rot[4], &acga.rot[5],
		       &acga.rot[6], &acga.rot[7], &acga.rot[8] )) != EOF )
        if ( npar==17 ) ArteTable<ACGA>::push_back( acga ) ;
    
      flag = 2;
    } else {
      printf ("ERROR(acReadGlobal): no file %s\n", glalfilename);
    }
  } else {
  //==================================================================
  // database server (default: /SETUP_ALIGN)
  //                  can be changed by setting DB_GLOBAL_ALIGNMENT
  //==================================================================
    char *defaultserver = "/SETUP_ALIGN";
    char *dbserver      = 0;
    if( getenv( "DB_GLOBAL_ALIGNMENT" ) == NULL ){
      dbserver = defaultserver;
    }
    else {
      dbserver = getenv( "DB_GLOBAL_ALIGNMENT" );
    }
  
  //==================================================================
  // get LUN number for database
  //==================================================================
    char LunName[5] = {'D','B','G','A',0};
    char *KuipCommand =  (char*) malloc( max(128, int(24+strlen(dbserver)) ) );
  
    if( first == 0 ){
      first = 1;
      sprintf ( KuipCommand, "IOFILE %s %s", LunName, dbserver );
      ackuexec( KuipCommand );
    }
 
  //==================================================================
  // access by key table
  //==================================================================
    if( KeyStatus >= 1 ){
      ARTE_mark( "acReadGlobalByKey", 0 );
      ACGAupdate = acReadGlobalByKey( KeyStatus, dbserver );
    }
    flag = ACGAupdate;
  
  //==================================================================
  // acces by date if KeySstatus < 1 or access by key not sucessfull
  //==================================================================
  
    if( KeyStatus < 1 || ACGAupdate < 0 ){
      ARTE_mark( "acReadGlobalByTime", 0 );
      ACGAupdate = acReadGlobalByTime( KeyStatus, dbserver );
      if( ACGAupdate <  0 ) flag = -4;
      if( ACGAupdate == 1 ) flag = 2;
    }
  }
  //==================================================================
  // check if ACGA table exist
  //==================================================================
  
  if( ArteTable<ACGA>::size() == 0 ){
    ARTE_mark( "acReadGlobal 1", 0 );
    ARTE_exit( GRR_SFATAL, "acReadGlobal: ACGA not existing", 0 );
    return -3;
  } else {
  //==================================================================
  // added for beckward db compatibility
  // if there is only one entry for ITR/OTR set the others
  // to the same values 
  //==================================================================  
    ACGA def_otr, def_itr ;
    int otr_missing[]={1,1,1,1} ; // 0x2, 0x10002, 0x20002, 0x30002
    int itr_missing[]={1,1,1,1} ; // 0x3, 0x10003, 0x20003, 0x30003
    
    for( ArteTable<ACGA>::iterator acga = ArteTable<ACGA>::begin();
	 acga != ArteTable<ACGA>::end(); ++acga ) {
      if ( ((*acga)->fz)<0.001 ) (*acga)->fz=1.0 ; // scale factor was 0
      if ( ((*acga)->cmp)==0x2 ) { //default OTR
	def_otr=**acga ;
      } else if ( ((*acga)->cmp)==0x3 ) { //default ITR
	def_itr=**acga ;
      }
      if ( ((*acga)->cmp & 0xFFFF)==0x2 ) { // all OTR
	int indx=((*acga)->cmp) >> 16 ;
	if ((indx<0)||(indx>3)) continue ;
	otr_missing[indx] = 0 ;
      }
      if ( ((*acga)->cmp & 0xFFFF)==0x3 ) { // all ITR
	int indx=((*acga)->cmp) >> 16 ;
	if ((indx<0)||(indx>3)) continue ;
	itr_missing[indx] = 0 ;
      }
      
    }

    for ( int i=1; i<4; i++ ) {
      if ( otr_missing[i] && (!otr_missing[0]) ) {
	def_otr.cmp = (i<<16)|0x2 ;
	ArteTable<ACGA>::push_back( def_otr ) ;
      }
      if ( itr_missing[i] && (!itr_missing[0]) ) {
	def_itr.cmp = (i<<16)|0x3 ;
	ArteTable<ACGA>::push_back( def_itr ) ;
      }
    }
   
  }
 
 return flag;
}



