
// *****************************************************************************
//
// source:    faFile.C
//
// created:   30 Jul 1997
//
// Author:    Andreas Gellrich
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// purpose:   Handle read-in of raw data event files (file.daq).
//
// symbols:   faReadFile
//            faOpenFile
//            faCloseFile
//
// called by: faRead
//
// changes:         multiple file input (AG: 27-Nov-98)
//                  includes (AG: 11 Jan 99)
//                  remove not needed evtNum (AG: 15 Jan 99)
//                  polishing (AG: 27 Jan 99)
//                  removed faUtil.hh (AG: 23 Feb 99)
//                  prepare for automatic staging (AG: 27 Apr 99)
//                  include staging function (AG: 12 May 99)
//                  stage disk (AG: 14 May 99)
//                  NO staging for HP-UX etc (AG: 25 May 99)
//                  put includes (AG: 1 Jun 99)
// 09 Sep 1999 (AG) always use cout
// 13 Sep 1999 (AG) polishing
// 31 Mar 2000 (AG) fix new defines for g++
// 23 May 2000 (AG) add realpath() to get real file name
// 17 Jan 2001 (AG) added commenst
//
// *****************************************************************************

// standard includes
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>

// ONLINE includes
#include "event.h"
#include "etb.h"
#include "grr.h"
#include "stage.h"
#include "log.h"

// Arte includes
#include "farm/fa.hh"

// externals
extern int  FADEB;               // debug flag
extern int  OLD;                 // old format
extern int  SEG_LEN;             // buffer size
extern char *evtBuf;             // buffer
extern char *IN_FILE[N_IN_FILE]; // input files
extern char *HBSTAGE;            // stage disk
extern FILE *in;                 // input streams
extern int  nFiles;              // number of files

//
// program code
//
int
faReadFile()
{
 // inits
  int         rc      = ETB_ok;
  static int  iFile   = 0;
  static int  newFile = 1;
  static FILE *in     = 0;
  static char *name   = 0;
    
  //
  // open file
  //
  while( 1 ){
    if( newFile ){
      
      // close old file
      if( in ){
	faCloseFile( name, in );
      }
      
      // new file
      if( iFile >= nFiles ){
	return ETB_file_err;
      }
      
      // open
      name = IN_FILE[iFile]; // current file name      
      if( ( in = faOpenFile( name ) ) ){
	newFile = 0; // flag
	iFile++;     // increase file counter
#ifdef _DEB2_
	if( FADEB ){
	  cout << "FILE: "
	       << "file="    << name            << " "
	       << "in="      << (void *) in     << " "
	       << "SEG_LEN=" << SEG_LEN         << " "
	       << "evtBuf="  << (void *) evtBuf << endl;
	}
#endif
      }
      // file open error
      else {
	in = 0;
	cout << "FILE: ERROR: Open FAILED for file=" << name << endl;
	rc = ETB_file_err;
	break;
      }
    }
    
    //
    // read event from file
    //
    int rc = ETB_ok;
    if( !OLD ){
      int bc = 0;
      LogHdr hdr;
      if( ( rc = l4aGetEvt( in, (ETB_event *) evtBuf, SEG_LEN, &bc, &hdr ) ) == GRR_OK ){
	if( FADEB ){
	  cout << endl;
	  cout << "FILE: read event from file=" << name << endl;
	  cout << "FILE: Log header: "
	       << "bc="     << bc         << " "
	       << "exp="    << hdr.exp    << " "
	       << "run="    << hdr.run    << " "
	       << "evt="    << hdr.evt    << " "
	       << "datime=" << hdr.datime << endl;
	}
	break;
      }
      // eof
      else {
	newFile = 1;
	cout << "FILE: End-of-file in file=" << name << " "
	     << "rc="                        << rc   << endl;
	rc = ETB_file_err;
	continue;
      }
    }
    //
    // read old format
    //
    else {
      if( ( rc = ETB_read_event( in, (ETB_event *) evtBuf, SEG_LEN ) ) == ETB_ok ){
	if( FADEB ){
	  cout << endl;
	  cout << "FILE: read event from OLD file=" << name << endl;
	}
	break;;
      }
      // eof
      else if( rc != ETB_ok ){
	newFile = 1;
	cout << "FILE: End-of-file in OLD file=" << name << " "
	     << "rc="                            << rc   << endl;
	continue;
      }
    }
  }
  
  // return value
  return rc;
}

//
// open file
//
FILE *
faOpenFile( char *name )
{
  // file name
  //static char *name = 0;
  //if( !name ) name = new char[MAXPATHLEN];
  //realpath( name0, name ); 

  // stage acs-files
  if( !strncmp( name, "/acs/", 5 ) ){

    cout << "FILE: Stage file=" << name << endl;

#if ( defined linux ) || ( defined sun ) || ( defined sgi )  || ( defined __linux ) || ( defined __sun ) || ( defined __sgi )

    // staging implemented for Linux, Sun, SGI
    int rc = l4aStageF_( name, HBSTAGE, FADEB );
    if( rc == L4A_OK ){
      cout << "FILE: Staging finished for file=" << name << endl;
    }
    else if( rc == L4A_FAILURE ){
      cout << "FILE: ERROR: Staging FAILED for file=" << name << endl; 
      return 0;
    }
    else if( rc == L4A_OSM_FAILURE ){
      cout << "FILE: ERROR: Staging FAILED due to OSM errors for file=" << name << endl; 
      return 0;
    }
    else if( rc == L4A_NO_ACS_FILE ){
      cout << "FILE: ERROR: No such file on acs file=" << name << endl;
      return 0;
    }
    else {
      cout << "FILE: ERROR: While staging file=" << name << " rc=" << rc << endl; 
      return 0;
    }

    // name of disk file
    char disk[80];
    memset( disk, 0, 80 );
    strcat( disk, HBSTAGE );
    strcat( disk, name + 4 ); // omit "/acs"
    memset( name, 0, 80 );
    strcpy( name, disk );

#else

    // staging NOT implemented
    cout << "FILE: Staging NOT implemented for your machine!" << endl; 
    return 0;

#endif
  }
  
  // open disk file
  FILE *in = 0;
  if( ( in = fopen( name, "rb" ) ) != 0 ){
    cout << "FILE: Opened file=" << name << endl;
    return in;
  }
  else {
    cout << "FILE: ERROR: Opening file=" << name << endl;
    return 0;
  }
}

//
// close file
//
int
faCloseFile( char *name, FILE *in )
{
  if( in ){
    cout << "FILE: Closed file=" << name << endl;
    fclose ( in );
  }  
  return GRR_OK;
}
