#include <iostream>
#include <math.h>

#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#endif

#include "grover/DumpFormatter.hh"
#include "grover/grover.hh"

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                GroverParticle member functions                     //
//                                                                    //
////////////////////////////////////////////////////////////////////////

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 00-09-27  M.Sang   Added new constructor for GroverParticle(GroverVertex*)
// 00-10-26  M.Sang   Default mass is charged pion mass (was zero)
// 00-10-26  M.Sang   Fixed bug in RSEG/RTRA/GroverVertex* constructors -
//                    some elements of covariance matrix were wrong :-(
// 00-11-02  M.Sang   Fixed bug in GroverParticle(ArtePointer<RTRA>) reported
//                    by C.Padilla
// 00-11-10  M.Sang   Fixed bug in Chi2DistTo(GroverParticle*) reported by
//                    Elmar Klaus
// 00-11-14  M.Sang   Chi2 bug fixed
// 00-12-01  M.Sang   fixed minor bug in GroverParticle(ArtePointer<RTRA>)
// 00-12-13  M.Sang   Added SetMass(ArtePointer<TOPP>)
// 00-12-13  M.Sang   Added some off-diagonals in GroverParticle(GroverVertex*)
// 01-01-12  M.Sang   Added operators ==, != and <<
// 01-01-25  M.Sang   Improved behaviour for MC tracks & vertices
// 01-02-15  M.Sang   Improved GroverParticle::ResetParameters(void)
// 01-02-28  M.Sang   GroverVertex is now a friend class. Several member functions
//                    removed, others have protected status.
// 01-02-28  M.Sang   const safe
// 01-02-28  M.Sang   Added mtra->topp initialisation in GroverParticle(MTRA)
// 01-02-29  M.Sang   PropagateBeginTo(z) improved. Now uses current parameters
//                    (not originals) if tampered==true (i.e. the user has manually
//                    set track parameters or covariance elements). If !tampered
//                    ResetParameters() is called automatically to prevent 
//                    incacuracies of covariance elements after a large number 
//                    of propagations.
// 01-03-13  M.Sang   Added preprocessor directives
// 01-04-28  Michael Change PropagateBeginTo to make GroveerCatsExtrapolate 
//                   usable for all particles
// 01-06-01  Michael  Bug fix in Chi2DistTo(const GroverVertex* vert) const
//                    and GroverParticle::Chi2DistTo(const float xx, 
//                                           const float yy, const float zz)
//

//
//
// SEE ALSO grover_inlines.hh !!!!!!!
//
//


// pre-definitions
float   scal(float v1[], float v2[]);
void     res(float v1[], float v2[], float R[]);
void     sum(float v1[], float v2[], float S[]);
void  mult_n(float v1[], float n,    float M[]);



GroverParticle::GroverParticle()
  : xb(0.), yb(0.), zb(0.), axb(0.), ayb(0.),
    xe(0.), ye(0.), ze(0.), axe(0.), aye(0.),
    oxb(0.), oyb(0.), ozb(0.), oaxb(0.), oayb(0.),
    oxe(0.), oye(0.), oze(0.), oaxe(0.), oaye(0.),
    p(FAILURE), mass(FAILURE), nhit(FAILURE), q(FAILURE), type(FAILURE),
    usedinprimary(0), usedinsecondary(0), index(0),
#ifndef GROVER_WITHOUT_ARTE
    parRTRA(0), parRSEG(0), parRCCL(0), parMTRA(0), 
#endif
    parGP(0), parGPL(0), parGV(0),
    inlocal(0), inglobal(0), wprim(0.), wsec(0.),
    dist(1e9), chi2(0.), 
#ifndef GROVER_WITHOUT_ARTE
    topp(0),
#endif
    tampered(false)
{
}



#ifndef GROVER_WITHOUT_ARTE
GroverParticle::GroverParticle(const ArtePointer<RSEG> rseg)
  : oxb(rseg->xf), oyb(rseg->yf), ozb(rseg->zf), 
    oaxb(rseg->txf), oayb(rseg->tyf),
    oxe(rseg->xe), oye(rseg->ye), oze(rseg->ze), 
    oaxe(rseg->txe), oaye(rseg->tye),
    op(fabs(rseg->pf)),
    mass(DEFAULTMASS), // <- note!
    nhit(rseg->nhit), 
    q(FAILURE), type(2),
    usedinprimary(0), usedinsecondary(0), index(0),
    parRTRA(0), parRSEG(rseg), parRCCL(0), parMTRA(0), 
    parGP(0), parGPL(0), parGV(0),
    inlocal(0), inglobal(1), wprim(0.0), wsec(0.),
    dist(1e9), chi2(rseg->chi), topp(0), tampered(false)
{ // Construct new instance from an entry in the RSEG table
  if (op>0.0) op = 1.0/op;
  else op = 0.0;
  for (int i=0; i<10; ++i) { ocovb[i]=rseg->cvf[i];  ocove[i]=rseg->cve[i]; }
  // zero off-diagonal p elements
  for (int i=10;i<14; ++i) ocovb[i] = ocove[i] = 0.0;
  float p4=op*op*op*op;
  ocovb[14] = rseg->cvf[14]*p4;
  ocove[14] = rseg->cve[14]*p4;
  // copy original start and finish points onto the variable ones
  p = op;
  xb=oxb; yb=oyb; zb=ozb; axb=oaxb; ayb=oayb;
  xe=oxe; ye=oye; ze=oze; axe=oaxe; aye=oaye;
  for (int i=0; i<15 ;++i) { covb[i]=ocovb[i];  cove[i]=ocove[i]; }
}
#endif



#ifndef GROVER_WITHOUT_ARTE
GroverParticle::GroverParticle(const ArtePointer<RTRA> rtra)
  : oxb(rtra->xf), oyb(rtra->yf), ozb(rtra->zf), 
    oaxb(rtra->txf), oayb(rtra->tyf), 
    op(fabs(rtra->pf)), mass(DEFAULTMASS), // <- note!
    nhit(rtra->hsi), type(1),
    usedinprimary(0), usedinsecondary(0), index(0),
    parRTRA(rtra), parRSEG(0), parRCCL(0), parMTRA(0),
    parGP(0), parGPL(0), parGV(0), 
    inlocal(0), inglobal(1), wprim(0.), wsec(0.),
    dist(1e9), chi2(rtra->chi), topp(0), tampered(false)
{ // Construct new instance from an entry in the RTRA table
  op = (op != 0.0) ? (fabs(1.0/op)) : 0.0;
  for (int i=0; i<10 ;++i) { ocove[i] = ocovb[i] = rtra->cvf[i];}
  // zero off-diagonal p elements
  for (int i=10;i<14; ++i) ocovb[i] = ocove[i] = 0.0;
  float p4=op*op*op*op;
  ocovb[14] = ocove[14] = rtra->cvf[14]*p4;
  if ((rtra->hin + rtra->hou + rtra->hsi) <= 0) q = 0; // no tracker hits
  else q = (rtra->pf<0.0) ? -1 : 1;
  // copy original start and finish points onto the movable ones
  xb=oxb; yb=oyb; zb=ozb; axb=oaxb; ayb=oayb;
  xe=oxe; ye=oye; ze=oze; axe=oaxe; aye=oaye;
  p = op;
  for (int i=0; i<15 ;++i) { covb[i]=ocovb[i];  cove[i]=ocove[i]; }
}
#endif



#ifndef GROVER_WITHOUT_ARTE
GroverParticle::GroverParticle(const ArtePointer<MTRA> mtra)
  : mass(mtra->mass), nhit(0), type(4),
    usedinprimary(0), usedinsecondary(0), index(0),
    parRTRA(0), parRSEG(0), parRCCL(0), parMTRA(mtra),
    parGP(0), parGPL(0), parGV(0),
    inlocal(0), inglobal(1),
    wprim(0.0), wsec(0.0), dist(0.0), chi2(0.0), chi2_m(0.0),
    topp(mtra->topp), tampered(false)
{ // Construct new instance from an entry in the MTRA table
  // *** NOT FINISHED!!!!!! ***
  // still needs proper setting of END parameters
  // zero errors
  for (int i=0;i<COVDIM;++i) covb[i]=ocovb[i]=cove[i]=ocove[i]=0.0;
  p = sqrt(mtra->px*mtra->px + mtra->py*mtra->py + mtra->pz*mtra->pz);
  // start point from parent MC vertex
  xb=oxb=mtra->bver->x;   yb=oyb=mtra->bver->y;   zb=ozb=mtra->bver->z;
  // start slopes
  axb = oaxb = (mtra->pz>0) ? (mtra->px/mtra->pz) : 0.0;   
  ayb = oayb = (mtra->pz>0) ? (mtra->py/mtra->pz) : 0.0;
  // end point from daughter MC vertex (if long-lived particle). Not done!
  if (mtra->leng > 0.0) {
    xe=oxe=oxb; ye=oye=oyb; ze=oze=ozb; axe=oaxe=oaxb; aye=oaye=oayb;
  }
  else {xe=oxe=oxb; ye=oye=oyb; ze=oze=ozb; axe=oaxe=oaxb; aye=oaye=oayb;}
  q = static_cast<int>(mtra->topp->chrg / 3.0);
  for (ArteRelationIterator<MIMP> imimp = mtra->mimp.begin();
       imimp != mtra->mimp.end(); ++imimp) {
    if ((*imimp)->gede->cmp == 1) {
      nhit += distance((*imimp)->hitb.begin(), (*imimp)->hitb.end());
    }
  } // is this correct? MIMP <=> HITB correspondence in VDS?
}
#endif



GroverParticle::GroverParticle(GroverVertex* vtx)
  : xb(FAILURE), yb(FAILURE), zb(FAILURE), axb(0.0), ayb(0.0),
    xe(FAILURE), ye(FAILURE), ze(FAILURE), axe(0.0), aye(0.0),
    oxb(FAILURE),oyb(FAILURE),ozb(FAILURE),oaxb(0.0),oayb(0.0),
    oxe(FAILURE),oye(FAILURE),oze(FAILURE),oaxe(0.0),oaye(0.0),
    p(0.0), op(0.0), mass(0.0), nhit(0), q(FAILURE),
    type(52), usedinprimary(0), usedinsecondary(0), index(0),
#ifndef GROVER_WITHOUT_ARTE
    parRTRA(0), parRSEG(0), parRCCL(0), parMTRA(0),
#endif
    parGP(0), parGPL(0), 
    parGV(vtx), inlocal(vtx->InLocal()), inglobal(vtx->InGlobal()),
    wprim(0.0), wsec(0.0), dist(0.0), chi2(0.0), chi2_m(0.0),
#ifndef GROVER_WITHOUT_ARTE
    topp(0), 
#endif
    tampered(false)
{
  if (vtx->Size()<1) { // screw-up type 1
    cout << "GroverParticle: ERROR - Constructed from empty GroverVertex!" << endl;
    for (int i=0; i<COVDIM; ++i) covb[i]=ocovb[i]=cove[i]=ocove[i]=0.0; 
  }
  else { 
    if (vtx->Momentum() == FAILURE) { // screw-up type 2
      cout<<"GroverParticle: ERROR - Constructed from unfitted GroverVertex!"<<endl;
      for (int i=0; i<COVDIM; ++i) covb[i]=ocovb[i]=cove[i]=ocove[i]=0.0;
    }
    else{
      mass = vtx->Mass();
      op   = p   = vtx->Momentum();
      oxb  = xb  = oxe = xe = vtx->X();
      oyb  = yb  = oye = ye = vtx->Y();
      ozb  = zb  = oze = ze = vtx->Z();
      oaxb = axb = oaxe = axe = vtx->SlopeX();
      oayb = ayb = oaye = aye = vtx->SlopeY();
      for (int i=0; i<COVDIM; ++i) covb[i] = ocovb[i] = 0.0;
      covb[0] = cove[0] = vtx->Cov(0);
      covb[2] = cove[2] = vtx->Cov(2);
      covb[5] = cove[5] = vtx->Cov(9);
      covb[9] = cove[9] = vtx->Cov(14);
      covb[14]= cove[14]= vtx->Cov(20);
      covb[8] = cove[8] = vtx->Cov(13);
      covb[12]= cove[12]= vtx->Cov(18);
      covb[13]= cove[13]= vtx->Cov(19);
      q=0;
      for (GroverParticleList::iterator ipar = vtx->particles.begin();
	   ipar!= vtx->particles.end(); ++ipar) q += ipar->Charge();
    }
  }
}



GroverParticle::~GroverParticle(void)
{ // Destructor 
}



void GroverParticle::Dump(void) const 
{ // Dump private members to stdout
  cout << "*** Grover: Dump() of GroverParticle object" << endl;
  DumpFormatter Nice;
  Nice.DumpClassName("GroverParticle");
  Nice.DumpComment("Part of the Grover vertexing package");
  Nice.DumpComment("Please send questions about this class "
		    "to the Grover authors");
  Nice.DumpHeader();
  Nice.DumpPrint("index",&index);
  Nice.DumpPrint("xb",&xb);
  Nice.DumpPrint("yb",&yb);
  Nice.DumpPrint("zb",&zb);
  Nice.DumpPrint("xe",&xe);
  Nice.DumpPrint("ye",&ye);
  Nice.DumpPrint("ze",&ze);
  Nice.DumpPrint("axb",&axb);
  Nice.DumpPrint("ayb",&ayb);
  Nice.DumpPrint("axe",&axe);
  Nice.DumpPrint("aye",&aye);
  Nice.DumpPrint("covb",covb,COVDIM);
  Nice.DumpPrint("cove",cove,COVDIM);
  Nice.DumpPrint("oxb",&oxb);
  Nice.DumpPrint("oyb",&oyb);
  Nice.DumpPrint("ozb",&ozb);
  Nice.DumpPrint("oxe",&oxe);
  Nice.DumpPrint("oye",&oye);
  Nice.DumpPrint("oze",&oze);
  Nice.DumpPrint("oaxb",&oaxb);
  Nice.DumpPrint("oayb",&oayb);
  Nice.DumpPrint("oaxe",&oaxe);
  Nice.DumpPrint("oaye",&oaye);
  Nice.DumpPrint("ocovb",ocovb,COVDIM);
  Nice.DumpPrint("ocove",ocove,COVDIM);
  Nice.DumpPrint("p",&p);
  Nice.DumpPrint("op",&op);
  Nice.DumpPrint("mass",&mass);
  Nice.DumpPrint("nhit",&nhit);
  Nice.DumpPrint("q",&q);
  Nice.DumpPrint("type",&type);
  Nice.DumpPrint("usedinprimary",&usedinprimary);
  Nice.DumpPrint("usedinsecondary",&usedinsecondary);
  Nice.DumpPrint("inlocal",&inlocal);
  Nice.DumpPrint("inglobal",&inglobal);
  Nice.DumpPrint("wprim",&wprim);
  Nice.DumpPrint("wsec",&wsec);
  Nice.DumpPrint("dist",&dist);
  Nice.DumpPrint("chi2",&chi2);
  Nice.DumpPrint("chi2_m",&chi2_m);
  Nice.DumpPrint("tampered",&tampered);
  Nice.DumpFooter();  
}



int GroverParticle::PropagateBeginTo(const float zplane = 0.0) 
{
  //
  // Propagate track to a given x-y plane
  // from Rover 1.05
  //
  // I.Kisel/D.Emeliyanov
  //

  if (zb == zplane) return SUCCESS;

  // Reset parameters unless user has explicitly changed them (prevents possible
  // blow-up of covariance elements if (for some reason) the user propagates
  // the particle backwards and forwards many times.
  if (!this->Tampered()) this->ResetParameters();

  float dz = zplane - zb;  // distance to current start of track

  bool  MC            = (this->Type() == 4);
  bool  CATSsegment   = (this->Type() == 2);
  bool  near_vds      = (zplane < 250.0 && zplane > -50.0);
  bool  tiny_distance = (fabs(dz) < 5.0);

  bool orig_global = this->InGlobal();

  if (orig_global) this->GlobalToLocal();

  //  bool  useCATSprop   = (CATSsegment       && 
  //			 this->InLocal()   && 
  //			 near_vds          && 
  //			 !tiny_distance      );

  bool  useCATSprop   = (this->InLocal()   && 
			 near_vds          && 
			 !tiny_distance      );

#ifdef GROVER_WITHOUT_ARTE  
  useCATSprop = false;
#endif

  if (useCATSprop) {
#ifndef GROVER_WITHOUT_ARTE    
    float nxb=xb, nyb=yb, nzb=zb, naxb=axb, nayb=ayb;
    float ninvp = (p!=0.0) ? 1.0/p : 0.0;
    float ncov[15]={covb[0], covb[1], covb[2], covb[3], covb[4],
		    covb[5], covb[6], covb[7], covb[8], covb[9],
		    covb[10],covb[11],covb[12],covb[13],covb[14]};
    GroverCatsExtrap(zplane, &nxb, &nyb, &nzb, &naxb, &nayb, &ninvp, ncov);
    xb=nxb; yb=nyb; zb=nzb; axb=naxb; ayb=nayb;
    p = (ninvp != 0.0) ? 1.0/ninvp : 0.0;
    for (int i=0;i<15;++i) covb[i] = ncov[i];
    zb = zplane;
    covb[10] += dz * covb[12];
    covb[11] += dz * covb[13];
#endif
  }
  else {
    xb  += dz*axb;
    yb  += dz*ayb;
    zb  =  zplane;  
    if (!MC) {
      covb[0]  += dz * (dz*covb[5] + covb[3] + covb[3]);
      covb[1]  += dz * (dz*covb[8] + covb[6] + covb[4]);
      covb[2]  += dz * (dz*covb[9] + covb[7] + covb[7]);
      covb[3]  += dz * covb[5]; 
      covb[4]  += dz * covb[8]; 
      covb[6]  += dz * covb[8];
      covb[7]  += dz * covb[9];
      covb[10] += dz * covb[12];
      covb[11] += dz * covb[13];
    }
  }

  if (orig_global) this->LocalToGlobal();
  
  return SUCCESS;
}



int GroverParticle::PropagateEndTo(const float zplane = 2000.) 
{
  // Very simplistic! Assumes the track is already heading out of the VDS and
  // simply propagates the track as a straight line, taking no account of
  // multiple scattering or magnetic field.

  // Reset parameters unless user has explicitly changed them (prevents possible
  // blow-up of covariance elements if (for some reason) the user propagates
  // the particle backwards and forwards many times.
  if (!this->Tampered()) this->ResetParameters();

  if (ze == zplane) return SUCCESS;
  float dz = zplane - ze;  // distance to current end of track

  bool MC = (this->Type() == 4);

  xe  = xe + dz*axe;
  ye  = ye + dz*aye;
  ze  = zplane;  
  if (!MC) {
    cove[0]  += dz * (dz*cove[5] + cove[3] + cove[3]);
    cove[1]  += dz * (dz*cove[8] + cove[6] + cove[4]);
    cove[2]  += dz * (dz*cove[9] + cove[7] + cove[7]);
    cove[3]  += dz * cove[5]; 
    cove[4]  += dz * cove[8]; 
    cove[6]  += dz * cove[8];
    cove[7]  += dz * cove[9];
    cove[10] += dz * cove[12];
    cove[11] += dz * cove[13];
  }

  return SUCCESS;
}



int GroverParticle::GlobalToLocal(void) {

  if (inglobal && !inlocal) {

#ifndef GROVER_WITHOUT_ARTE
  // Covariance unchanged (not a bad approximation, in fact)
    float oldb[5], olde[5], newb[5], newe[5];
    float ooldb[5], oolde[5], onewb[5], onewe[5];
    oldb[0] = xb; oldb[1] = yb; oldb[2] = zb; oldb[3] = axb; oldb[4] = ayb;
    olde[0] = xe; olde[1] = ye; olde[2] = ze; olde[3] = axe; olde[4] = aye;
    ooldb[0] = oxb;  ooldb[1] = oyb; ooldb[2] = ozb; 
    ooldb[3] = oaxb;  ooldb[4] = oayb;
    oolde[0] = oxe;  oolde[1] = oye; oolde[2] = oze; 
    oolde[3] = oaxe;  oolde[4] = oaye;
    acGlobalToLocal(newb, oldb, 1);
    acGlobalToLocal(newe, olde, 1);
    acGlobalToLocal(onewb, ooldb, 1);
    acGlobalToLocal(onewe, oolde, 1);
    xb = newb[0]; yb = newb[1]; zb = newb[2]; axb = newb[3]; ayb = newb[4];
    xe = newe[0]; ye = newe[1]; ze = newe[2]; axe = newe[3]; aye = newe[4];
    oxb = onewb[0];   oyb = onewb[1]; ozb = onewb[2]; 
    oaxb = onewb[3]; oayb = onewb[4];
    oxe = onewe[0];   oye = onewe[1]; oze = onewe[2]; 
    oaxe = onewe[3]; oaye = onewe[4];
#endif

    inglobal = false; inlocal = true;
  }

  return SUCCESS;
}



int GroverParticle::LocalToGlobal(void) {

  // Covariance unchanged (not a bad approximation, in fact)
  if (inlocal && !inglobal) {

#ifndef GROVER_WITHOUT_ARTE
    float oldb[5],  olde[5],  newb[5],  newe[5];
    float ooldb[5], oolde[5], onewb[5], onewe[5];
    oldb[0] = xb; oldb[1] = yb; oldb[2] = zb; oldb[3] = axb; oldb[4] = ayb;
    olde[0] = xe; olde[1] = ye; olde[2] = ze; olde[3] = axe; olde[4] = aye;
    ooldb[0] = oxb;  ooldb[1] = oyb;  ooldb[2] = ozb;
    ooldb[3] = oaxb; ooldb[4] = oayb;
    oolde[0] = oxe;  oolde[1] = oye;  oolde[2] = oze;
    oolde[3] = oaxe; oolde[4] = oaye;
    acLocalToGlobal(oldb,  newb,  1);
    acLocalToGlobal(olde,  newe,  1);
    acLocalToGlobal(ooldb, onewb, 1);
    acLocalToGlobal(oolde, onewe, 1); 
    xb = newb[0]; yb = newb[1]; zb = newb[2]; axb = newb[3]; ayb = newb[4];
    xe = newe[0]; ye = newe[1]; ze = newe[2]; axe = newe[3]; aye = newe[4];
    oxb  = onewb[0]; oyb  = onewb[1]; ozb = onewb[2];
    oaxb = onewb[3]; oayb = onewb[4];
    oxe  = onewe[0]; oye  = onewe[1]; oze = onewe[2];
    oaxe = onewe[3]; oaye = onewe[4];
#endif

    inlocal = false; inglobal = true;
  }
  return SUCCESS; 
}



int GroverParticle::ResetParameters(void)
{ // Reset all the track parameters to their 
  // original (constructed) values. BUT remember that a coordinate 
  // transformation changes these original values too. If you want the 
  // *really* original parameters, follow the ParentXX() pointer and 
  // construct a new particle (or transform back, although I do not 
  // guarantee that this is a unitary transformation as it uses 
  // acLocalToGlobal.C). 
  xb  = oxb;  yb  = oyb; zb = ozb;
  xe  = oxe;  ye  = oye; ze = oze;
  axb = oaxb; ayb = oayb;
  axe = oaxe; aye = oaye;
  p = op;
  for (int i=0;i<COVDIM;++i)   {covb[i]=ocovb[i];  cove[i]=ocove[i];}
  tampered = false;
  return SUCCESS;
}



float GroverParticle::DistanceTo(const GroverWire* wire) const
{
  //
  // 24.07.00  V.Kekelidze
  //
  float Dist = 0.0;
  float ax = axb;
  float ay = ayb;
  float fact = sqrt(ax*ax+ay*ay+1);
  float dz = this->BeginZ() - wire->Z();

  if(wire->Horizontal() == 1){
    float dy = this->BeginY() - wire->Y();
    Dist = fabs(dy-dz*ay)*fact/(1+ay*ay);
  }
  else if(wire->Vertical() == 1){
    float dx = this->BeginX() - wire->X();
    Dist = fabs(dz*ax-dx)*fact/(1+ax*ax);
  }
  return Dist;
}



float GroverParticle::Chi2DistTo(const GroverWire* wire) const 
{
  // Simple-minded 1D chi2 assuming infinitely long wires! 
  // If you don't like it, please write something better ;-)
  bool MC = (this->Type() == 4);
  if (MC) return 1e9;
  GroverParticle tmp = *this;
  tmp.PropagateBeginTo(wire->Z());
  double d = 0.0, s2 = 0.0;
  if (wire->Horizontal() == 1) {
    d  = tmp.BeginY() - wire->Y();
    s2 = tmp.BeginCov(2);
  }
  else if (wire->Vertical() == 1) {
    d  =  tmp.BeginX() - wire->X();
    s2 =  tmp.BeginCov(0);
  }
  float dist = (s2 > 0.0) ? static_cast<float>(sqrt(d*d/s2)) : 1e9;
  return dist;
}



float GroverParticle::DistanceTo(const GroverParticle* that) const 
{
  //
  // 20.07.00   D.Emeliyanov
  //
  GroverParticle p1 = *this, p2 = *that; // local copies
  p1.PropagateBeginTo(0.0);
  p2.PropagateBeginTo(0.0);
  float x1 = p1.BeginX(),      x2 = p2.BeginX();
  float y1 = p1.BeginY(),      y2 = p2.BeginY();
  float ax1= p1.BeginSlopeX(), ax2= p2.BeginSlopeX();
  float ay1= p1.BeginSlopeY(), ay2= p2.BeginSlopeY();
  float xy = ax1*ay2 - ay1*ax2;
  float sq_sum = (ax1-ax2)*(ax1-ax2) + (ay1-ay2)*(ay1-ay2) + xy*xy;
  float UP = (x2-x1)*(ay1-ay2) + (y2-y1)*(ax2-ax1);
  return  fabs(UP/sqrt(sq_sum));
}



float GroverParticle::Chi2DistTo(const GroverParticle* that) const 
{
  //
  // 19.07.00   D.Emeliyanov/V.Kekelidze
  //
  if (this->Type() == 4 && that->Type() == 4) return 1e9;
  GroverParticle p1 = *this, p2 = *that;
  p1.PropagateBeginTo(0.0);
  p2.PropagateBeginTo(0.0);
  float x1=p1.BeginX(),       x2=p2.BeginX();
  float y1=p1.BeginY(),       y2=p2.BeginY();
  float ax1=p1.BeginSlopeX(), ax2=p2.BeginSlopeX();
  float ay1=p1.BeginSlopeY(), ay2=p2.BeginSlopeY();
  float xy = ax1*ay2 - ay1*ax2;
  float sq_sum = (ax1-ax2)*(ax1-ax2)+(ay1-ay2)*(ay1-ay2)+xy*xy;
  float UP = (x2-x1)*(ay1-ay2)+(y2-y1)*(ax2-ax1);
  float dist = fabs(UP/sqrt(sq_sum));
  if (dist > 1e6) return 1e9;
  float Cov[8][8];
  for(int i=0; i<8; ++i) for(int j=i; j<8; ++j) Cov[i][j]=Cov[j][i]=0.0;

  int k=0;
  for(int i=0; i<4; ++i) {
    for(int j=0; j<=i; ++j){
      Cov[i][j]     = Cov[j][i]     = p1.BeginCov(k);
      Cov[i+4][j+4] = Cov[j+4][i+4] = p2.BeginCov(k);
      ++k;
    }
  }
  float Jac[8]={0.0};
  float sssum = sqrt(sq_sum);
  float sssum_times_sq_sum = sssum*sq_sum;
  Jac[0] = (ay2-ay1)/sssum;
  Jac[1] = (ax1-ax2)/sssum;
  Jac[2] = ((y1-y2)/sssum - UP*(ax1-ax2+ay2*xy))/sssum_times_sq_sum;
  Jac[3] = ((x2-x1)/sssum + UP*(ay1-ay2-ax2*xy))/sssum_times_sq_sum;
  Jac[4] = -Jac[0];
  Jac[5] = -Jac[1];
  Jac[6] = ((y2-y1)/sssum + UP*(ax2-ax1+ay1*xy))/sssum_times_sq_sum;
  Jac[7] = ((x1-x2)/sssum - UP*(ay2-ay1+ax1*xy))/sssum_times_sq_sum;
  float sum=0.0, sigma=0.0;
  for(int i=0; i<8; ++i){
    sum = 0.0;
    for(int j=0; j<8; ++j)  sum += Jac[j]*Cov[i][j];
    sigma += Jac[i]*sum;
  }
  return (sigma != 0.0) ? dist*dist/sigma : 1e9;
}



float GroverParticle::DistanceTo(const GroverVertex *vert) const 
{
  //
  // 20.07.00   D.Emeliyanov/V.Kekelidze
  // closest spatial distance to a vertex
  //
  float x0 = this->BeginX();
  float y0 = this->BeginY();      
  float ax = this->BeginSlopeX();
  float ay = this->BeginSlopeY();
  float xv = vert->X();
  float yv = vert->Y();
  float zv = vert->Z();
  float Sz = ax*ax + ay*ay + 1;
  float dx = x0-xv,    dy=y0-yv,     dz=this->BeginZ()-zv;
  float d1 = dy-dz*ay, d2=-dx+dz*ax, d3=dx*ay-dy*ax;
  float UP = d1*d1 + d2*d2 + d3*d3;
  return (UP >= 0.0) ? sqrt(UP/Sz) : 1e9;
}



float GroverParticle::Chi2DistTo(const GroverVertex* vert) const 
{
  //
  // 20.07.00   D.Emeliyanov/V.Kekelidze
  //
  bool MCver = false;
  int nmctr = 0;
  for (GroverParticleList::const_iterator ip = vert->particles.begin();
       ip != vert->particles.end(); ++ip) {
    if (ip->Type() == 4) ++nmctr;
  }
  MCver = (nmctr == vert->Size() && nmctr > 0);
  bool MCtrk = (this->Type() == 4);
  if (MCtrk && MCver) return 1e9;

  float x0 = this->BeginX();
  float y0 = this->BeginY();      
  float ax = this->BeginSlopeX();
  float ay = this->BeginSlopeY();
  float xv = vert->X();
  float yv = vert->Y();
  float zv = vert->Z();
  float Sz = ax*ax + ay*ay + 1;
  float dx=x0-xv,    dy=y0-yv,     dz=this->BeginZ()-zv;
  float d1=dy-dz*ay, d2=-dx+dz*ax, d3=dx*ay-dy*ax;
  float UP = d1*d1 + d2*d2 + d3*d3;
  float dist = sqrt(UP/Sz);
  if (dist>1e6) return 1e9;

  float Cov[7][7];
  for(int i=0; i<7; ++i) for(int j=i; j<7; j++) Cov[i][j]=Cov[j][i]=0.0;
  int k=0;
  for(int i=0;i<4; ++i) {
    for(int j=0;j<=i;++j){
      Cov[i][j] = Cov[j][i] = this->BeginCov(k);
      ++k;
    }
  }
  k=0;
  for(int i=4;i<7; ++i) {
    for(int j=4;j<=i;++j){
      Cov[i][j] = Cov[j][i] = vert->Cov(k);
      ++k;
    }
  }
  float Jac[7]={0.0};
  Jac[0] = (ay*d3 - d2)/(Sz*dist);
  Jac[1] = (d1 - ax*d3)/(Sz*dist);
  Jac[2] = ((dz*d2 - dy*d3)*Sz - UP*ax) / (dist*Sz*Sz);
  Jac[3] = ((dx*d3 - dz*d1)*Sz - UP*ay) / (dist*Sz*Sz);
  Jac[4] = -Jac[0];
  Jac[5] = -Jac[1];
  Jac[6] = (ay*d1 - ax*d2)/(Sz*dist);
  float sum=0.0, sigma=0.0;
  for(int i=0; i<7; ++i){
    sum=0;
    for(int j=0; j<7; ++j)  sum += Jac[j]*Cov[i][j];
    sigma += Jac[i]*sum;
  }

  return (sigma != 0.0) ? dist*dist/sigma : 1e9;
}



float GroverParticle::DistanceTo(const float xx, const float yy, const float zz) const 
{
  //
  //  21.07.00  V.Kekelidze
  //
  float x0  = this->BeginX();
  float y0  = this->BeginY();      
  float ax0 = this->BeginSlopeX();
  float ay0 = this->BeginSlopeY();
  float Sz = ax0*ax0 + ay0*ay0 + 1;
  float dx = x0-xx,     dy=y0-yy,      dz=this->BeginZ()-zz;
  float d1 = dy-dz*ay0, d2=-dx+dz*ax0, d3=dx*ay0-dy*ax0;
  float UP = d1*d1 + d2*d2 + d3*d3;
  return sqrt(UP/Sz);
}



float GroverParticle::Chi2DistTo(const float xx, const float yy, const float zz) const
{
  //
  // V.Kekelidze 21.07.00
  //
  bool MC = (this->Type() == 4);
  if (MC) return 1e9;
  float x0=this->BeginX();
  float y0=this->BeginY();      
  float ax0=this->BeginSlopeX();
  float ay0=this->BeginSlopeY();
  float Sz=ax0*ax0+ay0*ay0+1;
  float dx=x0-xx, dy=y0-yy, dz=this->BeginZ()-zz;
  float d1=dy-dz*ay0, d2=-dx+dz*ax0, d3=dx*ay0-dy*ax0;
  float UP=d1*d1+d2*d2+d3*d3;
  float dist=sqrt(UP/Sz);
  if (dist>1e6) return 1e9;
  float Cov[7][7];
  for(int i=0; i<4; ++i) for(int j=i; j<4; ++j) Cov[i][j]=Cov[j][i]=0.0;
  int k=0;
  for(int i=0;i<4; ++i) {
    for(int j=0;j<=i;++j){
      Cov[i][j] = Cov[j][i] = this->BeginCov(k);
      ++k;
    }
  }
  k=0;
  float Jac[7]={0.0};
  Jac[0] = (ay0*d3-d2)/(Sz*dist);
  Jac[1] = (d1-ax0*d3)/(Sz*dist);
  Jac[2] = ((dz*d2-dy*d3)*Sz-UP*ax0)/(dist*Sz*Sz);
  Jac[3] = ((dx*d3-dz*d1)*Sz-UP*ay0)/(dist*Sz*Sz);

  float sum=0.0, sigma=0.0;
  for(int i=0; i<4; ++i){
    sum=0;
    for(int j=0; j<4; ++j){
      sum += Jac[j]*Cov[i][j];
    }
    sigma += Jac[i]*sum;
  }
  return (sigma != 0.0) ? dist*dist/sigma : 1e9;
}



#ifndef GROVER_WITHOUT_ARTE
int GroverParticle::SetMass(const ArtePointer<TOPP> tp)
{ // set mass to whatever TOPP is pointing at
  if (tp == 0) return FAILURE;
  else {
    topp = tp;
    return this->SetMass(topp->mass);
  }
}
#endif



float GroverParticle::BeginCov(const int i, const int j) const
{ // square access method
  const int dim = 6;
  const int sq[dim][dim] = {
    { 0,  1,  3,  6, 10, 15},
    { 1,  2,  4,  7, 11, 16},
    { 3,  4,  5,  8, 12, 17},
    { 6,  7,  8,  9, 13, 18},
    {10, 11, 12, 13, 14, 19},
    {15, 16, 17, 18, 19, 20}
  };
  if (i<0 || j<0 || i>=dim || j>=dim) return FAILURE;
  else return this->BeginCov(sq[i][j]);
}



float GroverParticle::EndCov(const int i, const int j) const
{ // square access method
  const int dim = 6;
  const int sq[dim][dim] = {
    { 0,  1,  3,  6, 10, 15},
    { 1,  2,  4,  7, 11, 16},
    { 3,  4,  5,  8, 12, 17},
    { 6,  7,  8,  9, 13, 18},
    {10, 11, 12, 13, 14, 19},
    {15, 16, 17, 18, 19, 20}
  };
  if (i<0 || j<0 || i>=dim || j>=dim) return FAILURE;
  else return this->EndCov(sq[i][j]);
}




/////////////////////////////////////////////////////////////////
//
// Operators
//

bool operator==(const GroverParticle& a, const GroverParticle& b)
{
  int t = a.Type();
  if (t != b.Type()) return false;
  bool flag = false;
  switch (t) {

#ifndef GROVER_WITHOUT_ARTE
  case 1: {
    flag = (a.ParentRTRA() == b.ParentRTRA());
    break;
  }
  case 2: {
    flag = (a.ParentRSEG() == b.ParentRSEG());
    break;
  }
  case 3: {
    flag = (a.ParentRCCL() == b.ParentRCCL());
    break;
  }
  case 4: {
    flag = (a.ParentMTRA() == b.ParentMTRA());
    break;
  }
#endif

  case 50: {
    flag = (a.ParentGP()   == b.ParentGP());
    break;
  }
  case 51: {
    flag = (a.ParentGPL()  == b.ParentGPL());
    break;
  }
  case 52: {
    flag = (a.ParentGV()   == b.ParentGV());
    break;
  }
  default: {
    flag = false;
    break;
  }
  }
  return flag;
}



bool operator!=(const GroverParticle& a, const GroverParticle& b)
{
  return !(a==b);
}



ostream& operator<<(ostream& s, GroverParticle& p)
{
  s << "[" 
    << "idx=" << p.Index()
    << " x,y,z={" << p.BeginX()      << "," << p.BeginY()      << "," << p.BeginZ() << "}"
    << " ax,ay={" << p.BeginSlopeX() << "," << p.BeginSlopeY() << "}"
    << " p,m,q={" << p.Momentum()    << "," << p.Mass()        << "," << p.Charge() << "}";
  s << " nh=" << p.Nhit();
  if (p.InLocal()) s << " crd=LOC";
  else s << " crd=GLO";
  s << " type=";
  switch (p.Type()) {
  case 1:  { s << "RTRA"; break;}
  case 2:  { s << "RSEG"; break;}
  case 3:  { s << "RCCL"; break;}
  case 4:  { s << "MTRA"; break;}
  case 50: { s << "GP";   break;}
  case 51: { s << "GPL";  break;}
  case 52: { s << "GV";   break;}
  default: { s << "???";  break;}
  }
  s << "]";
  return s;
}



/////////////////////////////////////////////////////////////////
//
//  Vector functions (V.Kekelidze 15/06/00)
//

inline float scal(float v1[],float v2[])
{
  float scalar = 0.0;
  for(int i=0; i<3; i++) scalar+=v1[i]*v2[i];
  return scalar;
}

inline void res(float v1[],float v2[], float R[])
{
  for(int i=0; i<3; i++) R[i] = v1[i] - v2[i];
} 

inline void sum(float v1[], float v2[], float S[])
{
  for(int i=0; i<3; i++) S[i] = v1[i] + v2[i];
}

inline void mult_n(float v1[], float n, float M[])
{
  for(int i=0; i<3; i++) M[i] = v1[i] * n;
}

