#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#endif

#include "grover/grover.hh"

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// ------------------------------------------------------------------------
// 01-04-26 Michael   Comments if bad wires are found / if verbose
// 01-05-07 Michael   Make it usable for MC too.

bool ReasonableWires(const GroverWireList* wires, const int period = 1) 
{
  /////////////////////////////////////////////////////////////////////////////////
  //
  // Returns true if this list of wires looks reasonable for the
  // HERA-B data taking in 2000. Based on look-up table of usual
  // positions for each wire and TOLERANCE numbers which define
  // what 'reasonable' means. The second argument can be used to 
  // divide the data taking into blocks with similar wire positions,
  // but not necessary at the moment!
  //
  /////////////////////////////////////////////////////////////////////////////////

  // special bail-out
  const bool verbose = (ggCuts.Verbose == 1);

  if (ggCuts.GEFindSecondaries == 42) return true;

  if (wires->empty()) {
    if (verbose) cout << "*** ReasonableWires: wire number: wires are empty!" << endl;
    return false;
  }
  
  bool found_a_bad_wire = false;

  static int evno = 0;
  if (period > 1 && ++evno < 25) cout 
    << "*** ReasonableWires: WARNING: non-default periods not implemented!" 
    << endl;


  /////////////////////////////////////////////////////////////////////////////////
  // Limits for deviations from the expected positions
  const float RADIAL_TOLERANCE       = 0.3000; // The beam moves.                 
  const float TANGENTIAL_TOLERANCE   = 0.5000; // Not so critical anyway.             
  const float LONGITUDINAL_TOLERANCE = 1.0000; // z-positions don't seem to change
  /////////////////////////////////////////////////////////////////////////////////



  /////////////////////////////////////////////////////////////////////////////////
  // Common wire positions in the 2000 data (global coordinates) as found by the      
  // Grover wire following mechanism. Some wires (?) were never used.                 
  // Their values are estimates! The beam spot is at (0.300,-0.050)
  //                                                                                  
  //             1        2        3        4        5        6        7        8  
  //         
  const static float usual_x_2000[9] =
  {0.0,      0.300,   0.300,   0.320,   0.040,   0.300,   0.300,   0.320,   0.040};
  const static float usual_y_2000[9] =
  {0.0,      0.066,  -0.255,  -0.050,  -0.050,   0.066,  -0.255,  -0.050,  -0.050};
  const static float usual_z_2000[9]=
  {0.0,    - 1.350,  -1.440,  -0.880,  -1.060,  -5.440,  -5.300,  -4.880,  -4.940};


  const static float mc_x_2000[9] =
  {0.0,      0.500,   0.100,   0.300,   0.30,   0.500,   0.100,   0.300,   0.3000};;
  const static float mc_y_2000[9] =
  {0.0,      0.000,   0.000,  -0.200,   0.200,  0.000,    0.000,  -0.200,    0.200};
  const static float mc_z_2000[9]=
  {0.0,    - 4.800,  -5.100,  -5.700,  -5.600,  -0.900,  -1.000,  -1.400,  -1.600};
  
  /////////////////////////////////////////////////////////////////////////////////

  //
  // You can add more here if necessary
  // const static float usual_x_in_some_period[9] = {...}; etc...
  //

  const float* expected_x = usual_x_2000;
  const float* expected_y = usual_y_2000;
  const float* expected_z = usual_z_2000;

  // select which default positions to use
  if (period == 1) {
    if (ArteTable<MCIN>::booked()) {
      expected_x = mc_x_2000;
      expected_y = mc_y_2000;
      expected_z = mc_z_2000;
    } else {
      expected_x = usual_x_2000;
      expected_y = usual_y_2000;
      expected_z = usual_z_2000;
    }
  }
  //
  // else if (period == 2) {
  //   expected_x = usual_x_in_some_period;
  //   ... etc ...
  // }
  //
  else {
    //default - please do not change
    if (ArteTable<MCIN>::booked()) {
      expected_x = mc_x_2000;
      expected_y = mc_y_2000;
      expected_z = mc_z_2000;
    } else {
      expected_x = usual_x_2000;
      expected_y = usual_y_2000;
      expected_z = usual_z_2000;
    }
  }
      

  // Make the decision
  for (GroverWireList::const_iterator w = wires->begin();
       w != wires->end(); ++w) {
    bool bad_wire = false;
    GroverWire wire = *w;
    wire.LocalToGlobal();
    int wnr = wire.Number();
    if (wnr < 1 || wnr > 8) {
      if (verbose) cout << "*** ReasonableWires: wire number: wnr < 1 || wnr > 8" << endl;
      bad_wire = true;
    }
    else {
      float Dx = fabs(wire.X() - expected_x[wnr]);
      float Dy = fabs(wire.Y() - expected_y[wnr]);
      float Dz = fabs(wire.Z() - expected_z[wnr]);
      if (wire.Vertical()) {
	if (Dx > RADIAL_TOLERANCE     || 
	    Dy > TANGENTIAL_TOLERANCE || 
	    Dz > LONGITUDINAL_TOLERANCE ) {
	  if (verbose) cout << 
			 "*** ReasonableWires: vertical position wrong" << endl;
	  bad_wire = true;
	}
      }
      else if (wire.Horizontal()) {
	if (Dx > TANGENTIAL_TOLERANCE || 
	    Dy > RADIAL_TOLERANCE     || 
	    Dz > LONGITUDINAL_TOLERANCE ) {
	  if (verbose) cout << 
			 "*** ReasonableWires: horizontal position wrong" << endl;
	  bad_wire = true;
	}
      }
      else {
	if (verbose) cout << 
		       "*** ReasonableWires: wire not vertical and not horizontal" << endl;
	bad_wire = true;
      }
    }
    if (bad_wire) {
      found_a_bad_wire = true;
      if (verbose) cout << "*** ReasonableWires: wire number "
			<< wnr << " bad" << endl;
    }
  }


  return !found_a_bad_wire;
}

