#ifndef _NOARTE_

// ----------------------------------------------------------------------
//
// Tiny functions to get information from the EVHD table
//
// Author: Rainer Wanke, 18-Sep-1997
//
// ----------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <time.h>

// Arte banks,  C++ version
#include "arte/arte.hh"

// includes for output messages
extern "C" {
#include "iol.h"
#include "aframelib.h"
}
#include "grr.h"

// prototype declarations
#include "ITR_EVinfo.h" 

// ----------------------------------------------------------------------
//
// Functions to get the event number from the EVHD table
//
// Inputs:       None
//
// Return value:  event number (=0, if EVHD does not exist)
//
// Author: Rainer Wanke, 18-Sep-1997
//
// ----------------------------------------------------------------------
// First function for C++ calls,
int ITR_getEventNumber()
{

  char *routine = "ITR_getEventNumber";

  // take event # from EVHD table
  if ( !ArteTable<EVHD>::size() ) {      // check for existence of EVHD
    // sprintf(mess,"EVHD table not existent.");
    aframe_mess(routine,GRR_SINFO,"EVHD table not existent.");
    return 0;
  }

  return ArtePointer<EVHD>(1)->evt;
}


// Second function for calls from C programs
extern "C"
{

int ITR_getEventNumber_C()
{
  return ITR_getEventNumber();
}

// end of 'extern C'
}


// ----------------------------------------------------------------------
//
// Functions to get the run number from the EVHD table
// If EVHD is non-exsistent (or contains run # = 0), the run # is taken
// from the DDAQ bank and filled into EVHD.
//
// Inputs:        None
//
// Return value:  Run number
//
// Author: Rainer Wanke, 19-Sep-1997
//
// ----------------------------------------------------------------------
// the first for C++ calls, the second for calls from C programs
int ITR_getRunNumber()
{
  char *routine = "ITR_getRunNumber";

  // take run number from EVHD
  if ( !ArteTable<EVHD>::size() ) {      // check for existence of EVHD
    // sprintf(mess,"EVHD table not existent.");
    aframe_mess(routine,GRR_SINFO,"EVHD table not existent.");
    return 0;
  }

  ArtePointer<EVHD> evhd = ArtePointer<EVHD>(1);
  if ( evhd->run != 0)     // check run # in EVHD
      return evhd->run;
  // sprintf(mess,"EVHD table existent, but run number is 0.");
  aframe_mess(routine,GRR_SINFO,"EVHD table existent, but run number is 0.");

  // take run number from DDAQ
  ArtePointer<DDAQ> ddaq(2);
  int run = *((int*)ddaq->bits);
  // sprintf(mess,"Filling DDAQ run number %d into EVHD table.", run);
  aframe_mess(routine,GRR_SINFO,"Filling DDAQ run number into EVHD table.");

  // fill EVHD run #
  evhd->run = run;

  return evhd->run;
}


extern "C"
{

int ITR_getRunNumber_C()
{
  return ITR_getRunNumber();
}

/* end of 'extern C' */
}

// ----------------------------------------------------------------------
//
// Functions to get the event type from the EVHD table
//
// Inputs:       None
//
// Return value:  0: if real data
//                1: MC data
//               -1: Don't know (experiment # = 0 in EVHD table)
//               -2: EVHD table not existent
//
// Author: Rainer Wanke, 17-Sep-1997
//
// ----------------------------------------------------------------------

// First function for C++ calls,
int ITR_CheckMC()
{
  // take experiment # from EVHD table
  if ( ! ArteTable<EVHD>::size() )     // check for existence of EVHD
   return -2;

  ArtePointer<EVHD> evhd = ArtePointer<EVHD>(1);
  if ( evhd->exp > 0 )       // real data
    return 0;
  if ( evhd->exp <= 0 )  // MC data
    return 1;
  return -1;
  
}


// Second function for calls from C programs
extern "C"
{

int ITR_CheckMC_C()
{
  return ITR_CheckMC();
}

// end of 'extern C'
}

// ----------------------------------------------------------------------
//
// Functions to get the unix-time (in seconds) from the EVHD table
//
// Inputs:       None
//
// Return value:  unix-time in seconds after Jan 1, 1970
//                (= -1, if EVHD does not exist)
//
// Author: Rainer Wanke, 5-Nov-1997
//
// ----------------------------------------------------------------------
// First function for C++ calls
int ITR_getUTime()
{
  // get date and time from EVHD table
  if ( ! ArteTable<EVHD>::size() )       // check for existence of EVHD
   return -1;

  // Do date and time have sensible values ?
  ArtePointer<EVHD> evhd = ArtePointer<EVHD>(1);
  int date = evhd->date;
  int time = evhd->time;
  if ( date <= 0 || time < 0 )
   return -1;

  // fill tm structure from <time.h>
  struct tm mytime;
  //  cout << "Date = " << date << endl;
  //  cout << "Time = " << time << endl;
  mytime.tm_hour = (int) floor( time / 10000 );
  //  cout << "Hour = " << mytime.tm_hour << endl;
  time -= 10000 * mytime.tm_hour;
  mytime.tm_min = (int) floor( time / 100 );
  //  cout << "Min  = " << mytime.tm_min  << endl;
  time -= 100 * mytime.tm_min;
  mytime.tm_sec = time;
  //  cout << "Sec  = " << mytime.tm_sec  << endl;
  mytime.tm_year = (int) floor( date / 10000 );
  //  cout << "Year = " << mytime.tm_year << endl;
  date -= 10000 * mytime.tm_year;
  mytime.tm_mon = (int) floor( date / 100 ) - 1;
  //  cout << "Mon  = " << mytime.tm_mon  << endl;
  date -= 100 * ( mytime.tm_mon + 1 );
  mytime.tm_mday = date;
  //  cout << "Day  = " << mytime.tm_mday << endl;

  // take care of centennial switch
  if ( mytime.tm_year < 80 )   mytime.tm_year += 100;

  // convert to start unix-time
  int utime = mktime( &mytime );

  //  cout << "Converted  date " << evhd->date << " and time " << evhd->time
  //       << " to Unix time " << utime << endl;

  return utime;
}


// Second function for calls from C programs
extern "C"
{

int ITR_getUTime_C()
{
  return ITR_getUTime();
}

// end of 'extern C'
}

#endif   /* of #ifndef _NOARTE_ */

