
// *****************************************************************************
//
// source:    l4tEvnt.C
//
// type:      source code
//
// created:   14 Jan 1998
//
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// symbol:    l4tEvnt
//
// purpose:   finally select events (L4 trigger)
//
// called by: anywhere after rcevnt_()
//
// hints:     classification is done in rcClassify()
//            L4TKUM: -1: not set => write out ALL events (ignore L4TSET)
//                     0: 4LT off => reject ALL events
//                    >0: 4LT on  => use class mask
//
//            L4TSET:  0: 4LT off      => disable kumac => store all events
//                     1: 4LT on/unset => enable  kumac => evaluate kumac
//
// changes:
// 25 Nov 1999 (AG) move to l4tEvnt()
// 22 Feb 2000 (AG) handle various settings
// 13 Mar 2000 (AG) remove some printout
// 04 Apr 2000 (AG) return trigger result
// 27 Apr 2000 (AG) use ARTE_mark()
// 05 May 2000 (AG) set EDIR flag accoring to flagdst
// 08 May 2000 (AG) only flagdst needs to be set
// 16 May 2000 (AG) set flag = flagdst
// 24 May 2000 (AG) set classification bit for random triggers (8 + 9)
// 25 May 2000 (AG) set classification bit for FLT trigger (0 + 8)
// 29 May 2000 (AG) comment some printout
// 14 Jun 2000 (AG) polishing
// 16 Jun 2000 (AG) polishing
// 20 Jun 2000 (AG) use recent Arte interface
// 25 Jul 2000 (AG) added statistics
// 31 Jul 2000 (AG) added flag = flagdst again!!!
// 02 Aug 2000 (AG) rewrote 4LT to allow for class-wise triggering
// 03 Aug 2000 (AG) commissioned 4LT trigger setting
// 03 Aug 2000 (AG) polished descriptions
// 04 Aug 2000 (AG) polished printout
// 11 Aug 2000 (AG) introduced special cases
// 12 Mar 2001 (SE) set evhd pointer to the first row rather than last one
//
// *****************************************************************************

// standard includes
#include <iostream.h>

// Arte includes
#include "arte/arte.hh"
#include "arte/ArteUtil.h"
#include "farm/farm.hh"
#include "reco/eclass.hh"

// externals
extern int L4TKUM; // 4LT kumac
extern int L4TSET; // 4LT settings

extern int L4TMSK; // flag for special cases
extern int L4TSKP; // flag for special cases
extern int L4TMOD; // flag for special cases

extern int FADEB;  // debug flag

//
// program code
//
int
l4tEvnt()
{
  ARTE_mark( "l4tEvnt 0", 0 );

  //
  // EVHD table
  //
  ArtePointer<EVHD> evhd = *ArteTable<EVHD>::begin();
  evhd->flag = evhd->flagdst; // NEEDED!!!
  
  // debug
  if( FADEB ){
    cout << " 4LT: "
	 << "L4TKUM=0x" << hex << L4TKUM << dec << " "
	 << "L4TMOD=0x" << hex << L4TMOD << dec << " "
	 << "L4TSET=0x" << hex << L4TSET << dec << " "
	 << endl;
  }

  // preset
  arteutil_.eventout = 1; // write out event

  // counters
  static int nEVT = 0;
  static int nOUT = 0;

  //
  // check settings
  //

  //
  // 4LT enabled => use kumac
  //
  if( L4TSET ){

    // kumac NOT set => write out all events
    if( L4TKUM < 0 ){
      arteutil_.eventout = 1; // write out event
    }
    // kumac reject ALL => reject all events enabled
    else if( L4TKUM == 0 ){
      arteutil_.eventout = 0;
    }
    // kumac set => keep classes
    else if( L4TKUM  > 0 ){
      
      // preset
      arteutil_.eventout = 0;
      
      // debug
      if( FADEB ) cout << " 4LT: flagdst=0x" << hex << evhd->flagdst << dec << endl; 
      
      // evalute classification mask
      if( L4TMSK & evhd->flagdst ) arteutil_.eventout = 1;

      // evaluate mode 1 ("SKIP")
      if( L4TMOD == 1 ){
	if( FADEB ) cout << " 4LT: SKIP skip=" << L4TSKP << endl;
	if( L4TSKP ) arteutil_.eventout = 0; // skip event
      }
      
      // evaluate mode 2 ("SLTMMC")
      if( L4TMOD == 2 ){
	if( FADEB ) cout << " 4LT: SLTMMC flagdst=0x" << hex << evhd->flagdst << dec << endl;
	if( ( evhd->flagdst & ECLASS_SLTMM  ) &&
	    ( evhd->flagdst & ECLASS_SLTCUT )    ){
	  arteutil_.eventout = 1; // SLTMM && SLTCUT
	}
      }
    }
    
    // debug
    if( FADEB ) cout << " 4LT: eventout=" << arteutil_.eventout << endl;
    
    // statistics
    nEVT++;
    if( arteutil_.eventout ) nOUT++;
    if( !arteutil_.eventout || FADEB ) cout << " 4LT: events=" << nEVT << " written=" << nOUT << endl;
  }
  //
  // 4LT disabled => ignore kumac => keep all events
  //
  else {
    arteutil_.eventout = 1;
  }
  
  // debug
  if( FADEB ) cout << " 4LT: eventout=" << arteutil_.eventout << endl;
  
  ARTE_mark( "l4tEvnt 1", 0 );
  return arteutil_.eventout;
}
