/* This file defines 
     SetupBaseMenu<AnySetup>, SetupBaseMenu_1Par<AnySetup>
     SetupMenu<AnySetup>,     SetupMenu_1Par<AnySetup>
   classes.
   The difference between *Menu and *Menu_1Par is that the latter is the
   menu accepting the setup pointer as a parameter.
   This means that such a menu should be invoked like menu(&istr, &setup);
   The *Menu class defines just a menu without parameters, i.e. it can be
   invoked like menu(&istr); 
   See interactive.hh and interactive_1par.hh for more details.
   If there is only one setup, the programmer can normally use *Menu classes,
   such menu can work only with one setup, defined in creating the menu.
   If there are many setups, it can be convenient to use *Menu_1par, it can
   be used for many setups (the pointer to setup should be provided in every
   call to the menu). 

   The SetupBaseMenu and SetupBaseMenu_1Par both contain
     IstreamMenu_1Par<AnySetup::IChamber> chamber;
     IstreamMenu_1Par<AnySetup::ISection> section;
     IstreamMenu_1Par<AnySetup::ICell>    cell;

   and the SetupBaseMenu_1Par contain in addition
     IstreamMenu_1Par<AnySetup*>          setup;

   setup is the menu which contains commands for work with the whole setup;
   chamber is the menu for the chambers and so on. One can insert necessary
   interactive commands directly to the desired menu. Note that such
   menus all have one parameter (IChamber, for example). Your interactive
   command should be able to accept (pointer to istream) and (IChamber).

   Then, e.g., to process the top menu "setup" call: setup(is, AnySetup*);
   All next menu's are attached to previous in such a way that
   in e.g. "section" menu you can deal only with the sections from the 
   chambers and setups selected in the "setup" and "chamber" menu's.
   It can be seen from the following simple example (suppose we are using
   SetupBaseMenu, where there is no need to define setup: it is always fixed,
   since SetupBaseMenu works with only one setup)
   /chamber{i1- /z{468.75} /create{4}
                /section{0 /grid{grid 0.8x1.6, offset (0,0)}}
           }
   Here we should first define the chamber, which will be affected; its name
   is i1-. Then we are inside the "chamber" menu. Here we have the command 
   /z{...} defining the z position of the chamber, /create{4} which means 
   create 4 sections inside this chamber. Finally we invoke the "section" menu
   with the command /section{...}. As one can see we will be working with the 
   section #0. After defining the section number we are inside the "section" 
   menu, which has the command /grid{...} defining the grid, etc..

   SetupBaseMenu and SetupBaseMenu_1Par define no commands in their menus.
   The commands can be added later. Or one can derive other classes from
   them, where some interactive commands are defined just in the constructors.
   SetupMenu and SetupMenu_1Par are one example of such classes. They already 
   contain the interactive commands (like /z, /create above) for constructing
   the setups from scratch. 
   
   To insert your own new interactive command with one parameter in the menu
   you may write e.g. (again, you may look at interactive_1par.hh for details):
     section("command_name") = make_interactive_command_1par(
     <your_function_object>, (AnySetup::ISection*)0);
   Here <your_function_object> should be able to work with 
   istream* is and AnySetup::ISection i_section, i.e.
     expression <your_function_object>(is, i_section) should be valid.
*/   

// Author V.Balagura (balagura@mail.desy.de)   (28.02.99).

#ifndef _setup_menu_hh
#define _setup_menu_hh 1

#include "ptset/istream_menu_1par.hh"
#include "ptset/istream_menu.hh"
#include "ptset/call_via_istr.hh"
#include "ptset/string_primitives.hh"
#include "ptset/setup.hh"
#include "ptset/setup_figure_istream_menu.hh"

/* ----------------------- SetupMenuPrimitives ----------------------- */
template <class AnySetup>
class SetupMenuPrimitives {
 public:
  /* -------------------- ChamberMenu -------------------- */
  class ChamberMenu {
    IstreamMenu_1Par<typename AnySetup::IChamber>& menu;
  public:
    ChamberMenu(IstreamMenu_1Par<typename AnySetup::IChamber>& m) 
      : menu(m) {}
    void operator()(istream* is, AnySetup* p_setup) {
      string chamber_name;
      *is >> chamber_name;
      if ( !is ) {
	global_error()("can not read chamber name");
	return;
      }
      typename AnySetup::IChamber i_chamber = p_setup->find(chamber_name);
      if ( i_chamber == p_setup->end() ) {
	global_error()("there is no chamber \"" + chamber_name + "\"");
	return;
      }
      menu(is, i_chamber);
      if ( global_error() )
	global_error() << ("chamber \"" + chamber_name + "\"");
    }
  };
  /* -------------------- SectionMenu -------------------- */
  class SectionMenu {
    IstreamMenu_1Par<typename AnySetup::ISection>& menu;
  public:
    SectionMenu(IstreamMenu_1Par<typename AnySetup::ISection>& m) 
      : menu(m) {}
    void operator()(istream* is, typename AnySetup::IChamber i_chamber) {
      unsigned int n_section;
      *is >> n_section;
      if ( !is ) {
	global_error()("can not read section number");
	return;
      }
      if ( n_section >= (*i_chamber).second.size() ) {
	string tmp("there is no section #");
	tmp << n_section;
	global_error()(tmp);
	return;
      }
      typename AnySetup::ISection 
	i_section = (*i_chamber).second.begin() + n_section;
      menu(is, i_section);
      if ( global_error() ) {
	string tmp("section #");
	tmp << n_section;
	global_error() << tmp;
      }
    }
  };
  /* -------------------- CellMenu -------------------- */
  class CellMenu {
    IstreamMenu_1Par<typename AnySetup::ICell>& menu;
  public:
    CellMenu(IstreamMenu_1Par<typename AnySetup::ICell>& m) 
      : menu(m) {}
    void operator()(istream* is, typename AnySetup::ISection i_section) {
      Index ind;
      *is >> ind;
      if ( !is ) {
	global_error()("can not read cell index");
	return;
      }
      typename AnySetup::ICell i_cell = (*i_section).find(ind);
      if ( i_cell == (*i_section).end() ) {
	string tmp("there is no cell with index ");
	tmp << ind;
	global_error()(tmp);
	return;
      }
      menu(is, i_cell);
      if ( global_error() ) {
	string tmp("cell with index ");
	tmp << ind;
	global_error()(tmp);
	return;
      }
    }
  };
  /* -------------------- Functions -------------------- */
  /* 
     // Not used. CreateChambers below does the same job but can add several
     // chambers at once.
  static void create_chamber(const string& chamber_name, AnySetup* p_setup) {
    AnySetup& s = *p_setup;
    s.erase(chamber_name);
    s[chamber_name];
  }
  */
  struct CreateChambers {
    void operator()(istream* is, AnySetup* p_setup) {
      string name;
      while ( (*is) >> name ) {
	p_setup->erase(name);
	(*p_setup)[name];
      }
    }
  };
    
  static void create_sections(unsigned int n_sections,
			      typename AnySetup::IChamber i_chamber) {
    for (unsigned int i=0; i<n_sections; ++i)
      (*i_chamber).second.push_back(typename AnySetup::Section());
  }
  static void erase_section(unsigned int n_section,
			    typename AnySetup::IChamber i_chamber) {
    typename AnySetup::Chamber& ch = (*i_chamber).second;
    if ( n_section >= ch.size() ) {
      string tmp("there is no section #");
      tmp << n_section;
      global_error()(tmp);
      return;
    }
    ch.erase(ch.begin()+n_section);
  }
  static void set_chamber_z(double z,
			    typename AnySetup::IChamber i_chamber) {
    (*i_chamber).second.z = z;
  }
  static void set_section_grid(const Grid& grid, 
			       typename AnySetup::ISection i_section) {
    (*i_section).grid = grid;
  }
  static void set_section_figure(istream* is,
				 typename AnySetup::ISection i_section) {
    SetupFigureBase* fig;
    setup_figure_istream_menu()(is, &fig);
    if ( !global_error() ) (*i_section).ptr_figure = fig;
  }
  static void make_chamber_cells(const Rectangle& cover, 
				 typename AnySetup::IChamber i_chamber) {
    (*i_chamber).second.make_cells(cover);
  }

  // The following function was used before. Currently the
  // Make_Cells_Projectively_In class below is used instead of it
  // which applies "cover" not for the whole Setup, but for the given group of
  // chambers
  static void make_setup_cells_in_projective_rectangle
  (const Rectangle& cover_at_first_chamber, AnySetup* p_setup) {
    // find the minimal z (of the chamber with the lowest z)
    vector<vector<typename AnySetup::IChamber> > vv_ichamber;
    // z_precision = 0.1
    sort_setup_chambers_by_z(*p_setup, &vv_ichamber, 0.1);
    double z0 = (*vv_ichamber[0][0]).second.z;

    for (typename AnySetup::IChamber ich=p_setup->begin(); 
	 ich!=p_setup->end(); ++ich) {
      typename AnySetup::Chamber& ch = (*ich).second;
      ch.make_cells(cover_at_first_chamber * (ch.z/z0));
    }
  }

  struct Make_Cells_Projectively_In {
    void operator()(istream* is, AnySetup* p_setup) {
      Rectangle cover; string for__;
      if ( !(*is >> cover >> for__ ) || for__ != "for" ) {
	global_error()("can not read \"<rectangle> for\" block");
	return;
      }
      // find the minimal z (of the chamber with the lowest z)
      vector<vector<typename AnySetup::IChamber> > vv_ichamber;
      // z_precision = 0.1
      sort_setup_chambers_by_z(*p_setup, &vv_ichamber, 0.1);
      double z0 = (*vv_ichamber[0][0]).second.z;
      string name;
      while ( *is >> name ) {
	typename AnySetup::IChamber ich = p_setup->find(name);
	if ( ich == p_setup->end() ) {
	  string tmp; 
	  tmp<< "there is no chamber " << name << "(covering rectangle = "
	     << cover << " at z=" << z0 << ")";
	  global_error()(tmp);
	  return;
	}
	(*ich).second.make_cells(cover * ((*ich).second.z/z0));
      }
    }
  };
  static void dump_setup_structure(AnySetup* p_setup) {
    p_setup->dump_structure(cout);
  }
};

/* ------------------------- SetupBaseMenu ------------------------- */
template <class AnySetup>
class SetupBaseMenu : public IstreamMenu {
  AnySetup* p_setup;
 public:
  IstreamMenu_1Par<typename AnySetup::IChamber> chamber;
  IstreamMenu_1Par<typename AnySetup::ISection> section;
  IstreamMenu_1Par<typename AnySetup::ICell>    cell;

  SetupBaseMenu(AnySetup* p_any_setup) : p_setup(p_any_setup) {
    
    parser         = new SlashBlock_InteractiveParser;
    chamber.parser = new SlashBlock_InteractiveParser;
    section.parser = new SlashBlock_InteractiveParser;
    cell.parser    = new SlashBlock_InteractiveParser;

    (*this)("chamber") = make_interactive_command
      (bind2(SetupMenuPrimitives<AnySetup>::ChamberMenu(chamber),
	     (istream**)0, p_setup));
    chamber("section") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::SectionMenu(section),
       (typename AnySetup::IChamber*)0);
    section("cell") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::CellMenu(cell),
       (typename AnySetup::ISection*)0);
  }
  SetupBaseMenu(const SetupBaseMenu& m) { assign(m); }

  SetupBaseMenu& operator=(const SetupBaseMenu& m) {
    if ( &m == this ) return *this;
    assign(m);
    return *this;
  }
private:
  void assign(const SetupBaseMenu& menu) {
    p_setup = menu.p_setup;
    *((IstreamMenu*)this) = *((IstreamMenu*)&menu);
    chamber = menu.chamber;
    section = menu.section;
    cell    = menu.cell;
    // now make relations
    (*this)("chamber") = make_interactive_command
      (bind2(SetupMenuPrimitives<AnySetup>::ChamberMenu(chamber),
	     (istream**)0, p_setup));
    chamber("section") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::SectionMenu(section),
       (typename AnySetup::IChamber*)0);
    section("cell") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::CellMenu(cell),
       (typename AnySetup::ISection*)0);
  }
};
/* ------------------------- SetupMenu ------------------------- */
template <class AnySetup>
class SetupMenu : public SetupBaseMenu<AnySetup> {
 public:
  SetupMenu(AnySetup* p_any_setup) : SetupBaseMenu<AnySetup>(p_any_setup) {
    (*this)("create") = make_interactive_command
      (bind2(typename SetupMenuPrimitives<AnySetup>::CreateChambers(), 
	     (istream**)0, p_any_setup));
    chamber("create") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::create_sections,
	(unsigned int*)0, (typename AnySetup::IChamber*)0),
       (typename AnySetup::IChamber*)0);
    chamber("erase") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::erase_section,
	(unsigned int*)0, (typename AnySetup::IChamber*)0),
       (typename AnySetup::IChamber*)0);
    chamber("z") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::set_chamber_z,
	(double*)0, (typename AnySetup::IChamber*)0), 
       (typename AnySetup::IChamber*)0);
    section("grid") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::set_section_grid,
	(Grid*)0, (typename AnySetup::ISection*)0),
       (typename AnySetup::ISection*)0);
    section("figure") = make_interactive_command_1par
      (&SetupMenuPrimitives<AnySetup>::set_section_figure,
       (typename AnySetup::ISection*)0);
    chamber("make_cells") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::make_chamber_cells,
	(Rectangle*)0, (typename AnySetup::IChamber*)0),
       (typename AnySetup::IChamber*)0);
    // old version for old geometry definition: all chambers at once
    (*this)("make_cells_in_projective_rectangle") = 
     make_interactive_command
     (bind2(call_via_istr_check_EOF_1par
	    (&SetupMenuPrimitives<AnySetup>::
	     make_setup_cells_in_projective_rectangle,
	     (Rectangle*)0, (AnySetup**)0), (istream**)0, p_any_setup));
    (*this)("make_cells_projectively_in") =
      make_interactive_command
      (bind2
       (typename 
	SetupMenuPrimitives<AnySetup>::Make_Cells_Projectively_In(),
	(istream**)0, p_any_setup));
    (*this)("dump_structure") = make_interactive_command
      (bind2(call_via_istr_check_EOF_1par
	     (&SetupMenuPrimitives<AnySetup>::dump_setup_structure,
	      (AnySetup**)0), 
	     (istream**)0, p_any_setup));
  }
};
/* ------------------------- SetupBaseMenu_1Par ------------------------- */
template <class AnySetup>
class SetupBaseMenu_1Par {
 public:
  IstreamMenu_1Par<AnySetup*>                   setup;
  IstreamMenu_1Par<typename AnySetup::IChamber> chamber;
  IstreamMenu_1Par<typename AnySetup::ISection> section;
  IstreamMenu_1Par<typename AnySetup::ICell>    cell;

  SetupBaseMenu_1Par() {
    
    setup.parser   = new SlashBlock_InteractiveParser;
    chamber.parser = new SlashBlock_InteractiveParser;
    section.parser = new SlashBlock_InteractiveParser;
    cell.parser    = new SlashBlock_InteractiveParser;

    setup("chamber") = make_interactive_command_1par
     (SetupMenuPrimitives<AnySetup>::ChamberMenu(chamber), (AnySetup**)0);
    chamber("section") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::SectionMenu(section),
       (typename AnySetup::IChamber*)0);
    section("cell") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::CellMenu(cell),
       (typename AnySetup::ISection*)0);
  }

  SetupBaseMenu_1Par(const SetupBaseMenu_1Par& m) { assign(m); }

  SetupBaseMenu_1Par& operator=(const SetupBaseMenu_1Par& m) {
    if ( &m == this ) return *this;
    assign(m);
    return *this;
  }
private:
  void assign(const SetupBaseMenu_1Par& menu) {
    setup   = menu.setup;
    chamber = menu.chamber;
    section = menu.section;
    cell    = menu.cell;
    // now make relations
    setup("chamber") = make_interactive_command_1par
     (SetupMenuPrimitives<AnySetup>::ChamberMenu(chamber), (AnySetup**)0);
    chamber("section") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::SectionMenu(section),
       (typename AnySetup::IChamber*)0);
    section("cell") = make_interactive_command_1par
      (SetupMenuPrimitives<AnySetup>::CellMenu(cell),
       (typename AnySetup::ISection*)0);
  }
};
/* ------------------------- SetupBaseMenu_1Par ------------------------- */
template <class AnySetup>
class SetupMenu_1Par : public SetupBaseMenu_1Par<AnySetup> {
 public:
  SetupMenu_1Par() {
    /*
    setup("create") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::create_chamber,
	(string*)0, (AnySetup**)0), (AnySetup**)0);
	*/
    setup("create") = make_interactive_command_1par
      (typename SetupMenuPrimitives<AnySetup>::CreateChambers(), (AnySetup**)0);
    chamber("create") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::create_sections,
	(unsigned int*)0, (typename AnySetup::IChamber*)0),
       (typename AnySetup::IChamber*)0);
    chamber("erase") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::erase_section,
	(unsigned int*)0, (typename AnySetup::IChamber*)0),
       (typename AnySetup::IChamber*)0);
    chamber("z") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::set_chamber_z,
	(double*)0, (typename AnySetup::IChamber*)0), 
       (typename AnySetup::IChamber*)0);
    section("grid") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::set_section_grid,
	(Grid*)0, (typename AnySetup::ISection*)0),
       (typename AnySetup::ISection*)0);
    section("figure") = make_interactive_command_1par
      (&SetupMenuPrimitives<AnySetup>::set_section_figure,
       (typename AnySetup::ISection*)0);
    chamber("make_cells") = make_interactive_command_1par
      (call_via_istr_check_EOF_1par
       (&SetupMenuPrimitives<AnySetup>::make_chamber_cells,
	(Rectangle*)0, (typename AnySetup::IChamber*)0),
       (typename AnySetup::IChamber*)0);
    // old version for old geometry definition: all chambers at once
    setup("make_cells_in_projective_rectangle") = 
     make_interactive_command_1par
     (call_via_istr_check_EOF_1par
      (&
       SetupMenuPrimitives<AnySetup>::make_setup_cells_in_projective_rectangle,
	    (Rectangle*)0, (AnySetup**)0), (AnySetup**)0);
    setup("make_cells_projectively_in") =
      make_interactive_command_1par
      (typename 
       SetupMenuPrimitives<AnySetup>::Make_Cells_Projectively_In(),
       (AnySetup**)0);
    setup("dump_structure") = make_interactive_command_1par
    (call_via_istr_check_EOF_1par
     (&SetupMenuPrimitives<AnySetup>::dump_setup_structure,
      (AnySetup**)0), (AnySetup**)0);
  }
};

#endif
