// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 

// class rangerTrack
// basic ranger fit object

#ifndef RANGER_TRACK_HH
#define RANGER_TRACK_HH

#include <list.h>
#include <vector.h>
#include "CLHEP/Matrix/SymMatrix.h"
#include "CLHEP/Matrix/Vector.h"
#include "rang/FitMode.hh"
#include "rang/ParticleType.hh"
#include "rang/RMap.hh"

class rangerTrack {

public:

  rangerTrack();
  rangerTrack(ArtePointer<RSEG> rseg, rangerFitMode fm,
	      rangerParticleType pType= MUON,
	      rangerRMap* rmap=0);
  rangerTrack(const vector<ArtePointer<RSEG> >& rsegVec, 
	      rangerFitMode fm,
	      rangerParticleType pType= MUON,
	      vector<rangerRMap>* rmapVec=0);
  rangerTrack(const vector<ArtePointer<RSEG> >& rsegVec, 
	      ArtePointer<RCCL> rccl,int charge,rangerFitMode fm,
	      rangerParticleType pType= MUON,
	      vector<rangerRMap>* rmapVec=0);
  rangerTrack(ArtePointer<RTRA> rtra, rangerFitMode fm,
	      rangerParticleType pType= MUON,
	      rangerRMap* rmap=0);
  ~rangerTrack();

  typedef list<rangerNode*>::const_iterator rangerNodeConstIterator;
  typedef list<rangerNode*>::iterator rangerNodeIterator;
  typedef list<rangerNode*>::reverse_iterator rangerNodeReverseIterator; 
  typedef list<rangerNode*>::const_reverse_iterator 
                                         rangerNodeConstReverseIterator; 
  inline rangerNodeIterator  begin() {return nodes.begin();}
  inline rangerNodeIterator    end() {return nodes.end();}
  inline rangerNodeConstIterator  begin() const {return nodes.begin();}
  inline rangerNodeConstIterator    end() const {return nodes.end();}
  inline rangerNodeIterator   last() {
    rangerNodeIterator tmp = end(); tmp--; return tmp;}
  inline rangerNodeIterator next( rangerNodeIterator n ) {
    rangerNodeIterator tmp = n; tmp++; return tmp;}
  inline rangerNodeIterator previous( rangerNodeIterator n ) {
    rangerNodeIterator tmp = n; tmp--; return tmp;}
  inline bool empty() const {return nodes.empty();}
  inline rangerNodeReverseIterator  rbegin() {return nodes.rbegin();}
  inline rangerNodeReverseIterator    rend() {return nodes.rend();}
  inline rangerNodeConstReverseIterator  rbegin() const 
    {return nodes.rbegin();}
  inline rangerNodeConstReverseIterator    rend() const 
    {return nodes.rend();}
  rangerParticleType getParticleType() const { return tPartType ; }

  void  append(rangerNode& n);  // append new node at end of list
  void prepend(rangerNode& n);  // prepend new node at begin of list
  rangerNodeIterator insertAfter(rangerNodeIterator baseNode,
			  rangerNode* newNode); // insert new node
  rangerNodeIterator insertBefore(rangerNodeIterator baseNode,
			  rangerNode* newNode); // insert new node
  void covInit();
  int fit(const int firstPass=0,const int thirdPass=1,
	  const int insertMaterial=1,const int updateInternal=0); // (re)fit track
  int fitV();  // (re)fit track in VDS 
  int fitVf();  // (re)fit track in VDS fast 
  int fitVs();  // (re)fit track in VDS special
  int fitFast();  // (re)fit track fast
  int filter();   // perform upstream Kalman filter on track candidate
  int smooth();   // perform backward Kalman smoother .  .    .
  int filterDownStr();   // perform downstream Kalman filter on track candidate
  int insertWalls();  // insert multiple scattering objects
  int insertWallsV();  // insert multiple scattering objects (special for VDS)
  void updInternal(); // update internal degrees of freedom
  void trackList();
  void dump();
  void storeback(rangerRMap* rmap);  // store track back using rmap info
  void storeback(vector<rangerRMap>* rmapVec);  // same for multiple rmaps
  ArtePointer<RTRA> storeRtra() const;  // store into new RTRA entry
  ArtePointer<RSEG> storeRseg() const;  // store into new RSEG entry
  double getLowZ() const;
  const HepVector* getLowZVec() const;
  const HepSymMatrix* getLowZCov() const;
  float totalChiSq() const ;
  int   totalNdof() const ; // return number of degrees of freedom
  float chiSqProb() const ; // return chiSquare probability 
  double totalRadThickness() const; // calculate total rad thickness
  int genCmp() const; // generate cmp bit map for RSEG
  void lastNoise(); // apply process noise to last node
  inline rangerFitMode fitMode() { return tFitMode; };

  list<rangerNode*>& getNodes() { return nodes; }
  void zOrder();      // bring nodes into standard z ordering
  void orderIfNeeded();  // order nodes if z(begin) lt z(end)
  int fitGlobal() ;
private:
  list<rangerNode*> nodes;          // list of node pointers
  float qpEstimate;                 // current estimate for Q/p
  int electron;                     // true if electron
  int nIteration;                   // number of fit iterations
  int smoothed;                     // true if smoothed
  rangerFitMode tFitMode;           // variant of fit parameters
  rangerParticleType tPartType;     // particle species
  int status;                       // fit status

  friend ostream& operator<<(ostream &os,const rangerTrack &t);
};

#endif
