//***********************************************************************
//
//  Routine:      rcCreateRtraBasePatt
//                
//  Author:       W. Hulsbergen from rccreateRTRA (20/01/01)
//
//  Description:  
//  Routines for rtra creation in `rtra2001' scheme.
// 
//
//  Last modifications:
//  P.Conde (19/01/01) added last step of RTRA creation: rcCreateRtraLast.
//
//***********************************************************************

#include <map>
#include <set>
#include <arte/arte.hh>
#include <arte/RSEG_cmp.hh>
#include <reco/reco.hh>
#include <reco/rcflags.hh>
#include <marple/marple.hh>

ArtePointer<RTRA> rcCreateRtraBaseRseg(ArtePointer<RSEG> rseg) ;

//********************************************************************
//
//  Routine:      rcCreateRtraBasePatt
//
//  Description: 
// Create rtra tracks for all Tracker-VDS matches and for remaining
// tracker segments. Relations to hitb are set and hits are
// counted. Momentum is taken from the marple segment if that exists
// and from the tracker segment otherwise.
//
// Called from: `new reco chain'
// Calls to: rcSetRtraHitb and rcMakeRtraBaseRseg
//
//********************************************************************

int rcCreateRtraBasePatt() {
  // Loop over all pattern segments. Create seperate rtra for each
  // match to vds segment. Create only one rtra if no vds segments are
  // matched.

  for(ArteTable<RSEG>::iterator iseg = ArteTable<RSEG>::begin();
      iseg != ArteTable<RSEG>::end() ; ++iseg) 
    if( !((*iseg)->cmp & Rsegc::bitsign) &&
	(( *iseg)->fit == 0)  &&
	((*iseg)->cmp & Rsegc::patt) ) {
      ArtePointer<RSEG> patt = *iseg;
      
      // find tc segment
      ArtePointer<RSEG> trit ;
      for( ArteRelationIterator<RSEG> jseg = patt->rend.begin() ;
	   jseg != patt->rend.end() ; ++jseg ) 
	if( ((*jseg)->cmp & Rsegc::trit) != 0 )  trit = *jseg;
      
      // find all related vds segments
      // (make set because relations can be set twice ?)
      set<ArtePointer<RSEG> > vdssegs ;
      //vector<ArtePointer<RSEG> > vdssegs ;
      for( ArteRelationIterator<RSEG> ivds = patt->rbeg.begin();
	   ivds != patt->rbeg.end(); ++ivds )
	if( !((*ivds)->cmp & Rsegc::bitsign) &&
	    (( *ivds)->fit == 0)  &&
	    ((*ivds)->cmp & Rsegc::vxd) ) vdssegs.insert(*ivds) ;
      
      // Create rtras. Note: there is always at least one rtra created,
      // even if there are no vds matches.
      set<ArtePointer<RSEG> >::iterator ivds = vdssegs.begin() ;
      do {
	ArtePointer<RTRA> rtra = rcCreateRtraBaseRseg(patt) ;
	rtra->cmain = patt.index() ;
	rtra->cmp |= patt->cmp ; 
	if(trit.index()) rtra->cmp |= trit->cmp ;
	if(patt->cmp&Rsegc::magt) rtra->mommask = RCMOMOTRMAGNET;
	else                      rtra->mommask = RCMOMOTRPCKICK;
	
	if(ivds != vdssegs.end() ) {
	  ArtePointer<RSEG> vds = *ivds ;
	  createRelation( rtra, vds, &RTRA::rseg );
	  rtra->cmp  |= vds->cmp ;
	  rtra->csvd = vds.index() ;
	  // copy the vds pars
	  rtra->xf  = vds->xf;
	  rtra->yf  = vds->yf;
	  rtra->zf  = vds->zf;
	  rtra->txf = vds->txf;
	  rtra->tyf = vds->tyf;
	  memcpy(rtra->cvf,vds->cvf,10*sizeof(float)) ;	

	  // find the marple segment that connects OTR-VDS
	  ArtePointer<RSEG> marpleseg ;
	  for(ArteRelationIterator<RSEG> mseg = vds->rbeg.begin() ;
	      mseg != vds->rbeg.end() && !(marpleseg.index()); mseg++)
	    if( (*mseg)->fit==7 && ((*mseg)->cmp& Rsegc::vxd) &&
		((*mseg)->cmp&Rsegc::patt) )
	      for(ArteRelationIterator<RSEG> kseg = patt->rbeg.begin() ;
		  kseg != patt->rbeg.end() && !(marpleseg.index()); kseg++)
		if( (*mseg) == (*kseg) ) marpleseg = *mseg ;
	  
	  // if marple segment exists, take momentum from there. (else
	  // take momentum from patt segment)
	  if(marpleseg.index()) {
	    rtra->pf      = marpleseg->pf ;
	    rtra->cvf[14] = marpleseg->cvf[14] ;
	    // I am not sure about this line
	    rtra->mommask = RCMOMOTRTARGET;
	  } 
	}
	
	// relations to the rsegs must be set from vds to tc:
	createRelation( rtra, patt, &RTRA::rseg );
	if(trit.index()) createRelation( rtra, trit, &RTRA::rseg ) ;
	
	// relations to hitb
	rcSetRtraHitb( rtra ) ;
	rcCountRtraHits( rtra ) ;

      } while (ivds != vdssegs.end() && (++ivds) != vdssegs.end()) ; 
    }
  
  return 0;
}

//********************************************************************
//
//  Routine:      rcCreateRtraLast
//
//  Description: 
//  Last step of RTRA creation:
//            Add VDS-ECAL matches and SVD-not-matched as new RTRA's.
//            Add RICH to the existing RTRA's, from RICH-OTR matches.
//
// Called from: rcevnt
// Calls to: rcSetRtraHitRelation and rcCountRtraHits
// 
// Modified:
//  - Added check on vds-ecal clone               Wouter Hulsbergen
//  - Revised scheme, including following fixes:          22/03/2001
//    - also add vds that are only used in invalid rtras
//    - first add all vds-ecal, then check if vds used
//    - add rich only to existing rtra
//  - Removed RTRA-RICH match: now in seperate marple routine
//
//********************************************************************

int rcCountRtraHits();

int  rcCreateRtraLast()
{
  //---------------------------------------------------------
  // Add VDS-ECAL that have not been used
  //---------------------------------------------------------

  for(ArteTable<RSEG>::iterator matchseg = ArteTable<RSEG>::begin();
      matchseg != ArteTable<RSEG>::end() ; ++matchseg)
    if( !((*matchseg)->cmp & Rsegc::bitsign) &&
	((*matchseg)->fit == 7) &&
	((*matchseg)->cmp & Rsegc::vxd) && ((*matchseg)->cmp & Rsegc::ecal) ) {

      ArtePointer<RCCL> rccl = *(*matchseg)->rccl.begin();
      ArtePointer<RSEG> vds ;
      for( ArteRelationIterator<RSEG> iter = (*matchseg)->rend.begin();
	   iter != (*matchseg)->rend.end(); ++iter )
	if( ((*iter)->cmp & Rsegc::vxd) )  vds = *iter;
      
      if( vds.index()==0 || rccl.index()==0) 
	cout << "REAL PROBLEM IN rcCreateRtraLast()" << endl ;

      // check that there is no rtra with this combination yet
      bool found = false ;
      for(ArteRelationIterator<RTRA> ivrtra = vds->rtra.begin() ;
	  ivrtra != vds->rtra.end() && !found; ++ivrtra)
	if( !((*ivrtra)->cmp & Rsegc::bitsign))
	  for(ArteRelationIterator<RTRA> iertra = rccl->rtra.begin() ;
	      iertra != rccl->rtra.end() && !found; ++iertra)
	    if( (*ivrtra) == (*iertra) ) found=true ;
      
      if(!found) {
	ArtePointer<RTRA> rtra = rcCreateRtraBaseRseg(vds) ;
	rtra->csvd = vds.index() ;
	rtra->cmp |= Rsegc::ecal ;
	rtra->cecal = rccl.index() ;
	rtra->hca = 1;

	createRelation( rtra, rccl, &RTRA::rccl );
	createRelation( rtra, vds, &RTRA::rseg );

	// copy from match
	rtra->pf = (*matchseg)->pf;
	rtra->cvf[10] = rtra->cvf[11] = rtra->cvf[12] = rtra->cvf[13] = 0;
	rtra->cvf[14] = (*matchseg)->cvf[14]; // cov(5,5)
	rtra->mommask = RCMOMVDSECAL;
	
	// set relation to vds hits and count
	rcSetRtraHitb(rtra) ;
	rcCountRtraHits( rtra ) ;
      }
    }


  //---------------------------------------------------------
  // Add vds segments that have not been used yet
  //---------------------------------------------------------

  for(ArteTable<RSEG>::iterator iseg = ArteTable<RSEG>::begin();
      iseg != ArteTable<RSEG>::end() ; ++iseg)
    if( !((*iseg)->cmp & Rsegc::bitsign) &&
	((*iseg)->fit == 0) &&
	((*iseg)->cmp & Rsegc::vxd) ) {
      
      ArtePointer<RSEG> vds = (*iseg);
      
      // See if there is a _valid_ rtra in which rseg is used.
      bool found = false ;
      for(ArteRelationIterator<RTRA> irtra = vds->rtra.begin() ;
	  irtra != vds->rtra.end() && !found; ++irtra)
	if( !((*irtra)->cmp & Rsegc::bitsign) ) found=true ;
      
      if(!found) {
	ArtePointer<RTRA> rtra = rcCreateRtraBaseRseg(vds) ;
	rtra->csvd = vds.index() ;
	rtra->mommask = RCMOMVDSONLY ;
	
	createRelation( rtra, vds, &RTRA::rseg );
	
	// set relation to vds hits and count
	rcSetRtraHitb(rtra) ;
	rcCountRtraHits( rtra ) ;
      }
    }

  return 0;
}


//********************************************************************
//
//  Routine:      rcCreateRtraBaseRseg
//
//  Description: 
//  Create rtra entry from an existing rseg. This where the default
//  values are set.
//
//  Called from: rcCreateRtraBasePatt
//
//********************************************************************

ArtePointer<RTRA> rcCreateRtraBaseRseg(ArtePointer<RSEG> seg)
{
  // new rtra object
  RTRA rtra;
  
  // reset rtra entries that will not be copied from rseg
  rtra.hsi = rtra.hin = rtra.hou = rtra.hri = rtra.hca = rtra.htr = rtra.hmu = 0;
  rtra.fit = 0 ;
  rtra.lrme = rtra.lre = rtra.lrmu = rtra.lrpi = rtra.lrk = rtra.lrp = rtra.flr = 0;
  rtra.lte = rtra.lth = rtra.flt = 0 ;
  rtra.lee = rtra.leh = rtra.lemi= rtra.fle = 0;
  rtra.lmmu = rtra.lmh = rtra.flm = 0;
  rtra.le = rtra.lmu = rtra.lpi = rtra.lk = rtra.lp = rtra.fid = 0 ;
  rtra.csvd = rtra.cmain = rtra.cecal = rtra.crich = rtra.cmuon = -10;
  rtra.txr = rtra.tyr = rtra.cer = 0 ;
  rtra.mommask = -10 ;
  rtra.bse = rtra.bsi = rtra.bmain = 0;
  rtra.ndof = -1;
  
  // copy track parameters and cmp flag from rseg
  rtra.xf  = seg->xf;
  rtra.yf  = seg->yf;
  rtra.zf  = seg->zf;
  rtra.txf = seg->txf;
  rtra.tyf = seg->tyf;
  rtra.pf  = seg->pf;
  memcpy(rtra.cvf,seg->cvf,15*sizeof(float)) ;
  // if no magnet info, set part of cvf to 0
  if(seg->zf>680 || seg->ze<220){
    rtra.cvf[10]=rtra.cvf[11]=rtra.cvf[12]=rtra.cvf[13]=0 ;
    rtra.cvf[14]=1;
  }
  rtra.chi = seg->chi ;
  rtra.cmp = seg->cmp;
  
  return ArteTable<RTRA>::push_back( rtra );
}

