// **********************************************************************************************************************************
//
//    K0, Lambda and Lambda_bar Selection routine. 
//
//     Author F.Sanchez (Nov'00)
//
//      Include "rcSelect_K0S_Lambda.hh"
//
//      ARTE ONLINE VERSION:  call "rcSelect_K0S_Lambda()" after reconstruction. It initialize automatically and 
//                            is set to true if one K0S or lambda is found. Optimal cuts are set, and no ntuple 
//                            is created. The RTRA associated to V0 has a toop :
//
//                                      kinema::findTopp(310) for K short.
//                                      kinema::findTopp(3122) for Lambda.
//                                      kinema::findTopp(-3122) for anti Lambda. 
//
//
//       To initialize: K0LambdaSelectInit(void);                                -------> No cuts, does not produce an ntuple.
//                      K0LambdaSelectInit(char ntuplefile )                     -------> No cuts. It does produe an ntuple.
//                      K0LambdaSelectInit(cu1,cu2,cu3)                          -------> Cuts, does not produce an ntuple.
//                      K0LambdaSelectInit(cu1,cu2,cu3,char ntuplefile )         -------> Cuts, produce an ntuple.
//                         cu1  = minimum distance between tracks. (cm)
//                         cu2  = flight path of K0S in K0S restframe. (cm)
//                         cu3  = Impact parameter to the closest primary (cm)
//
//                      K0LambdaSelectMassWindow(kshortlow,kshorthigh,lambdalow,lambdahigh) -> define mass windows
//                     
//
//       Each event:    K0LambdaSelect() returns the number of K0S in the event. It creates a new VERTEX with 
//                                      k0s particle as an rtra with fid == 310.
//
//       Stop funct:    K0LambdaClose()  close ntuples if opened.
//
//       For each K0 or lambda found a new vertex is created. The new vertex can be identified as follows:
//
//   if( K0LambdaSelect() ) 
//     {
//       for( ArteTable<RVER>::iterator rverIt = ArteTable<RVER>::begin();
//  		       rverIt != ArteTable<RVER>::end() ; ++rverIt) 
//         {
//  	     ArtePointer<RTRA> part = rverIt->btra;
//  	     switch( (part->topp)->lund ) 
//  	       {
//               case 310:
//                  .................Kshort.............
//                  break;
//               case 3122:
//  	            ................LAMBDA..............
//                  break;
//               case -3122:
//  	            ................LAMBDA_BAR..............
//                  break;
//               default:
//                  .............OTHER VERTECES............
//                  break;
//  	       }
//         }
//     }
//
// Changes:
//
//   04 Dec 2000 (AG) corrected include path (reco/).
//   12 Dec 2000 (FS) more efficient lambda <-> k0s selection in the kinematically shared region.
//   14 Dec 2000 (FS) RHP histograms were not initialized.
//   09 Feb 2000 (FS) Block second initialization.
//   12 Feb 2000 (FS) Define spacename for internal functions.
//
// **********************************************************************************************************************************
//
// Includes
//
#include <stdio.h>
#include <math.h>
#include <vector.h>
#include <map.h>

#include "CLHEP/Vector/LorentzVector.h"
#include "CLHEP/Vector/LorentzRotation.h"
#include "marple/marple.hh"
#include "arte/ArteUtil.h"
#include "reco/rcFitFlag.hh"

#include "rhp.h"
#include "dq.h"

extern "C" {
  void hgevnt_();
  void dgevnt_();
  void l1evnt_();
  void rcevnt_();
}

extern "C" {
#include "hbook.h"
void hcdir_(char *,char *,int, int);
}

#include "reco/rcXkinema.h"

#include "arte/arte.hh"
#include "arte/ArteFlag.h"
#include "arte/ArteFlag.h"
#include "farm/farm.hh"
#include "farm/fa.hh"

#include "rcSelect_K0S_Lambda.hh"


ntuple evt;

struct {
    int iquest[100];
  } quest_;

static int FillNtuple;
static bool Online = FALSE;
static float DistCut_saved;
static float FlightCut_saved;
static float ImpactParamCut_saved;
static float K0LowMassCut_saved;
static float K0HighMassCut_saved;
static float LambdaLowMassCut_saved;
static float LambdaHighMassCut_saved;

float Pion   = 0.13957;
float Proton = 0.93827;
  
void K0SLAMBDA::rc_Select_K0S_Lambda_OnlineMode()
{
  Online = TRUE;
}

bool K0SLAMBDA::rcSelect_K0S_Lambda()
{
  ARTE_mark( "rcSelect_K0S_Lambda 0", 0 );

  static bool first = FALSE;

  if ( !first )
    {  
      K0SLAMBDA::rc_Select_K0S_Lambda_OnlineMode();
      K0SLAMBDA::K0LambdaSelectInit(0.04,0.13,0.06); 
      first = TRUE;
    }
    
  if( K0SLAMBDA::K0LambdaSelect() ) return TRUE;


  ARTE_mark( "rcSelect_K0S_Lambda 1", 0 );
  return FALSE;
}


void K0SLAMBDA::K0LambdaSelectMassWindow( float k0low, float k0high, float Lmbdlow, float Lmbdhigh )
{
  K0LowMassCut_saved = k0low;
  K0HighMassCut_saved = k0high;
  LambdaLowMassCut_saved = Lmbdlow;
  LambdaHighMassCut_saved = Lmbdhigh;
}


void K0SLAMBDA::K0LambdaSelectClose(void)
{
  int i = 0;

  if( !FillNtuple ) return;

  cout << " Calling HCDIR \n";

  hcdir_("//K0Lambda"," ",strlen("//K0Lambda"),strlen(" "));

  cout << " Calling HROUT \n";

  hrout(0,i,"T");

  cout << " Calling HREND \n";

  hrend("K0S");

}

void K0SLAMBDA::K0LambdaSelectInit(void)
{
  K0SLAMBDA::K0LambdaSelectInit(0.1,-10000.,10000.,0," ");
}


void K0SLAMBDA::K0LambdaSelectInit(float DistCut, float FlightCut, float ImpactParamCut )
{

 K0SLAMBDA::K0LambdaSelectInit(DistCut, FlightCut, ImpactParamCut,0," ");

}

void K0SLAMBDA::K0LambdaSelectInit(float DistCut, float FlightCut, float ImpactParamCut,char *ntuplefile )
{

 K0SLAMBDA::K0LambdaSelectInit(DistCut, FlightCut, ImpactParamCut,1,ntuplefile);

}


void K0SLAMBDA::K0LambdaSelectInit(char *ntuplefile)
{
  K0SLAMBDA::K0LambdaSelectInit(0.1,-10000.,10000.,1,ntuplefile);
}


void K0SLAMBDA::K0LambdaSelectInit(float DistCut, float FlightCut, float ImpactParamCut , int EnableNtuple, char *ntuplefile )
{
  static bool first = FALSE;
  FillNtuple = EnableNtuple;
  DistCut_saved = DistCut;
  FlightCut_saved = FlightCut;
  ImpactParamCut_saved = ImpactParamCut;
  K0LowMassCut_saved = 0.428;
  K0HighMassCut_saved = 0.568;
  LambdaLowMassCut_saved = 1.10;
  LambdaHighMassCut_saved = 1.13;

  if ( first ) {
    cout << " rcSelect_K0S_lambda already initialized " << endl;
    return;}

  first = TRUE;

  cout<< endl << endl ;
  cout<<"---------------------------------------------------------------- \n"; 
  cout<<" Initialization of K short selection \n";
  cout<<"   Cuts : \n";
  cout<<"     Maximum distance between tracks " << DistCut << " cm \n";
  if( FlightCut != -10000. ) 
  cout<<"     Minimum distance of K0 flight path (K0S rest frame) " << FlightCut_saved << " cm \n";
  if(  ImpactParamCut != 10000.  ) 
  cout<<"     Impact parameter to closest primary vertex " << ImpactParamCut_saved << " cm \n";
  cout<<"     Mass(K0S)= (" << K0LowMassCut_saved << " , " << K0HighMassCut_saved << " )  GeV \n";
  cout<<"     Mass(LAMBDA)= (" << LambdaLowMassCut_saved << " , " << LambdaHighMassCut_saved << " )  GeV \n";
  cout<<"   Options: \n";
  if( FillNtuple ) 
  cout<<"     Ntuple output " << ntuplefile <<" is selected. \n"; 
  else
  cout<<"     Ntuple output disable \n " ;
  cout<<"---------------------------------------------------------------- \n\n\n";
  

  if( FillNtuple ) 
    {
      int istat = 0;      
  
      cout << " Open " << ntuplefile << " for ntuples \n";
      
      quest_.iquest[9] = 65000;

      marpleinit_();
      
      hropen(1,"K0Lambda",ntuplefile,"N",8191,istat);
      
      if( istat ) cout << " Error opening file \n";
      
      cout << " Book " << Nt  << " ntuple \n";

      hbook1( 8000 ,"Selected events (1=k0s,2=lamb,3=lamb_b)",  3,  0.5,  3.5 ); 
      hbook1( 8001 ,"K short    mass", 50,  0.45,  0.545 ); 
      hbook1( 8002 ,"Lambda     mass", 25,   1.1,   1.13 ); 
      hbook1( 8003 ,"Lambda bar mass", 25,   1.1,   1.13 ); 

      if( FillNtuple == 2 ) 
	hbookn(3,"Selection","//K0Lambda",Nt);
      else
	hbookn(3,"Selection","//K0Lambda",Nt_mini);
      
      return;
    }
  else if ( Online )  
    {

      cout<<"     Online version enabled \n " ;

      rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 8000 ),"Selected events (1=k0s,2=lamb,3=lamb_b)",  3,  0.5,  3.5 ); 
      rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 8001 ),"K short    mass", 50,  0.45,  0.545 ); 
      rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 8002 ),"Lambda     mass", 25,   1.1,   1.13 ); 
      rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 8003 ),"Lambda bar mass", 25,   1.1,   1.13 ); 

    }


  return;  
}


int K0SLAMBDA::K0LambdaSelect()
{
  int n_selected = 0;
 
  ArtePointer<TOPP> tpk0s = kinema::findTopp(310);
  ArtePointer<TOPP> tplmb = kinema::findTopp(3122);
  ArtePointer<TOPP> tplmb_b = kinema::findTopp(-3122);

  // get VDS-OTR tracks - quality check

  vector< ArtePointer<RTRA> > rtra;
  int pos = 0;
  int neg = 0;

  int ntall = K0SLAMBDA::trSelect(rtra, pos, neg);
 
  // At least two track with opposite sign 

  if ( ntall <= 1 || pos == 0 || neg == 0 ) return 0;
 
  // fill kinematic quantities for selected tracks into STL vectors

  int first = 1;

  vector< float > tx;
  vector< float > ty;
  vector< float > px;
  vector< float > py;
  vector< float > pz;
  vector< float > p;
  vector< float > Q;

  vector< ArtePointer<RTRA> > track;
  vector< ArteTable<RVER>::iterator > primary;

  int   ntr = 0;
 
  // Some definitions for kinematic
  
  for( int i = 0; i != ntall ; i++ ) {

    // px/pz
    float Tx    = rtra[i]->txf;
    // py/pz
    float Ty    = rtra[i]->tyf;
    // Q/p  
    float P     = fabs(1./rtra[i]->pf);
    // Pz
    float Pz    = P/sqrt(1.+Tx*Tx+Ty*Ty);
    // Px
    float Px    = Pz*Tx;
    // Py
    float Py    = Pz*Ty;

    float Charge = 1.;
    if( rtra[i]->pf < 0 ) Charge = -1.;

    ntr++;
    track.push_back( rtra[i] );
    tx.push_back(Tx);
    ty.push_back(Ty);
    px.push_back(Px);
    py.push_back(Py);
    pz.push_back(Pz);
    p.push_back( P );
    Q.push_back(Charge);
  }

  if ( ntr < 2 ) return 0;

  //
  // 2-track vertexing
  // -----------------
  
  float dz;
  float xs, ys, zs;
   
  int nprimary = 0;
  
  evt.ntr = (float)ntr;

  for(int i=1; i!=ntr; ++i) 
    {
      for(int j=0; j!=i; ++j) 
	{
	  GroverVertexList vtxsel;
	  float mass1,mass2;
          // Look for opposite charge tracks.

	  if( rtra[i]->csvd == rtra[j]->csvd ) continue;
 	  if( rtra[i]->cmain == rtra[j]->cmain ) continue;


	  if( Q[i] == Q[j] ) continue;

	  // Look for candidate 2-track secondary vertices
	  GroverParticle pi1(track[i]), pi2(track[j]);
	  pi1.GlobalToLocal(); 
	  pi2.GlobalToLocal();
	  
	  // If the momemtum cocient is smaller than 4.5 we start to check the k0 short.
	  if( max(p[i]/p[j],p[j]/p[i]) < 4.0 ) 
	    {
	      evt.part = 1;
	      mass1 = Pion;
	      mass2 = Pion;
	    }
	  else
	    {
	      //  If not, we check first the Lambda or the Lambda_bar.
	      if( p[i] > p[j] )
		{
		  mass1 = Proton;
		  mass2 = Pion;
		  if( Q[i] > 0 ) 
		    evt.part = 2;
		  else
		    evt.part = 3;
		}
	      else
		{
		  mass1 = Pion;
		  mass2 = Proton;
		  if( Q[i] < 0 ) 
		    evt.part = 2;
		  else
		    evt.part = 3;
		}
	    }
	  
	  pi1.SetMass(mass1);
	  pi2.SetMass(mass2);
	  
	  GroverVertex vtx(&pi1,&pi2);
	  //and make an unconstrained fit
	  vtx.Fit(-1.0);
	  
	  xs = vtx.X();
	  ys = vtx.Y();
	  zs = vtx.Z();
	  dz = zs+5.;
	  
	  if( dz < 2.0 ) continue;
	  
	  evt.dist = pi1.DistanceTo(&pi2);  
	  
	  if( evt.dist > DistCut_saved ) continue;
	  
	  evt.mass = vtx.Mass();
	  
	  evt.chi2 = vtx.Chi2();
	  
	  evt.prob =  vtx.Prob();
	  
	  if( !MassCut((int)evt.part,evt.mass ) ) 
	    {
	      K0SLAMBDA::SelectNextTry(&evt.part, (int) Q[i],&mass1,&mass2);
	      pi1.SetMass(mass1);
	      pi2.SetMass(mass2);
	      GroverVertex vtx(&pi1,&pi2);
	      vtx.Fit(-1.0);
	      evt.mass = vtx.Mass();
	      evt.chi2 = vtx.Chi2();
	      if( !MassCut((int)evt.part,evt.mass ) ) 
		{
		  K0SLAMBDA::SelectNextTry(&evt.part,(int) Q[i],&mass1,&mass2);
		  pi1.SetMass(mass1);
		  pi2.SetMass(mass2);
		  GroverVertex vtx(&pi1,&pi2);
		  vtx.Fit(-1.0);
		  mass1 = Pion;
		  mass2 = Proton;
		  evt.mass = vtx.Mass();
		  evt.chi2 = vtx.Chi2();
		  if( !MassCut((int)evt.part,evt.mass ) ) 
		    continue;
		  else
		    vtxsel.push_back(vtx);
		}
	      else
		vtxsel.push_back(vtx);
	    }
	  else 
	    vtxsel.push_back(vtx);
	  
	  
	  float mom[3],pos[3],vertex[3];
	  int primvtx = 0;
	  
	  evt.q1 = Q[i];
	  evt.q2 = Q[j];
	  
	  // Impact parameter to vertex.
	  
	  mom[0] = px[i]+px[j];
	  mom[1] = py[i]+py[j];
	  mom[2] = pz[i]+pz[j];
	  
	  pos[0] = xs;
	  pos[1] = ys;
	  pos[2] = zs;
	  
	  if( first )
	    {
	      for( ArteTable<RVER>::iterator rverIt = ArteTable<RVER>::begin();
		   rverIt != ArteTable<RVER>::end() ; ++rverIt) 
		{
		  if( rverIt->nt >= 4 && rverIt->type == 1 )
		    {
		      nprimary++;
		      primary.push_back( rverIt );
		    }
		}
	      
	      evt.nprim =(float)nprimary;
	      first = 0;
	    }
	  
	  float mindist = 10.;
	  
	  for(int  ll=0; ll<nprimary ; ll++) 
	    {
	      float local;
	      
	      vertex[0] =  primary[ll]->xv;
	      vertex[1] =  primary[ll]->yv;
	      vertex[2] =  primary[ll]->zv;
	      
	      local = K0SLAMBDA::distance(mom,pos,vertex);
	      
	      if( local < mindist && local > 0.0 )
		{
		  mindist = local;
		  primvtx = ll;
		  evt.primx = primary[ll]->xv;
		  evt.primy = primary[ll]->yv;
		  evt.primz = primary[ll]->zv;
		  evt.primtr = primary[ll]->nt;
		}
	    }
	  
	  if ( mindist > ImpactParamCut_saved ) continue;
	  
	  // Boost to K0Lambda rest frame.

	  HepLorentzVector plab1(px[i],py[i],pz[i],sqrt(px[i]*px[i]+py[i]*py[i]+pz[i]*pz[i]+mass1*mass1));
	  HepLorentzVector plab2(px[j],py[j],pz[j],sqrt(px[j]*px[j]+py[j]*py[j]+pz[j]*pz[j]+mass2*mass2));

	  HepLorentzVector boost = plab1+plab2;
	  HepLorentzRotation RotBoost(boost.boostVector());
	  
	  if( nprimary != 0 ) 
	    {
	      evt.flight = sqrt((xs-primary[primvtx]->xv)*(xs-primary[primvtx]->xv)+
				(ys-primary[primvtx]->yv)*(ys-primary[primvtx]->yv)+
				(zs-primary[primvtx]->zv)*(zs-primary[primvtx]->zv))/(float)RotBoost.tt();
	    }
	  else
	    evt.flight = -100. ;
	  
	  if ( evt.flight < FlightCut_saved ) continue;

          // -----------------------------------------------
          //    Create RVER
	  switch( (int) evt.part )
	    {
	    case 1:
	      vtx.Fit(tpk0s);
	      GroverAddToRVER(&vtx, FIT_SHORT, tpk0s);
	      break;
	    case 2:
	      vtx.Fit(tplmb);
	      GroverAddToRVER(&vtx, FIT_SHORT, tplmb);
	      break;
	    case 3:
	      vtx.Fit(tplmb_b);
	      GroverAddToRVER(&vtx, FIT_SHORT, tplmb_b);
	      break;
	    }
	  // -----------------------------------------------
	  
	  n_selected++;

	  if( Online ) 
	    {
	      rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 8000 ), evt.part , 1. ); 
		       
	      rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 8000+(int)evt.part ), evt.mass , 1. ); 

	    }
	  else
	    {
	      hf1( 8000 , evt.part , 1. ); 
	      hf1( 8000+(int)evt.part , evt.mass , 1. ); 
	    }
	  
	  if( !FillNtuple ) continue;

	  if( evt.part == 1 && max(p[i]/p[j],p[j]/p[i]) > 4. ) continue;
	  
	  evt.run    =(float)(*ArtePointer<EVHD>(1)).run;
	  evt.evt    =(float)(*ArtePointer<EVHD>(1)).evt;
	  evt.time   =(float)(*ArtePointer<EVHD>(1)).time;
	  evt.bx     =(float)faDDAQphysBX();
	  evt.tcode  =(float)faDDAQFCS_tcode();
	  
	  evt.distv = mindist;    

	  evt.nhtr1 = track[i]->hou+1000*track[i]->hin+1000000*track[i]->hsi;
	  evt.nhtr2 = track[j]->hou+1000*track[j]->hin+1000000*track[j]->hsi;
	  
	  evt.xs = xs;
	  evt.ys = ys;
	  evt.zs = zs;
	  
	  evt.px1 = px[i];
	  evt.px2 = px[j];
	  
	  evt.py1 = py[i];
	  evt.py2 = py[j];
	  
	  evt.pz1 = pz[i];
	  evt.pz2 = pz[j];
	  
	  plab1 *= RotBoost.inverse();
	  plab2 *= RotBoost.inverse();
	  
	  evt.px1lab = plab1.px();
	  evt.py1lab = plab1.py();
	  evt.pz1lab = plab1.pz();
	  
	  evt.px2lab = plab2.px();
	  evt.py2lab = plab2.py();
	  evt.pz2lab = plab2.pz();


	  if( FillNtuple == 1 ) 
	    {
	      evt.vxf1 = track[i]->xf ;
              evt.vyf1 = track[i]->yf ;
              evt.vzf1 = track[i]->zf ;
              evt.vtxf1 = track[j]->xf ;
              evt.vtyf1 = track[j]->yf ;
              evt.vpf1 = track[j]->zf ;      
	    }
	  else if( FillNtuple == 2 ) 
	    {
	      // RSEG loop 1 (many RSEG per RTRA)
	      for (ArteRelationIterator<RSEG> rseg = track[i]->rseg.begin();
		   rseg != track[i]->rseg.end(); ++rseg){
		
		if( (*rseg)->fit <= 0)
		  {
		    
		    if ((*rseg)->cmp & Rsegc::vxd) 
		      {
			evt.vzf1 = (*rseg)->zf;
			evt.vyf1 = (*rseg)->yf;
			evt.vxf1 = (*rseg)->xf;
			
			evt.vze1 = (*rseg)->ze;
			evt.vye1 = (*rseg)->ye;
			evt.vxe1 = (*rseg)->xe;
			
			evt.vtxf1 = (*rseg)->txe;
			evt.vtyf1 = (*rseg)->tye;
			
			evt.vpf1  = (*rseg)->pf;
			
			evt.vtxe1 = (*rseg)->txe;
			evt.vtye1 = (*rseg)->tye;
			
			evt.vpe1  = (*rseg)->pe;
			
		      }
		    else if ((*rseg)->cmp &  Rsegc::patt )
		      {
			evt.tzf1 = (*rseg)->zf;
			evt.tyf1 = (*rseg)->yf;
			evt.txf1 = (*rseg)->xf;
			
			evt.tze1 = (*rseg)->ze;
			evt.tye1 = (*rseg)->ye;
			evt.txe1 = (*rseg)->xe;
			
			evt.ttxf1 = (*rseg)->txf;
			evt.ttyf1 = (*rseg)->tyf;
			
			evt.tpf1   = (*rseg)->pf;
			
			evt.ttxe1 = (*rseg)->txe;
			evt.ttye1 = (*rseg)->tye;
			
			evt.tpe1  = (*rseg)->pe;
		      }
		  }
	      }

	      for (ArteRelationIterator<RSEG> rseg = track[j]->rseg.begin();
		   rseg != track[j]->rseg.end(); ++rseg){
		
		if( (*rseg)->fit <= 0)
		  {
		    
		    if ((*rseg)->cmp & Rsegc::vxd) 
		      {
			evt.vzf2 = (*rseg)->zf;
			evt.vyf2 = (*rseg)->yf;
			evt.vxf2 = (*rseg)->xf;
			
			evt.vze2 = (*rseg)->ze;
			evt.vye2 = (*rseg)->ye;
			evt.vxe2 = (*rseg)->xe;
			
			evt.vtxf2 = (*rseg)->txf;
			evt.vtyf2 = (*rseg)->tyf;
			
			evt.vpf2   = (*rseg)->pf;
			
			evt.vtxe2 = (*rseg)->txe;
			evt.vtye2 = (*rseg)->tye;
			
			evt.vpe2   = (*rseg)->pe;
			
		      }
		    else if ((*rseg)->cmp &  Rsegc::patt )
		      {
			evt.tzf2 = (*rseg)->zf;
			evt.tyf2 = (*rseg)->yf;
			evt.txf2 = (*rseg)->xf;
			
			evt.tze2 = (*rseg)->ze;
			evt.tye2 = (*rseg)->ye;
			evt.txe2 = (*rseg)->xe;
			
			evt.ttxf2 = (*rseg)->txf;
			evt.ttyf2 = (*rseg)->tyf;
			
			evt.tpf2 = (*rseg)->pf;
			
			evt.ttxe2 = (*rseg)->txe;
			evt.ttye2 = (*rseg)->tye;
			
			evt.tpe2 = (*rseg)->pe;
			
		      }
		  }
	      }
	    }


	  hfn(3,(float*)&evt);
       
	}
    }
  
return n_selected;
}


bool K0SLAMBDA::MassCut( int candidate , float mass )
{

  switch ( candidate ) 
    {
    case 1: 
      if ( mass > K0LowMassCut_saved && mass < K0HighMassCut_saved ) return TRUE; 
      break;
    case 2: 
    case 3: 
      if ( mass > LambdaLowMassCut_saved && mass < LambdaHighMassCut_saved ) return TRUE;
      break;
    }

  return FALSE;
}


void K0SLAMBDA::SelectNextTry(float *candidate, int Q, float *mass1, float *mass2) 
{
	
  switch( (int) *candidate )
    {
    case 1:  // Last try was a K0s.
      // Lambda 
      if( Q > 0 ) 
	{ 
	  *mass1 = Proton;
	  *mass2 = Pion;
	}
      else 
	{
	  *mass1 = Pion;
	  *mass2 = Proton;  
	}
      *candidate = 2.;
      break;
      
    case 2:  // Last try was a Lambda.
      // Lambda bar
      if( Q < 0 ) 
	{ 
	  *mass1 = Proton;
	  *mass2 = Pion;
	}
      else 
	{
	  *mass1 = Pion;
	  *mass2 = Proton;  
	}
      *candidate = 3.;
      break; 
    case 3: // Last try was a Lambda_bar.
      *mass1 = Pion;
      *mass2 = Pion;
      *candidate = 1.;
      break;
    }
} 

int K0SLAMBDA::trSelect( vector< ArtePointer<RTRA> > &rtra, int &clpos, int &clneg)
// *---------------------------------------------------------------------
// * return a vector with ArtePointers into RTRA
// *---------------------------------------------------------------------
{
  static int MINHITS_VDS  =  5;   // required minimum number of VDS hits
  static int MINHITS_MAIN =  5;  // required minimum number of Main tracker hits
  static int MINHITS_ITR  =  5;  // required minimum number of Main tracker hits

  map< float, ArtePointer<RTRA> > mrtra;

  //------------------------------------

  // put all VDS-OTR tracks into a sorted map

  for(ArteTable<RTRA>::iterator itrtra = ArteTable<RTRA>::begin();
      itrtra != ArteTable<RTRA>::end(); ++itrtra) 
  {
    if( (*itrtra)->fit != 0 ) continue;
    
    int nvds  = (*itrtra)->hsi;  //number of hits in inner tracker
    int nmain = (*itrtra)->hou+(*itrtra)->hin;  //number of hits in inner/outer tracker
    int notr  = (*itrtra)->hou;
    int nitr  = (*itrtra)->hin;
    
    if( nvds  < MINHITS_VDS ) continue;
    if( nmain < MINHITS_MAIN && nitr < MINHITS_ITR && notr < MINHITS_MAIN ) continue;
    
    float fom = float(nvds+nmain)-(*itrtra)->chi/float(nvds+nmain-5);
    mrtra.insert(pair< float, ArtePointer<RTRA> >(-fom,*itrtra));
  }
  // copy the tracks into the return vector, omitting all clones
  
  clpos = clneg = 0;
  for(map< float, ArtePointer<RTRA> >::iterator it=mrtra.begin();it != mrtra.end(); ++it) 
    {
      int OK = 1;
      for(unsigned int i=0; i!=rtra.size(); ++i) 
	{
      //	  if(rtra[i]->csvd  == it->second->csvd ) OK=0;
     	  if(rtra[i]->cmain == it->second->cmain) OK=0;
	  if(!OK) break;
	}
      if(OK) 
	{
	  rtra.push_back(it->second);
	  if(it->second->pf < 0) clneg++;
	  if(it->second->pf > 0) clpos++;
	}
    }

  return rtra.size();
}

float K0SLAMBDA::distance(float vect[3],float pos[3],float poin[3])
{
  float diff;
  float vectn=0.0;
  float dd=0.0;
  float diffn=0.0;
  
  for(int i=0;i<3;i++)
    {
      diff = poin[i]-pos[i];
      dd += diff*vect[i];
      vectn += vect[i]*vect[i];
      diffn += diff*diff;
    }

  if( dd > 0 ) 
    return -sqrt(diffn-dd*dd/vectn);
  else 
    return sqrt(diffn-dd*dd/vectn);

}
