// ********************************************************************
//
//  Routine:      rcSetRtraHitb( ArtePointer<RTRA> )
//                rcSetRtraSegHitb( ArtePointer<RTRA>, ArtePointer<RSEG> )
//                rcCreateRtraRhit( ArtePointer<RTRA> )
//                rcCountRtraHits( ArtePointer<RTRA> )
//  Author:       T. Oest
//
//  Description:  set relations between hitb and rtra; create rhit
//                it is important that the order of the related segments
//                is correct
//
//  Modification:
//  - adapted to new reco chain,  W.Hulsbergen, 20-01-2001
//  - Added rcSetRtraSegHitb, to add hitb's from a particular segment.
//                                P.Conde (24-01-2001)
//
// ********************************************************************

#include <iostream>
#include <vector>
#include <arte/arte.hh>
#include <gean/ZGESL.h>

// ********************************************************************
//
//  Routine:     rcSetRtraHitb 
//  Description: creates relation between rtra and the hitbs related
//               to its rsegs
// 
// ********************************************************************

int rcSetRtraHitb( ArtePointer<RTRA> rtra )
{
  // Create relations between this rtra and the hitbs related to its
  // rsegs. Take care the order is correct for the refit.
  int rc=0;
  if( rtra->hitb.begin() !=  rtra->hitb.end() ) { 
    cout << "rcSetRtraHitb (rtra = " << rtra << "):" 
	 << "  Hitbs already assigned." << endl ;
    rc = 1;
  } else {
    for( ArteRelationIterator<RSEG> rseg = rtra->rseg.begin();
	 rseg != rtra->rseg.end(); ++rseg ) {
      // the createRelation inverts the order. we do not want that.
      vector<ArtePointer<HITB> > hitbs ;
      for( ArteRelationIterator<HITB> hitb = (*rseg)->hitb.begin();
	   hitb != (*rseg)->hitb.end(); ++hitb ) 
	hitbs.push_back(*hitb) ;
	for(vector<ArtePointer<HITB> >::reverse_iterator 
	      hitb = hitbs.rbegin() ; hitb != hitbs.rend() ; hitb++)
	  createRelation( rtra, *hitb, &RTRA::hitb );
    }
  }
  return rc;
}

// ********************************************************************
//
//  Routine:     rcSetRtraSegHitb 
//  Description: creates relation between rtra and the hitbs related
//               to the segment rseg. Take care the order is correct 
//               for the refit.
// 
// ********************************************************************


int rcSetRtraSegHitb( ArtePointer<RTRA> rtra, ArtePointer<RSEG> rseg )
{

  int rc=0;
  // the createRelation inverts the order. we do not want that.

  vector<ArtePointer<HITB> > hitbs ;
  for( ArteRelationIterator<HITB> hitb = rseg->hitb.begin();
       hitb != rseg->hitb.end(); ++hitb ) 
    hitbs.push_back(*hitb) ;

  for(vector<ArtePointer<HITB> >::reverse_iterator 
	hitb = hitbs.rbegin() ; hitb != hitbs.rend() ; hitb++){

    // Check that this hit is not already in rtra:

    int skip=0;
    for(ArteRelationIterator<HITB> h2 = rtra->hitb.begin();
	h2 != rtra->hitb.end(); h2++){
      if((*hitb) == (*h2)) skip =1; // Hit already in RTRA. Skip it.
    }
    if(skip == 0 ) createRelation( rtra, *hitb, &RTRA::hitb );
  }
  return rc;
}



// ********************************************************************
//
//  Routine:     rcCreateRtraRhit( ArtePointer<RTRA> rtra ) 
//  Description: creates rhits for this rtra
//
// ********************************************************************

int rcCreateRtraRhit( ArtePointer<RTRA> rtra ) {

  // check if there already exist rhits for this rtra
  int nrhit=0;
  for( ArteRelationIterator<HITB> ihitb = rtra->hitb.begin();
       ihitb != rtra->hitb.end() && !nrhit; ++ihitb ) 
    for ( ArteRelationIterator<RHIT> irhit = (*ihitb)->rhit.begin();
	  irhit != (*ihitb)->rhit.end() && !nrhit; ++irhit ) 
      if( (*irhit)->rtra == rtra.index() ) nrhit++ ;
  
  if(nrhit) {
    static int n=0;
    if(++n<100) cout << "rcCreateRtraRhit (rtra = " << rtra << "):" 
		     << "  Rhits already assigned." << endl ;
  } else {
    for( ArteRelationIterator<RSEG> rseg = rtra->rseg.begin();
	 rseg != rtra->rseg.end(); ++rseg )
      for( ArteRelationIterator<RHIT> rhit = (*rseg)->rhit.begin();
	   rhit != (*rseg)->rhit.end(); ++rhit) {
	ArtePointer<RHIT> rhitnew = ArteTable<RHIT>::push_back(RHIT()) ;
	rhitnew->x  = (*rhit)->x;
	rhitnew->y  = (*rhit)->y;
	rhitnew->tx = (*rhit)->tx;
	rhitnew->ty = (*rhit)->ty;
	rhitnew->q  = (*rhit)->q;
	rhitnew->p  = rtra->pf ; // !!
	rhitnew->rtra = rtra ;
	createRelation( (*rhit)->hitb, rhitnew, &HITB::rhit );
      } 
  }
  return 0;
}


// ********************************************************************
//
//  Routine:     rcCountRtraHits( ArtePointer<RTRA> rtra ) 
//               rcCountRtraHits()
//  Description: fill hca, hou, hin, hsi, bsi and bmain for this rtra
//  bsi:    bit 1-32 are vds layers 1-32 (note that 2,6,10 are not built :)
//  bmain:  bit 1-13  are otr super layer 1-13
//          bit 14-22 are itr superlayers 1-9
//
// ********************************************************************

int rcCountRtraHits( ArtePointer<RTRA> rtra) {

  // loop over clusters and count
  rtra->hca = 0;
  for( ArteRelationIterator<RCCL> rccl = rtra->rccl.begin();
       rccl != rtra->rccl.end(); ++rccl )
    rtra->hca += int((*rccl)->wid);

  // loop over hitbs and count
  rtra->bmain = rtra->bsi = 0;
  rtra->hsi = rtra->hou = rtra->hin = 0;
  for( ArteRelationIterator<HITB> hitb = rtra->hitb.begin();
       hitb != rtra->hitb.end(); ++hitb ) {
    if(       (*hitb)->gede->cmp == LGESL_SVD ) {
      rtra->hsi++ ;
      int layer = 4 * ((*hitb)->gede->gesl->nsl - 1) + (*hitb)->gede->lay - 1;
      rtra->bsi |= 1 << layer ;
    } else if((*hitb)->gede->cmp == LGESL_OTR ) {
      rtra->hou++ ;
      int layer = (*hitb)->gede->gesl->nsl - 1 ;
      rtra->bmain |= 1 << layer ;
    } else if((*hitb)->gede->cmp == LGESL_ITR ) {
      rtra->hin++ ;
      int layer = (*hitb)->gede->gesl->nsl - 1 ;
      rtra->bmain |= 1 << ( 13 + layer ) ;
    }
  }

  return 0;
}


int rcCountRtraHits() {
  for(ArteTable<RTRA>::iterator irtra = ArteTable<RTRA>::begin() ;
      irtra != ArteTable<RTRA>::end() ; ++irtra)
    if( (*irtra)->fit%100 == 0 ) rcCountRtraHits(*irtra) ;
  return 0;
}


// ********************************************************************
//
//  Routine: rcSetRTRArelations 
//
// Description: Kept for backward compatibility. Sets the relations
// from RTRA to HITB (via RSEG) and creates the RHITs
//
// ********************************************************************

int rcSetRtraHitRelations() {
  
  for(ArteTable<RTRA>::iterator irtra = ArteTable<RTRA>::begin() ;
      irtra != ArteTable<RTRA>::end() ; ++irtra) 
    if( (*irtra)->fit%100 == 0 ) {
      rcSetRtraHitb(*irtra) ;
      rcCreateRtraRhit(*irtra) ;
    }

  return 0 ;
}


