// ********************************************************************
//
//  Routine:      rcSetTrackPoints
//  Author: Wouter Hulsbergen (January 2001)
//
//  Description: Create RPNTs from tracker tracks that have not been
//  refitted. The points are made at the end of pc segments and at
//  beginning and end of tc segments.
//
//  Modification:
//
// ********************************************************************
 
#include "arte/arte.hh"
#include "reco/rcflags.hh"
#include "arte/RSEG_cmp.hh"

int rcSetTrackPoints() {

  for(ArteTable<RTRA>::iterator rtra=ArteTable<RTRA>::begin();
      rtra!=ArteTable<RTRA>::end();rtra++) 
    if( !( rtra->cmp & Rsegc::bitsign) &&
	( rtra->cmp & Rsegc::patt) &&
	distance(rtra->rpnt.begin(),rtra->rpnt.end())==0 ) {

      // let's keep this really simple: create rpnts at end of pc and
      // begin and end of tc segment.
      
      for(ArteRelationIterator<RSEG> rsegIt = rtra->rseg.begin() ;
	  rsegIt != rtra->rseg.end() ; rsegIt++) {
	ArtePointer<RSEG> rseg = *rsegIt ;
	if( !(rseg->cmp & Rsegc::bitsign) &&
	    (rseg->cmp & ( Rsegc::patt | Rsegc::trit) ) ) {

	  if( rseg->cmp & Rsegc::trit) {
	    RPNT rpnt;
	    rpnt.zf = rseg->zf;
	    rpnt.xf = rseg->xf;
	    rpnt.yf = rseg->yf;
	    rpnt.txf = rseg->txf;
	    rpnt.tyf = rseg->tyf;
	    rpnt.pf = rseg->pf;
	    memcpy(rpnt.cvf,rseg->cvf,15*sizeof(float)) ;
	    ArtePointer<RPNT> Prpnt = ArteTable<RPNT>::push_back(rpnt);
	    createRelation( *rtra, Prpnt, &RPNT::rtra );
	  }

	  RPNT rpnt;
	  rpnt.zf = rseg->ze;
	  rpnt.xf = rseg->xe;
	  rpnt.yf = rseg->ye;
	  rpnt.txf = rseg->txe;
	  rpnt.tyf = rseg->tye;
	  rpnt.pf = rseg->pe;
	  memcpy(rpnt.cvf,rseg->cve,15*sizeof(float)) ;
	  ArtePointer<RPNT> Prpnt = ArteTable<RPNT>::push_back(rpnt);
	  createRelation( *rtra, Prpnt, &RPNT::rtra );
	}
      }
    }

  return 0; 
}


// ********************************************************************
//
//  Routine:      rcSetTrackPointsVdsEcal
//  Author: Wouter Hulsbergen (January 2001)
//
// Description: Create RPNTs for vds-ecal tracks that were _not_
// refitted with the vds-ecal fit. The first RPNT is created at the
// end of the vds segment. The second RPNT is created at the z
// position of the ecal cluster, by extrapolation of the first RPNT.
//
//  Modification:
//  -updated for new vds-ecal refit    WH (June 2001)
//
// ********************************************************************

#include "marple/RcRpntPoint.hh"

int rcSetTrackPointsVdsEcal() {

  for(ArteTable<RTRA>::iterator rtra=ArteTable<RTRA>::begin();
      rtra!=ArteTable<RTRA>::end();rtra++) 
    if( !( rtra->cmp & Rsegc::bitsign) &&
       !( rtra->cmp & Rsegc::patt) &&
       (rtra->cmp & Rsegc::vxd) && (rtra->cmp & Rsegc::ecal) &&
	rtra->mommask!=RCMOMREFIT) {
 
      ArtePointer<RPNT> rpnt1 = 0 ;
      if(distance(rtra->rpnt.begin(),rtra->rpnt.end())==0) {
	// Track has not been refitted. Create 1 rpnt at end of vds segment.
	ArtePointer<RSEG> rseg(rtra->csvd) ;
	RPNT rpnt;
	rpnt.zf = rseg->ze;
	rpnt.xf = rseg->xe;
	rpnt.yf = rseg->ye;
	rpnt.txf = rseg->txe;
	rpnt.tyf = rseg->tye;
	rpnt.pf = rtra->pf;
	memcpy(rpnt.cvf,rseg->cve,10*sizeof(float)) ;
	memcpy(rpnt.cvf+10,rtra->cvf+10,5*sizeof(float)) ;
	rpnt1 = ArteTable<RPNT>::push_back(rpnt);
	createRelation( *rtra, rpnt1, &RPNT::rtra );
      } else {
	// The refit can create more than one rpnt. Appearently, the
	// momentum is not set by the refit, otherwise
	// momflag==RCMOMREFIT would be true. We assume that only the
	// vds part was refitted, as in the old vds-ecal refit. Then
	// we overwrite the RPNT momentum with that from the RTRA.
	for(ArteRelationIterator<RPNT> irpnt = rtra->rpnt.begin() ;
	    irpnt != rtra->rpnt.end() ; ++irpnt) {
	  (*irpnt)->pf = rtra->pf;
	  memcpy((*irpnt)->cvf+10,rtra->cvf+10,5*sizeof(float)) ;
	  if(rpnt1.index()==0 || rpnt1->zf < (*irpnt)->zf) rpnt1 = *irpnt ;
	}
      }

      // create new rpnt from first rpnt and transport to ecal
      RPNT rpnt;
      rpnt.zf = rpnt1->zf;
      rpnt.xf = rpnt1->xf;
      rpnt.yf = rpnt1->yf;
      rpnt.txf = rpnt1->txf;
      rpnt.tyf = rpnt1->tyf;
      rpnt.pf = rpnt1->pf;
      memcpy(rpnt.cvf,rpnt1->cvf,15*sizeof(float)) ;
      ArtePointer<RPNT> rpnt2 = ArteTable<RPNT>::push_back(rpnt);
      createRelation( *rtra, rpnt2, &RPNT::rtra );
      
      RcRpntPoint rcrpn(rpnt2);
      int ierr ;
      rcrpn.prolongate(ArtePointer<RCCL>(rtra->cecal)->z,ierr) ;
      if(ierr==0) rcrpn.updateARTE(ierr) ; 
    }
  
  return 0; 
}


