// *****************************************************************************
//
// source:    
//
// type:      source code
//
// created:   26 Jul 2000
//	
// author:    Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// changes:
// 26 Jul 2000 (AG) creation (code from Denis)
// 31 Jul 2000 (DD) scan for primary vertex/wire added
// 04 Aug 2000 (DD) vtchi not called before vtfit (T.Glebe)
// 16 Aug 2000 (DD) option for v0 constraint
// 17 Aug 2000 (DD) option for exclusion of v0 tracks
// 08 Sep 2000 (DD) option to skip changing of rtra's in vtmass 
// 09 Sep 2000 (DD) decay covariance matrix fixed (vtfit, vtmass)
// 10 Sep 2000 (DD) chi2 fixed (vtmass)
// 13 Sep 2000 (DD) interface to T.L.'s secondary vertex pattern recognition
// 28 Oct 2000 (DD) Grover vertex fitting
// 15 Jan 2001 (DD) removed printout
//
// *****************************************************************************

#include "reco/rcXvertex.h"
#include "reco/rcXsegment.h"
#include "reco/rcXmessage.h"
#include "grover/grover.hh"


map<int, RVER, less<int> >           vertexmass::v0_average_list;
map<int, RVER, less<int> >::iterator vertexmass::v0_average_list_it;


vertexmass::vertexmass() {
  track_list = new vector<TRACK>;
  wire_list  = new vector<ArtePointer<GTAR> >;
  findUsedWires();
}


void vertexmass::bookRVER() {
  // -- book rver
  RVER rver;
  rver.xv=v[0];
  rver.yv=v[1];
  rver.zv=v[2];
  for (int i=0; i<3; i++)
    for (int j=i; j<3; j++) rver.cv[segment::covij[i][j]]=cs[i][j];
  rver.mass=Mass;
  rver.merr=Merr;
  rver.nt=track_list->size();
  rver.wire=0;
  ArteTable<RVER>::push_back(rver);
  ArtePointer<RVER> prver(ArteTable<RVER>::size());

  // -- book rtra-rrel-rver
  for (unsigned int i=0; i<track_list->size(); i++) {
    RREL rrel;
    rrel.flag = 1;
    rrel.chi  = chi2;
    ArteTable<RREL>::push_back(rrel);
    ArtePointer<RREL> prrel(ArteTable<RREL>::size());
    createRelation(                 prver, prrel, &RVER::rrel );
    createRelation( (*track_list)[i].rtra, prrel, &RTRA::rrel );
  }
}



int vertexmass::fitVertex() {

  if (track_list->size()<2&&vattrib.rlevel!=6)  
    return RCX_VERTEX_CUT;

  // -- fitting algorithms
  switch (vattrib.rlevel) {
  case 1: vtchi();  break;
  case 2:
  case 3: vtfit();  break;
  case 4: vtrver(); break;
  case 5:
  case 6: grover(); break;
  case 7: vtsec();  break;
  };

  // -- apply cuts
  switch (vattrib.rlevel) {
  case 1: 
  case 2:
  case 3: 
  case 4: 
  case 5: 
    if (vattrib.verb) printVertex();
    if (chi2>vattrib.max_vertex_chi2)          return RCX_VERTEX_CUT;
    float fpar[]={v[0],v[1],v[2], par[0],par[1]};
    // -- old approach: find wire, primary vertex
    //if (vattrib.check_wire && checkWire(fpar)) return RCX_VERTEX_CUT;
    //if (findPrimaryVertex());//                   return RCX_VERTEX_CUT;
    // -- new approach: find primary vertex, wire
    int vrf=checkPrimaryVertex(fpar);
    if (vrf && vattrib.check_wire)             return vrf;
    break;
  }
  return RCX_OK;
}


void vertexmass::printVertex() {
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    cout<<"track "<<(*it).rtra
	<<"  x="<<(*it).x[0]<<","<<(*it).x[1]<<","<<(*it).x[2]
	<<"  n="<<(*it).n[0]<<","<<(*it).n[1]<<","<<(*it).n[2]
	<<endl;
  }
  cout<<"vertex: "<<"x="<<v[0]<<","<<v[1]<<","<<v[2]
      <<"  chi="<<chi2<<endl<<endl;
}

void vertexmass::addVertexTrack(ArtePointer<RTRA> tp, ArtePointer<TOPP> to) {
  TRACK tr;
  tr.rtra=tp;
  tr.x[0]=tp->xf;
  tr.x[1]=tp->yf;
  tr.x[2]=tp->zf;
  tr.n[0]=tp->txf; tr.n[1]=tp->tyf;
  float inorm=1/sqrt(tr.n[0]*tr.n[0]+tr.n[1]*tr.n[1]+1);
  tr.n[0]=tp->txf*inorm; tr.n[1]=tp->tyf*inorm; tr.n[2]=inorm;
  tr.topp=to;
  track_list->push_back(tr);
}


void vertexmass::removeLastVertexTrack() {
  if (!track_list->size()) return;
  vector<TRACK>::iterator it=track_list->end();
  track_list->erase(--it); 
}

vector<ArtePointer<GTAR> >::iterator vertexmass::findGTAR(float *fpar) {
  // fpar[5]: x,y,z,tx,ty
  float min_drho2=vattrib.max_drho_from_wire*vattrib.max_drho_from_wire;
  vector<ArtePointer<GTAR> >::iterator iwire=wire_list->end();
  for (vector<ArtePointer<GTAR> >::iterator igtar=wire_list->begin();
       igtar!=wire_list->end(); ++igtar) {
    // -- decay behind/at a wire
    if ((*igtar)->z-fpar[2]>.5) continue; /////////// hard-coded vertex resolution!!!!
    float dz=(*igtar)->z-fpar[2];
    float dx=fpar[0]+fpar[3]*dz - (*igtar)->x;
    float dy=fpar[1]+fpar[4]*dz - (*igtar)->y;
    float drho2=dx*dx+dy*dy;
    if (drho2<min_drho2) { min_drho2=drho2; iwire=igtar; }
  }
  //cout<<"best drho2="<<min_drho2<<endl;
  return iwire;
}


void vertexmass::followRVER() {
  static int old_event=0;
  if (old_event==ArtePointer<EVHD>(1)->evt && v0_average_list.size()>0) return;
  old_event=ArtePointer<EVHD>(1)->evt;

  for (ArteTable<RVER>::iterator irver=ArteTable<RVER>::begin();
       irver!=ArteTable<RVER>::end(); ++irver) {
    if ((*irver)->nt<4)                        continue; 
    if ((*irver)->wire<1||(*irver)->wire>8)    continue;  
    int iw=(*irver)->wire;
    v0_average_list_it = v0_average_list.find(iw);
    if (v0_average_list_it!=v0_average_list.end()) {
      (*v0_average_list_it).second.xv = (*v0_average_list_it).second.xv*.99 + (*irver)->xv*.01;
      (*v0_average_list_it).second.yv = (*v0_average_list_it).second.yv*.99 + (*irver)->yv*.01;
      (*v0_average_list_it).second.zv = (*v0_average_list_it).second.zv*.99 + (*irver)->zv*.01;
    }
    else {
      //cout<<"vertexmass: adding new wire to v0 follower:"
      //<<"  wire="<<(*irver)->wire<<"  "<<(*irver)->xv<<" "<<(*irver)->yv<<" "<<(*irver)->zv<<endl;
      RVER newv0;
      memset(&newv0,0x0,sizeof(RVER));
      newv0.xv=(*irver)->xv;
      newv0.yv=(*irver)->yv;
      newv0.zv=(*irver)->zv;
      newv0.wire=-(*irver)->wire;
      for (int i=0; i<6; i++) newv0.cv[i]=(*irver)->cv[i];
      newv0.nt=(*irver)->nt;
      v0_average_list.insert(pair<int,RVER> (iw, newv0));
    }
  }

  int local_verb=0;
  if (local_verb) {
    cout<<"Primary vertex follower for event="<<ArtePointer<EVHD>(1)->evt<<":"<<endl;
    for (v0_average_list_it=v0_average_list.begin(); 
	 v0_average_list_it!=v0_average_list.end(); 
	 ++v0_average_list_it) {
      cout<<"wire="<<(*v0_average_list_it).first
	  <<"  x="<<(*v0_average_list_it).second.xv
	  <<"  y="<<(*v0_average_list_it).second.yv
	  <<"  z="<<(*v0_average_list_it).second.zv<<endl;
    }
  }

}


ArteTable<RVER>::iterator vertexmass::findRVER(float *fpar) {
  // fpar[5]: x,y,z,tx,ty
  float min_drho2=vattrib.max_drho_from_wire*vattrib.max_drho_from_wire;
  ArteTable<RVER>::iterator iv0=ArteTable<RVER>::end();
  for (ArteTable<RVER>::iterator irver=ArteTable<RVER>::begin();
       irver!=ArteTable<RVER>::end(); ++irver) {
    // -- decay before v0
    float z_error=((*irver)->cv[5]>0) ? 10*sqrt((*irver)->cv[5]) : .5;/////////// hard-coded vertex resolution!!!!
    if ((*irver)->zv-fpar[2]>.5)               continue; /////////// hard-coded vertex resolution!!!!replace with z_error
    if ((*irver)->nt<vattrib.min_tracks_in_v0) continue; 
    if ((*irver)->wire<1||(*irver)->wire>8)    continue;  
    float dz=(*irver)->zv-fpar[2];
    float dx=fpar[0]+fpar[3]*dz - (*irver)->xv;
    float dy=fpar[1]+fpar[4]*dz - (*irver)->yv;
    float drho2=dx*dx+dy*dy;
    if (drho2<min_drho2) { min_drho2=drho2; iv0=(*irver); }
  }
  return iv0;
}

int vertexmass::checkPrimaryVertex(float *fpar) {
  ArteTable<RVER>::iterator irver=findRVER(fpar);
  if (irver==ArteTable<RVER>::end())                    return RCX_WIRE_CUT;
  v0_rver=(*irver);
  decay_wire=ArtePointer<GTAR>(v0_rver->wire);
  if (vattrib.verb)
    cout<<"primary vertex: "<<v0_rver<<" "<<v0_rver->xv<<" "<<v0_rver->yv<<" "<<v0_rver->zv<<"  wire="<<v0_rver->wire<<endl;
  // -- check decay length
  if (fpar[2]-v0_rver->zv<vattrib.min_dz_from_wire)     return RCX_VERTEX_CUT;
  // -- check if tracks come from primary vertex
  int nv0t=0;
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    if (IsInVertex(v0_rver, (*it).rtra)) nv0t++;
  }
  if (nv0t<vattrib.require_v0_tracks)                   return RCX_EXCLUDE_V0;
  if (nv0t>vattrib.allowed_v0_tracks)                   return RCX_EXCLUDE_V0;
  return RCX_OK;
}


bool vertexmass::IsInVertex(ArtePointer<RVER> rver, ArtePointer<RTRA> rtra) {
  for (ArteRelationIterator<RREL> irrel=rver->rrel.begin();
       irrel!=rver->rrel.end(); ++irrel) {
    ArtePointer<RTRA> prong=(*irrel)->rtra;
    if (prong==rtra)                   return true;
  }
  return false;
}



int vertexmass::checkWire(float *fpar) {
  // -- try to find the best wire in drho
  vector<ArtePointer<GTAR> >::iterator iwire=findGTAR(fpar);
  if (iwire==wire_list->end())                        return RCX_VERTEX_CUT;
  decay_wire=(*iwire);
  if (vattrib.verb>2) {
    float dz=fpar[2]-(*iwire)->z;
    float dx=fpar[0]+fpar[3]*dz - (*iwire)->x;
    float dy=fpar[1]+fpar[4]*dz - (*iwire)->y;
    float drho2=dx*dx+dy*dy;
    cout<<"input param ="<<fpar[0]<<" "<<fpar[1]<<" "<<fpar[2]<<" "<<fpar[3]<<" "<<fpar[4]<<endl;
    cout<<"nearest wire="<<(*iwire)<<"  dz="<<dz<<"  drho="<<sqrt(drho2)<<endl;
    cout<<"             "<<(*iwire)->x<<" "<<(*iwire)->y<<" "<<(*iwire)->z<<endl;
  }
  // -- check decay length
  if (fpar[2]-(*iwire)->z<vattrib.min_dz_from_wire)   return RCX_VERTEX_CUT;
  return RCX_OK;
}


void vertexmass::findUsedWires() {
  float maxBeamTargetDx=1, maxBeamTargetDy=1;
  for (ArteTable<GTAR>::iterator igtar=ArteTable<GTAR>::begin();
       igtar!=ArteTable<GTAR>::end(); ++igtar) {
    //cout<<(*igtar)<<" "<<(*igtar)->x<<" "<<(*igtar)->y<<" "<<(*igtar)->z<<endl;
    if ( fabs((*igtar)->x)<maxBeamTargetDx &&
         fabs((*igtar)->y)<maxBeamTargetDy) { 
      wire_list->push_back((*igtar));
      if (vattrib.verb) cout<<"using wire="<<(*igtar)<<"  "
			    <<(*igtar)->x<<" "<<(*igtar)->y<<" "<<(*igtar)->z<<endl;
    }
  }
  if (!wire_list->size()) cout<<"vertexmass: cannot find target wires"<<endl;
}


int vertexmass::findPrimaryVertex() {
  int   ret=RCX_VERTEX_CUT;
  float min_dz=.3;
  float min_drho2=.11;
  for (ArteTable<RVER>::iterator irver=ArteTable<RVER>::begin();
       irver!=ArteTable<RVER>::end(); ++irver) {
    if ((*irver)->nt<vattrib.min_tracks_in_v0)  continue;
    if ((*irver)->wire!=(int)decay_wire.index())  continue; 
    float dz=fabs(decay_wire->z-(*irver)->zv);
    if (dz>min_dz)                           continue; 
    float dx=(decay_wire->x-(*irver)->xv);
    float dy=(decay_wire->y-(*irver)->yv);
    float drho2=dx*dx+dy*dy;
    if (drho2>min_drho2) continue;
    min_drho2=drho2;
    v0_rver=(*irver);
    ret=RCX_OK;
  }
  return ret;
}


void vertexmass::findVertexMC() {
  if (!ArteTable<MTRA>::size()) return; 

  // -- find primary vertex
  ArtePointer<MTRA> bmtra = decay_mtra;
  v0_mver                 = bmtra->bver;
  int               flag  = v0_mver->flag;
  while (v0_mver->flag!=1) {
    bmtra=bmtra->btra;
    v0_mver=bmtra->bver;
  }

  // -- find MC vertex
  if (decay_mtra->ever.begin()==decay_mtra->ever.end()) decay_ever=decay_mtra->bver;
  else for (ArteRelationIterator<MVER> iever=decay_mtra->ever.begin(); 
	    iever!=decay_mtra->ever.end(); ++iever) decay_ever=(*iever);

  if (vattrib.verb) cout<<" primary vertex "<<v0_mver<<": "<<v0_mver->x<<" "<<v0_mver->y<<" "<<v0_mver->z<<endl;
  if (vattrib.verb) cout<<" mc vertex "<<decay_ever<<": "<<decay_ever->x<<" "<<decay_ever->y<<" "<<decay_ever->z<<endl;
}



float vertexmass::bchi2(ArtePointer<RTRA> t, ArtePointer<RVER> v0) {
  ArtePointer<RVER> vc=t->ever;

  float dz=v0->zv - t->zf;
  float dz2=dz*dz;

  float sigmax2 = v0->cv[0] + 
    t->cvf[0] + 
    dz2*t->cvf[5] +
    t->txf*t->txf*(vc->cv[5] + v0->cv[5]);

  float sigmay2 = v0->cv[2] + 
    t->cvf[2] + 
    dz2*t->cvf[9] +
    t->tyf*t->tyf*(vc->cv[5] + v0->cv[5]);

  float dx = t->xf + t->txf*dz - v0->xv;
  float dy = t->yf + t->tyf*dz - v0->yv;

//   cout<<"dx="<<dx<<"  sigx="<<sqrt(sigmax2)<<"  dy="<<dy<<"  sigy="<<sqrt(sigmay2)<<"  z0="<<v0->zv<<"  z="<<t->zf<<" "<<endl;

  return (dx*dx/sigmax2 + dy*dy/sigmay2);
}



//////////////////////////////// vertex fit 1 ////////////////////////////////
// -- minimizes chi2 deviations of tracks from unknown vertex
// -- problem can be solved algebraically, w/o fitting
// -- based on code by P.Schlein
extern "C" void rinv_(int &ndim, float *a, int &n, float *s, int &iflag);
void vertexmass::vtchi() {

  static float d[3];
  memset(&cs[0][0],0x0,sizeof(float)*9);
  memset(&d[0],0x0,sizeof(float)*3);
  memset(&v[0],0x0,sizeof(float)*3);
  chi2=1e5;

  // -- loop over tracks
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    float xdotn=0;
    for (int i=0; i<3; i++)  xdotn+=(*it).x[i]*(*it).n[i];
    for (int i=0; i<3; i++) {
      cs[i][i]+=1-(*it).n[i]*(*it).n[i];
      d[i]    +=(*it).x[i]-xdotn*(*it).n[i];
      for (int j=i+1; j<3; j++) {
	cs[i][j]+= -(*it).n[i]*(*it).n[j];
	cs[j][i]=cs[i][j];
      }
    }
  }
  
  // -- compute vertex position
  int   iflag=0, ndim=3;
  float sc[3];
  rinv_(ndim, &cs[0][0], ndim, sc, iflag);
  if (iflag) { return; }
  for (int i=0; i<3; i++) {
    v[i]=0;
    for (int j=0; j<3; j++) {
      v[i]+=cs[i][j]*d[j];
      Cov[segment::covij[i][j]] = cs[i][j];
    }
  }
  
  // -- compute chi^2
  chi2=0;
  float del[3] , isg[3];
  memset(&del[0],0x0,sizeof(float)*3);
  memset(&par[0],0x0,sizeof(float)*5);
  isg[0]=isg[1]=1/.0070; isg[2]=1/.100;
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    float xvdotn=0;
    memset(&del[0],0x0,3*sizeof(float));
    for (int j=0; j<3; j++) xvdotn += ((*it).x[j]-v[j])*(*it).n[j];
    for (int j=0; j<3; j++) del[j]  = ( xvdotn*(*it).n[j] - ((*it).x[j]-v[j]) )*isg[j];
    for (int j=0; j<3; j++) chi2   += ( del[j]*del[j] );
    float pf=fabs((*it).rtra->pf);
    if (pf>1e-4) pf=1/pf;
    for (int i=0; i<3; i++) par[i] +=(*it).n[i]*pf; 
  }
  float pz=sqrt(par[0]*par[0]+par[1]*par[1]+par[2]*par[2]);
  par[0]/=pz; par[1]/=pz; par[4]=par[2];

  chi2/=(track_list->size()*2-3);
}



//////////////////////////////// vertex fit 2 ////////////////////////////////
// -- Kalman filtering method for vertexing 
// -- based on v2fit by T.Lohse
extern "C" {
  void  vtzero_();
  void  vtrini_(int &i);
  void  vtztr_(int &it1, int &it2, double &z1, double &z2);
  void  vtinit_(double *x, double *a, int &iv, int &iflag);
  void  vtadd_(int &iv, int &i, int &iflag);
  void  vtfit_(int &iv, int &momflg, int &iflag);
  float vtchi2_(int &iv, int &k, double *xk, double *qk, double &b13, 
		double &a33, double &a34, int &momflg);
  void  vtsum_(int &iv, double *par, double *cov, int &iflag);
  void  vtaddc_(int &iv, double &xM, bool *mask, int &iflag);
  void  vtmass_(int &iv, double *par, double *cov, double &chi2m, int &iflag);
}
#define NVTMAX 50
#define NTRK   100
#define NTRTOT 300
#define NCOMAX 5
#define NVTTRK 5
extern struct {
  double xv[(NTRK+1)*NVTMAX*3], 
    xvs[NVTMAX*3], 
    qv[NTRK*NVTMAX*3], 
    qvs[NTRK*NVTMAX*3], 
    c[(NTRK+1)*NVTMAX*3*3], 
    cinv[(NTRK+1)*NVTMAX*3*3], 
    f[NTRK*NVTMAX*3*3], 
    gb[NTRK*NVTMAX*5*5], 
    w[NTRK*NVTMAX*3*3], 
    cs[3*NVTMAX*3], 
    ds[3*3*NVTMAX*NTRK], 
    es[NTRK*NVTMAX*3*3], 
    chi2[NTRK*NVTMAX], 
    chi2s[NTRK*NVTMAX],
    xnk[NTRK*NVTMAX*3],
    cnkinv[NTRK*NVTMAX*3*3], 
    cnk[NTRK*NVTMAX*3*3], 
    Ak33[NTRK*NVTMAX], 
    Ak34[NTRK*NVTMAX], 
    Bk13[NTRK*NVTMAX];
  int    fit[NVTMAX], 
    fitm[NVTMAX], 
    smooth[NVTMAX], 
    dist[NVTMAX],
    invflg[NTRTOT];
} vtkalm_;
extern struct {
  double p[NTRTOT*5], v[5*NTRTOT*5], g[5*NTRTOT*5], xmass[NTRTOT];
  int    trmult;
} vttrbf_; 
extern struct {
  int    nconst[NVTMAX];
  double xmc[NVTMAX*NCOMAX];
  bool   mask[NVTMAX*NCOMAX*NTRK];
} vtcons_;
void vertexmass::vtfit() {

  if (track_list->size()<2) return;

  // -- initialize and fill track buffer
  static double cov[3][3]={{1e5,0,0},{0,1e5,0},{0,0,1e5}};
  memset(&v[0],0x0,sizeof(float)*3);
  chi2=1e5;
  static int old_event=0;
  if (vattrib.remove_old_vertices||old_event!=ArtePointer<EVHD>(1)->evt) { 
    vtzero_(); 
    old_event=ArtePointer<EVHD>(1)->evt; 
  }
  int n0=vttrbf_.trmult, ni=0;
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    int itr=(*it).rtra.index();
    if (vattrib.verb) cout<<"vtrini itr="<<itr<<endl;
    vtrini_(itr);
    vttrbf_.xmass[n0+ni]=(*track_list)[ni].topp->mass;
    ni++;
  }
  int n1=n0+ni;
  if (vattrib.verb) {
    cout<<"vertexmass: new tracks="<<ni<<endl;
    for (int i=n0; i<n1; i++) {
      cout<<"p= "<<vttrbf_.p[i*5]<<" "<<vttrbf_.p[i*5+1]<<" "<<vttrbf_.p[i*5+2]<<" "<<vttrbf_.p[i*5+3]<<" "<<vttrbf_.p[i*5+4];
      cout<<"  m="<<vttrbf_.xmass[i]<<"("<<(*track_list)[i].topp->lund<<")"<<endl;
    }
  }

  // -- tmp. vertex position
  double x[]={0,0,0};

  // -- initialize fit, add tracks to vertex
  int iv, iflag=0;
  vtinit_(x, &cov[0][0], iv, iflag);
  if (vattrib.verb) cout<<"vtinit: iv="<<iv<<"  flag="<<iflag<<endl;
  if (iflag<0) return;
  for (int i=n0+1; i<=n1; i++) {
    vtadd_(iv, i, iflag);
    if (vattrib.verb) cout<<"vtadd flag="<<iflag<<endl; 
    if (iflag) return;
  }


  switch (getVertAttrib().rlevel) {
  case 2:
    // -- kalman fit w/o mass constraint
    vtfit_(iv, vattrib.momentum_flag, iflag);
    if (vattrib.verb) cout<<"vtfit flag="<<iflag<<endl;
    if (iflag) return;
    
    // -- get results
    for (int i=0; i<3; i++) v[i] = float(vtkalm_.xvs[(iv-1)*3+i]);
    chi2  = float(vtkalm_.chi2[(iv-1)+NVTMAX*(track_list->size()-1)]);
    chi2 /= (2*track_list->size()-3);
    break;
    
  case 3:
    // -- kalman fit with mass constraint
    double xM=vattrib.topp->mass;
    bool   mask[NTRK];
    for (unsigned int i=0; i<track_list->size(); i++) mask[i]=true;
    vtaddc_(iv, xM, mask, iflag);
    if (vattrib.verb) cout<<"mass constraint="<<vtcons_.xmc[iv-1]<<endl;
    if (vattrib.verb) cout<<"vtmassc flag="<<iflag<<endl;
    double mPar[NTRK*3+3];
    double mCov[(NTRK*3+3)*(NTRK*3+3)];
    double mChi2=1e5;
    vtmass_(iv, mPar, mCov, mChi2, iflag);
    if (vattrib.verb) cout<<"vtmass flag="<<iflag<<endl;
    if (iflag) return;
    // -- get results
    for (int i=0; i<3; i++) v[i]=mPar[i];
    ///////wrong/////////chi2 = float(mChi2)/(2*track_list->size()-3);
    chi2  = float(vtkalm_.chi2[(iv-1)+NVTMAX*(track_list->size()-1)]);
    chi2 /= (2*track_list->size()-3);
    // -- update daughter parameters
    if (vattrib.change_track_rtra) {
      for (unsigned int i=0; i<track_list->size(); i++) {
	ArtePointer<RTRA> rtra=(*track_list)[i].rtra;
	if (vattrib.verb)  cout<<"old pars: "<<rtra->xf<<" "<<rtra->yf<<" "<<rtra->zf<<" "
			       <<rtra->txf<<" "<<rtra->tyf<<" "<<rtra->pf<<endl;
	rtra->xf=v[0];
	rtra->yf=v[1];
	rtra->zf=v[2];
	rtra->txf=mPar[i*3+3];
	rtra->tyf=mPar[i*3+4];
	rtra->pf =((*track_list)[i].topp->chrg<0) ? -1/mPar[i*3+5] : 1/mPar[i*3+5];
	if (vattrib.verb)  cout<<"new pars: "<<rtra->xf<<" "<<rtra->yf<<" "<<rtra->zf<<" "
			       <<rtra->txf<<" "<<rtra->tyf<<" "<<rtra->pf<<endl;
      }
    }
    break;
  }

  
  // -- get covariance matrix
  double dCov[5][5];
  float  tCov[5][5];
  vtsum_(iv,par,&dCov[0][0],iflag);
  //if (vattrib.verb) 
  if (iflag && vattrib.verb) cout<<"vtsum: iflag="<<iflag<<endl;
  // tx,ty,x,y,p -> x,y,tx,ty,1/p
  float pf2=1/(par[4]*par[4]), z2=v[2]*v[2];
  tCov[0][0]=  float(dCov[2][2]) + v[2]*float(dCov[0][2])*2 + z2*float(dCov[0][0]);
  tCov[0][1]=  float(dCov[2][3]) + v[2]*float(dCov[0][3]+dCov[1][2]) + z2*float(dCov[0][1]);
  tCov[0][2]=  float(dCov[2][0]) + v[2]*float(dCov[0][0]);
  tCov[0][3]=  float(dCov[2][1]) + v[2]*float(dCov[0][1]);
  tCov[0][4]=-(float(dCov[2][4]) + v[2]*float(dCov[0][4]))*pf2;

  tCov[1][1]=  float(dCov[3][3]) + v[2]*float(dCov[1][3])*2 + z2*float(dCov[1][1]);
  tCov[1][2]=  float(dCov[3][0]) + v[2]*float(dCov[0][1]);
  tCov[1][3]=  float(dCov[3][1]) + v[2]*float(dCov[1][1]);
  tCov[1][4]=-(float(dCov[3][4]) + v[2]*float(dCov[1][4]))*pf2; 

  tCov[2][2]= float(dCov[0][0]);
  tCov[2][3]= float(dCov[0][1]);
  tCov[2][4]=-float(dCov[0][4])*pf2;

  tCov[3][3]= float(dCov[1][1]);
  tCov[3][4]=-float(dCov[1][4])*pf2; 

  tCov[4][4]= float(dCov[4][4])*pf2*pf2;
  
  for (int i=0; i<5; i++) for (int j=i; j<5; j++) Cov[segment::covij[i][j]]=tCov[i][j];

  // -- also book short cov. matrix
  for (int i=0; i<3; i++) { 
    for (int j=0; j<3; j++) {
      cs[i][j]=float(vtkalm_.cs[((iv-1)*3+j)*3+i]);
    }
  }

  if (vattrib.verb) 
    {
      for (int i=0; i<5; i++) {
	for (int j=0; j<5; j++) {
	  if (j<i) cout<<"\t\t";
	  else cout<<" "<<Cov[segment::covij[i][j]]<<" ";
	}
	cout<<endl;
      }
      for (int i=0; i<3; i++) {
	for (int j=0; j<3; j++) {
	  if (j<i) cout<<"\t\t";
	  else     cout<<" "<<cs[i][j]<<" ";
	}
	cout<<endl;
      }
    }
}

//////////////////////////////// vertex rec 3 ////////////////////////////////
// -- Check for common vertex in RVER bank, dont make any fit
// -- it relies on previously found vertices
void vertexmass::vtrver() {
  for (ArteTable<RVER>::iterator irver=ArteTable<RVER>::begin();
       irver!=ArteTable<RVER>::end(); ++irver) {
    unsigned int nv=0;
    memset(&par[0],0,sizeof(par));
    for (ArteRelationIterator<RREL> irrel=(*irver)->rrel.begin();
	 irrel!=(*irver)->rrel.end(); ++irrel) {
      // -- check if rtra on list of vertex tracks
      ArtePointer<RTRA> vtrtra=(*irrel)->rtra;
      for (vector<TRACK>::iterator it=track_list->begin();
	   it!=track_list->end(); ++it) {
	if (vtrtra==(*it).rtra) { 
	  par[nv] += (*it).n[nv]/fabs(vtrtra->pf); 
	  nv++;
	  break; 
	}
      }
    }
    if (nv!=track_list->size()) continue;
    v[0]=(*irver)->xv;
    v[1]=(*irver)->yv;
    v[2]=(*irver)->zv;
    par[0]/=par[2];
    par[1]/=par[2];
    chi2=(*irver)->chi2;
    break;
  }
}



//////////////////////////////// vertex fit 4 ////////////////////////////////
// -- GROVER fit
// -- based on GROVER package
void vertexmass::grover() {

  // -- init
  GroverWireList     mywires;
  GroverVertexList   myvertices;
  GroverParticleList myparticles;
  GroverFillFromGTAR(&mywires);
  GroverFillFromRSEG(&myparticles);
  mywires.clear();

  // -- fill tracks
  GroverVertex gv;
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    GroverParticle temp(it->rtra);
    gv.Add(&temp);
    gv.particles.back().SetMass(it->topp->mass);
  }


  // -- fit
  float mass_constraint=0;
  if (vattrib.rlevel==6) mass_constraint = vattrib.topp->mass;
  gv.Fit(mass_constraint);  // mass_constraint==0 <=> no mass constraint

  // -- get results for vertex:
  chi2=gv.Chi2();                           
  v[0]=gv.X();  v[1]=gv.Y(); v[2]=gv.Z();    
  cs[0][0] = gv.Cov(0);                     
  cs[1][1] = gv.Cov(2);
  cs[2][2] = gv.Cov(5);
  cs[0][1] = cs[1][0] = gv.Cov(1);
  cs[2][0] = cs[0][2] = gv.Cov(3);
  cs[2][1] = cs[1][2] = gv.Cov(4);

  /////////////  mother particle not yet correctly filled !!! 

  // -- form a mother track
  //gv.EstimateMomentum();
  //GroverParticle mother(&gv);

  //at the position of the decay vertex:
  //Cov[0] = mother.Cov(0);      //x
  //Cov[2] = mother.Cov(2);      //y
  // for the slopes and the momentum part, please wait

  // mother.PropagateBeginTo(wirez);
  // cov. as above

}


////////////// secondary vertex patter recognition ///////////////////////////
// Make sec. vertex reconstruction
// Base on vt package by T. Lohse
extern "C"   {
  void vtprim_(int &itb,int &iflag); 
  void vtsec_ (int &itb,int &iflag);
  void rc99avtrec_(int &iflag);
}
extern struct {
  int nvsec, ivsec[NVTMAX];
} ivsec_;
extern struct {
  int  nvt, ktr[NVTMAX*NTRK], ntr[NVTMAX];
  int  kv[NTRTOT*NVTTRK], nv[NTRTOT], jv[NVTMAX*NTRTOT];
  bool vtset[NVTMAX];
} vtpntr_;
extern struct {
  int vt_rtra[NTRTOT];
} vt_rtra_;
void vertexmass::vtsec() {

  // -- initialize and fill track buffer
  chi2=1e5;
  static int old_event=0;
  if (vattrib.remove_old_vertices||old_event!=ArtePointer<EVHD>(1)->evt) { 
    vtzero_(); 
    old_event=ArtePointer<EVHD>(1)->evt; 
  }
  int iflag=-1;
  int n0=vttrbf_.trmult, ni=0;
  for (ArteTable<RTRA>::iterator irtra=ArteTable<RTRA>::begin();
       irtra!=ArteTable<RTRA>::end(); ++irtra) {
    /*
    // -- check assignment to primary vertex (vertex driven method)
    // -- don;t do this and you get (candidate driven method)
    ArteRelationIterator<RREL> irrel=(*irtra)->rrel.begin();
    for (; irrel!=(*irtra)->rrel.end(); ++irrel) {
      if ((*irrel)->rver->wire && 
	  (*irrel)->rver->nt>=vattrib.min_tracks_in_v0) break; 
    }
    if (irrel!=(*irtra)->rrel.end()) continue;
    */


    // -- use only tracks that are not from primary vertex (MC)
    v0_mver = (*(*irtra)->mtra.begin())->bver;
    if (v0_mver->flag<=1) continue;

    int itr=(*irtra).index();
    if (vattrib.verb) cout<<"vtrini itr="<<itr<<endl;
    vtrini_(itr);
  }
  ni=vttrbf_.trmult;
  if (vattrib.verb) cout<<"vtsec: total added tracks = "<<ni-n0<<endl;
 
  int itb=0;
  vtsec_(itb,iflag);
  if (iflag) { cout<<"vertexmass: vtsec fitting error = "<<iflag<<endl; }
  if (vattrib.verb) {
    cout<<"Total sec. vertices="<<ivsec_.nvsec<<"  flag="<<iflag<<endl;
  }

  // -- save vertices to ARTE banks
  for (int i=0; i<ivsec_.nvsec; i++) {
    int iv=ivsec_.ivsec[i];
    int nvt=vtpntr_.ntr[iv-1];
    if (vattrib.verb) cout<<"iv="<<i<<"  vt="<<iv<<" with ntr="<<nvt<<endl;
    // -- use vertexmass data structure to book RVER
    removeAllVertexTracks();
    for (int j=0; j<nvt; j++) {
      int itr=vtpntr_.ktr[(iv-1)+j*NVTMAX];
      int irtra=vt_rtra_.vt_rtra[itr-1];
      addVertexTrack(ArtePointer<RTRA>(irtra));
      if (vattrib.verb) cout<<itr<<" rtra="<<irtra<<"  mtra="<<(*ArtePointer<RTRA>(irtra)->mtra.begin())<<endl;
    }
    for (int i=0; i<3; i++) v[i] = float(vtkalm_.xvs[(iv-1)*3+i]);
    chi2 = float(vtkalm_.chi2[(iv-1)+NVTMAX*(track_list->size()-1)]);
    chi2/= (2*track_list->size()-3);
    for (int i=0; i<3; i++) { 
      for (int j=0; j<3; j++) {
	cs[i][j]=float(vtkalm_.cs[((iv-1)*3+j)*3+i]);
      }
    }
    Mass=Merr=0;
    bookRVER();
  }

}
