#ifndef __DIELECTRIC_H__
#define __DIELECTRIC_H__

#include "rich/ricutil.h"

// ------------------------------------------------------------------------

extern "C" {
  extern void F77NAME(rgsdim)
       (const char* name,int* num,float* e,float* abso, float *eff,float *n);
  }
  
inline void rgsdim(const char* name,int num,float* e,float* abso, float
*eff,float *n) {
  F77NAME(rgsdim)(name,&num,e,abso,eff,n);
  }
  
/// description of a dieletric material
/**
  In the class DiElectric the properties of a dielectric material
  are described. This includes the refractive index,absortion,Snell's laws,
  Rayleigh-scattering as well as the production of Cherenkov radiation.
  */
class DiElectric {
char   *nname;
int    ichflag;
ArtePointer<CDIM> mat;
double rayleigh_coeff;
Table absorp,refr;  
//
//  Cherenkov Radiation
//
int    numphot,nphot;
double beta2,epsmax,step,emin,emax;
Table  dncdx;
CoordinateSystem track;

void init();

public:
/// returns the absorption length as function of the wavelength [m] 
  inline double Absorption(double lambda) { 
    return absorp.f(lambda); 
    }
/// returns the refractive index as function of the wavelength [m] 
  inline double RefractiveIndex(const double lambda)  { 
    return refr.f(lambda); 
    }
/// returns the relative dielectric constant as function of the wavelength [m] 
  inline double epsilon(double lambda) {
    return pow(RefractiveIndex(lambda),2);
    }
/// Fresnel laws
/**
 Describe a photon transistion from a dieletric materials with refractive index
 n1 to one with refractive index n2. The vector ns determine the normal
 of the surface on the hit position. It returns 1 in the case of total
 reflexion, 2 in the case of a reflexion and 3 in the case of a refraction.
 */
  int Fresnel(Photon* p,Vec3 ns,double n1,double n2);
/// calls Fresnel with n1=nold,n2 taken from this material
  int entering(Photon* p,Vec3 ns,double nold);
/// calls Fresnel with n2=nnew,n1 taken from this material
  int leaving(Photon* p,Vec3 ns,double nnew);

/// initialisation of the Cherenkov-Radiation monte carlo
  void Cherenkov(const Track& t,const double sstep);
/// total number of photon which will be produced
  inline int NumCherenkov() { return numphot; }
/// return 1, as long as further Cherenkov--photon will be radiated
  int  radiate(Photon* p);
/// return 1, if the photon could be traced through the material over a distance 'length'
  int trace(Photon* p,double length);
//
//   ParticleDensity: for gases:
//
//  float NA  = 6.0221367E23; // N/mol
//  float Vmol= 22.41410E-3;   // volume per mol m^3/mol
//  float N   = NA/Vmol;       // ideal gas at T=273.15K, p=1 atm
//
// returns index in CDIM, where the information can be accessed 
  inline        int    ArteTableIndex() const { return mat.index(); }
/// returns the particle density of the material
  inline        float  ParticleDensity()  const { return mat->dens; }
/// set the particle density of the material
  inline        float& ParticleDensity()        { return mat->dens; }
/// returns a generally assumed (if not zero) reflexion loss
  inline        float  ReflexionLoss()    const { return mat->rlos; }
/// set the generally assumed (if not zero) reflexion loss
  inline        float& ReflexionLoss()          { return mat->rlos; }
/// returns the Table describing the absorption length
  inline const  Table& Absorption()        const { return absorp; }
/// set the Table describing the absorption length
  inline        Table& Absorption()              { return absorp; }
/// returns the Table describing the refractive index
  inline const  Table& RefractiveIndex()   const { return refr; }
/// set the Table describing the refractivw index
  inline        Table& RefractiveIndex()         { return refr; }
/// returns the name of the material
  inline const char* name() const { return nname; } 
/// set the name of the material
  inline void name(char * n) {
    //if (nname) delete nname; 
    /*nname=strdup(n);*/
    nname=new char[strlen(n)+1];
    strcpy(nname,n);
    memset(mat->name,' ',sizeof(mat->name));
    size_t nn=strlen(nname);
    if (nn>sizeof(mat->name)) nn=sizeof(mat->name);
    memcpy(mat->name,nname,nn);
    }
/// print the contents of the class to an ASCII file.
  void print(FILE *fp) const;
/// constructor reading the information from an ASCII-FILE. See also page:
  DiElectric(ScriptFile& f);
/// constructor using a record in the ARTE-Table CDIM
    DiElectric(ArtePointer<CDIM> mat);
/// destructor
  ~DiElectric();
/// copy the material infromation to GEANT. A detection efficiency curve has to be provided
  void gsckov(int n,Table& efficiency);
// constructor only for special purposes.
  DiElectric();
  };
  

#endif
