//
//               Minimuzation of fitting function
//
//  Minimizes linearized likelihood function provided by RingFit::likelihood().
//
//
//  Created: Apr 11, 2001
//
//  Modified:
//
//   2001/05/15  (DD) Bug fix in covariance matrix
//   2001/05/16  (DD) Few bug fixes, minimization should work fine
//   2001/06/26  (DD) Mumbo-jumbo fixes
//
#include "rise/RObject.h"
#include "rise/RRingFit.h"
#include "rise/RErrorMsg.h"
#include "rise/RAttribute.h"


void invert3x3triangle(float A[3][3]) {
  float det=0;
  det  = A[2][2]*A[1][1]*A[0][0] + A[0][2]*A[1][2]*A[0][1]*2;
  det -= A[0][2]*A[1][1]*A[0][2] + A[1][2]*A[1][2]*A[0][0] + A[2][2]*A[0][1]*A[0][1];
  det  = fabs(det);
  if (det<1e-10) {
    RErrorMsg err("Matrix inversion problem, determinant zero",1);
    throw err;
  }

  float M[3][3];
  M[0][0] =           (  A[2][2]*A[1][1] - A[1][2]*A[1][2] ) / det;
  M[0][1] = M[1][0] = (- A[2][2]*A[0][1] + A[0][2]*A[1][2] ) / det;
  M[0][2] = M[2][0] = (  A[1][2]*A[0][1] - A[0][2]*A[1][1] ) / det;
  M[1][1] =           (  A[2][2]*A[0][0] - A[0][2]*A[0][2] ) / det;
  M[1][2] = M[2][1] = (- A[1][2]*A[0][0] + A[0][2]*A[0][1] ) / det;
  M[2][2] =           (  A[1][1]*A[0][0] - A[0][1]*A[0][1] ) / det;

  memcpy(&A[0][0], &M[0][0], sizeof(M));
}



void RGradient() {

  // ... get references for current ring, list of photons
  RRing            &ring    =*gRingFit->getRing();
  RTrack           &track   =*gRingFit->getRing()->itrack;
  vector<RCorrHit> &hit_list=*gRingFit->getHitList();

  // ... check if dof>0
  if (ring.nROI<RAttribute::minROI) {
    RErrorMsg err("Not enough hits in ROI",0);
    err.idle();
    return;
  };

  // ... iterate
  ring.nIt=0;
  float A[3][3], B[3];
  while (1) {

    if (++ring.nIt>=RAttribute::maxIter) break;

    // ... calculate likelihood, weights, fluxes
    try  { gRingFit->likelihood(); } 
    catch (RErrorMsg &err) { throw; }
    float iW  = 1/ring.weight;
    float rms = 0;
    float s2  = 0;

    memset(&A[0][0],0x0,sizeof(A));
    memset(&B[0],0x0,sizeof(B));
    float stx = track.stx();
    float sty = track.sty();
    for (vector<RCorrHit>::iterator ihit=hit_list.begin();
	 ihit!=hit_list.end(); ++ihit) {
      //float si2  = (ring.nIt>-1) ? 1/s2 : 1/(ihit->ihit->sigma*ihit->ihit->sigma);///// -1 !!!!!
      float si2  = 1/(ihit->ihit->sigma*ihit->ihit->sigma);
      float wsi2 = ihit->w * si2;
      float dx   = ihit->X - ring.X;
      float dy   = ihit->Y - ring.Y;
      float dr   = sqrt(dx*dx+dy*dy);
      float ux   = dx/dr;
      float uy   = dy/dr;
      rms       += wsi2 * (dr - ring.radius) * (dr - ring.radius);
      // ... covariance matrix
      A[0][0]   += wsi2;
      A[0][1]   += wsi2 * ux;
      A[0][2]   += wsi2 * uy;
      A[1][1]   += wsi2 * ux * ux;
      A[1][2]   += wsi2 * ux * uy;
      A[2][2]   += wsi2 * uy * uy;
      // ... deviation vector
      B[0]      += wsi2 * (dr - ring.radius);
      B[1]      += wsi2 * (dr - ring.radius) * ux;
      B[2]      += wsi2 * (dr - ring.radius) * uy;
    }
    A[0][0]  = (A[0][0] + 4*ring.radiusH*ring.radiusH/ring.R2P2error2) * iW;
    A[0][1] *= iW;
    A[0][2] *= iW;
    A[1][1]  = (A[1][1] + 1/(stx*stx)) * iW;
    A[1][2] *= iW;
    A[2][2]  = (A[1][1] + 1/(sty*sty)) * iW;
    A[1][0]  = A[0][1];
    A[2][0]  = A[0][2];
    A[2][1]  = A[1][2];
    B[0]    *= iW;
    B[1]    *= iW;
    B[2]    *= iW;
    s2       = rms * iW;

    //cout<<"A:"<<endl; for (int i=0; i<3; i++) { for (int j=0; j<3; j++) { cout<<A[i][j]<<" "; } cout<<endl; }
    //cout<<"B:"<<endl; for (int i=0; i<3; i++) cout<<B[i]<<endl;

    // ... invert matrix
    try  { invert3x3triangle(A); } 
    catch (RErrorMsg &err) { throw; }

    // ... new ring paramters
    //cout<<"OLD:"<<endl; ring.print();
    float dR=0, dX=0, dY=0;
    for (int i=0; i<3; i++) dR += A[0][i]*B[i];
    for (int i=0; i<3; i++) dX += A[1][i]*B[i];
    for (int i=0; i<3; i++) dY += A[2][i]*B[i];
    if (RAttribute::fitR) ring.radius += dR/ring.weight;
    if (RAttribute::fitX) ring.X      += dX/ring.weight;
    if (RAttribute::fitY) ring.Y      += dY/ring.weight;
    //cout<<"NEW:"<<endl; ring.print();

    if (ring.radius<=0) {
      RErrorMsg err("Ring radius less than zero, restoring hyp. radius",0);
      err.idle();
      ring.radius=ring.radiusH;
      ring.X=track.X;
      ring.Y=track.Y;
      return; // NOT throw err;
    }
    if (ring.radius>RAttribute::maxR) {
      RErrorMsg err("Ring radius greater than maximum value, restoring hyp. radius",0);
      err.idle();
      ring.radius=ring.radiusH;
      ring.X=track.X;
      ring.Y=track.Y;
      return; // NOT throw err;
    }

    // ... check convergence
    if (fabs(dR)<1e-5 && fabs(dX)<1e-5 && fabs(dY)<1e-5) break;
  }

}

