//--*-C++-*--
//
//  Created: May 11 2001
//
//  Modifications:
// 
//  2001/05/23   DD  fillFitFlag added
//  2001/05/31   DD  0.007 changed to RAttribute::deltaR
//  2001/06/26   DD  fillFitFlag corrected
//
#include "rise/RRingFit.h"
#include "rise/RErrorMsg.h"
#include "rise/RAttribute.h"
#ifdef USE_ARTE
#include "rich/PhiLambda.h"
#endif



RRingFit *gRingFit=0;

void dummyMinFunction()  { 
  RErrorMsg err("Minimization function NOT set",-1);
  throw err;
}

void dummySelectROI()  { 
  RErrorMsg err("ROI selection function NOT set",-1);
  throw err;
}



RRingFit::RRingFit() { 
  RICH.N0T     = 11000.; 
  RICH.radius0 = 0.0526; 
  globalFit    = false;
  minimize     = dummyMinFunction;
  selectROI    = dummySelectROI;
  gRingFit     = this;
} 




void RRingFit::Rb(bool signal) {
  float A       = ring->SArea * RICH.N0T * PI32;
  float &newRb  = (signal) ? ring->Rbs : ring->Rbb;
  float curRb   = -1;
  newRb         = 1/A; 
  while (fabs(newRb-curRb)>.001) {
    curRb = newRb;
    float fRb=A, dfRb=0;
    for (vector<RCorrHit>::iterator ihit=hit_list.begin();
	 ihit!=hit_list.end(); ++ihit) {
      float x = (signal) ? 1/(curRb + ihit->flux + ihit->beta) : 1/(curRb + ihit->beta);      
      fRb  -= x;
      dfRb += x*x;
    }
    if (fabs(dfRb)<1e-10) {
      RErrorMsg err("dfRb too small",0);
      if (RAttribute::verb>2) err.print(); 
      break;
    }
    newRb = curRb - fRb/dfRb;
  }
  if (newRb<-10) newRb=1e-5;
}




float RRingFit::likelihood() {
  fluxAll();
  betaAll();
  RbAll();
  ring->L=-1e5;
  if (ring->Rbs>=ring->Rbb) {
    RErrorMsg err("Negative number of signal photons",0);
    throw err;    
  }
  ring->S      =   RICH.N0T * ring->SArea * PI32 * (ring->Rbb-ring->Rbs);
  float L1     = - RICH.N0T * ring->radiusH*ring->radiusH*ring->Fs + ring->S;
  float L2     = 0;
  ring->weight = 0;
  for (vector<RCorrHit>::iterator ihit=hit_list.begin();
       ihit!=hit_list.end(); ++ihit) {
    ihit->w = ihit->flux / ( ihit->flux + ring->Rbs + ihit->beta );
    if (ihit->w<0) ihit->w = 0;
    else ring->weight += ihit->w;
    if (ring->Rbb<=1e-5) {
      RErrorMsg err("Rbb too small",0);
      if (RAttribute::verb>2) err.print(); 
      continue;
    }
    float t = (ihit->flux + ring->Rbs + ihit->beta)/(ring->Rbb + ihit->beta);
    if (t<=0) {
      RErrorMsg err("Negative t in likelihood calculation",0);
      if (RAttribute::verb>2) err.print(); 
      continue; 
    }
    L2 += log( t );
  }
  float L3 = - 0.5 * matchingChi2();
  ring->L  = L1+L2+L3;
  //cout<<"L="<<ring->L<<"  L1="<<L1<<"  L2="<<L2<<"  L3="<<L3<<"  Rbs="<<ring->Rbs<<"  Rbb="<<ring->Rbb<<endl;
  return ring->L;
}



float RRingFit::matchingChi2() {
  float DX  =  (track->X - ring->X)/dX();
  float DY  =  (track->Y - ring->Y)/dY();
  float DR2 =  ring->radius*ring->radius - ring->radiusH*ring->radiusH;
  return   DX*DX + DY*DY + DR2*DR2/ring->R2P2error2;
}



float RRingFit::flux(const RCorrHit &h) {
  float  dx = ring->X - h.X;
  float  dy = ring->Y - h.Y;
  float t   = (ring->radius - sqrt(dx*dx+dy*dy))/(h.ihit->sigma);
  if (t>6) return 0;
  return (ring->radius)/(h.ihit->sigma)*exp(-.5*t*t); // should this be ring->radius?
}

void RRingFit::fluxAll() {
  for (vector<RCorrHit>::iterator ihit=hit_list.begin();
       ihit!=hit_list.end(); ++ihit) {
    ihit->flux = flux(*ihit);
  }
}


float RRingFit::beta(const RCorrHit &h) {
  float Beta=0;
  for (vector<RRing>::iterator iring=RichObject<RRing>::list.begin();
       iring!=RichObject<RRing>::list.end(); ++iring) {
    if (ring==iring) continue;
    RTrack &track = *iring->itrack;
    RICH_LambdaPhi->Correct()->newRing(track.X, track.Y, iring->radius, track.bx(), track.by());
    float  dx     = - RICH_LambdaPhi->Correct()->dPhi   ( h.ihit->x, h.ihit->y, h.ihit->plane);
    float  dy     = - RICH_LambdaPhi->Correct()->dLambda( h.ihit->x, h.ihit->y, h.ihit->plane);
    float  t      =  (iring->radius-sqrt(dx*dx + dy*dy))/(h.ihit->sigma);
    if (t<6) Beta += (iring->radius)/(h.ihit->sigma)*exp(-.5*t*t);
  }
  return Beta;
}

void RRingFit::betaAll() {
  for (vector<RCorrHit>::iterator ihit=hit_list.begin();
       ihit!=hit_list.end(); ++ihit) {
    ihit->beta = (globalFit) ? beta(*ihit) : 0;
  }  
}


void RRingFit::fitQuality() {
  try {
    likelihood();
    fitChi2();
    fitDh();
    fitDr();
  }
  catch (RErrorMsg &err) { throw; }
}


void RRingFit::fitChi2() {
  float chi2=0;
  ring->chi2=1e5;
  float avgsigma=0;
  for (vector<RCorrHit>::iterator ihit=hit_list.begin();
       ihit!=hit_list.end(); ++ihit) {
    float dx  = ring->X - ihit->X;
    float dy  = ring->Y - ihit->Y;
    float t   = (ring->radius - sqrt(dx*dx+dy*dy))/(ihit->ihit->sigma);
    chi2     += ihit->w * t*t;
    avgsigma += ihit->ihit->sigma;
  }
  if (ring->weight<=0 || chi2>1e5 || avgsigma<=0) {
    RErrorMsg err("Ring quality insufficient or problems with determination",0);
    throw err;
  }
  avgsigma   /= hit_list.size();
  ring->Delta = sqrt(2*log(ring->radiusH/(ring->Rbs*avgsigma)));
  ring->chi2  = chi2/ring->weight;
}


void RRingFit::fitDh() {
  float m2   = ring->mass*ring->mass;
  float r2   = ring->radius*ring->radius;
  float rH2  = ring->radiusH*ring->radiusH;;
  float ip2  = ring->itrack->ip*ring->itrack->ip;
  float dp   = dpp();
  float sip2 = 4*(dp*dp)*(ip2*ip2);
  ring->Dh   = (r2 - rH2)/sqrt(ring->R2P2error2 + m2*m2*sip2);
  if (ring->Dh>100) {
    RErrorMsg err("Ring too far from R2P2 line",0);
    throw err;
  }
}


void RRingFit::fitDr() {
  static float isqrt2=1/sqrt(2);
  float r2  = ring->radius*ring->radius;
  float rH2 = ring->radiusH*ring->radiusH;
  ring->Dr  = fabs(r2 - rH2)/sqrt(ring->R2P2error2)*isqrt2;
}




extern "C" float erfc_(float &x);
extern "C" float prob_(float &x, int &dof);
void RRingFit::probability() {
  float lr[5];
  memset(&lr[0],0x0,sizeof(lr));
  float lrsum=0;
  int   npar = RAttribute::fitR + RAttribute::fitX + RAttribute::fitY;
  if (!globalFit) {
    /* 
       Each hypothesis has its own R2P2 distance, likelihood, radius...
       so fill them in lr[5] of the best ring. Only the best ring
       will be preserved for this track.
    */
    int i=0, dof=0;
    for (vector<RRing>::iterator iring=init_ring_list.begin();
	 iring!=init_ring_list.end(); ++iring) {
      switch (RAttribute::probType) {
      case 0: // use only radius 
	lr[i] = erfc_(iring->Dr);
	break;
      case 1: // use radius and momentum
	lr[i] = exp(-0.5*iring->Dh*iring->Dh);
	break;
      case 2: // use likelihood from the fit
	lr[i] = iring->L;
	break;
      case 3: // use chi2 probability
	dof=iring->nROI-npar;
	lr[i] = (dof>0) ? prob_(iring->chi2, dof) : 0;
	break;
      default: 
	RErrorMsg err("Wrong RAttribute::probType",-1);
	throw err;
      }
      lrsum += lr[i++];
    }
  }  else {
    /* 
       Only one ring (hypothesis) exists so likelihoods for other
       hypothesis are calculated using its radius
    */
    static float mass[]={.0005109989, .105658357, .13957018, .493677, .93827};
    RRing save_ring(*ring);
    for (int i=0; i<5; i++) {
      ring->mass=mass[i];
      ring->setR2P2error2();
      ring->radiusH=RICH.radius0*RICH.radius0 - 
	ring->mass*ring->mass * ring->itrack->ip*ring->itrack->ip;
      switch (RAttribute::probTypeGlobal) {
      case 0: // use only radius 
	fitDr();
	lr[i] = erfc_(ring->Dr);
	break;
      case 1: // use radius and momentum
	fitDh();
	lr[i] = exp(-0.5*ring->Dh*ring->Dh);
	break;
      default: 
	RErrorMsg err("Wrong RAttribute::probTypeGlobal",-1);
	throw err;
      }
    }
    *ring=save_ring;
  }

  // ... check likelihood sum and normalize if needed
  if (fabs(lrsum)<1e-15) {
    RErrorMsg err("Likelihood sum zero",-1);
    throw err;
  }
  for (int i=0; i<5; i++) {
    ring->lr[i] = (RAttribute::normL) ? lr[i]/lrsum : lr[i];
  }
}


void RRingFit::fillFitFlag() {
  /*
    Evaluate truth table and fill the fit flag:
    0=below threshold
    1=abolutely no
    2=maybe, or best hypothesis
    3=ok

    Save truth table in a human readable format:
    e + mu*10 + pi*100 + K*1000 + p*10000
    (less than 2 bytes!)
  */
  int truth_table[5];
  int shift[5]={1,10,100,1000,10000};
  int i=0;
  memset(&truth_table[0],0x0,sizeof(truth_table));
  ring->flag = 0;
  for (vector<RRing>::iterator iring=init_ring_list.begin();
       iring!=init_ring_list.end(); ++iring) {
    if      (iring->L<RAttribute::minL) truth_table[i]=1;
    else if (iring->L<RAttribute::okL)  truth_table[i]=2;
    else                                truth_table[i]=3;
    ring->flag += truth_table[i]*shift[i];
    i++;
  }  
}


void RRingFit::printFitSummary() {
  static int multi=0;
  if (multi%10==0) {
    cout<<"MC                e       \t         mu          \t       pi         \t        K        \t       p"<<endl;
    cout<<"          L/Lb/Dh/chi2    \t L/Lb/Dh/chi2        \t L/Lb/Dh/chi2     \t L/Lb/Dh/chi2    \t  L/Lb/Dh/chi2"<<endl;
  }
  //if (ArteTable<MTRA>::size()&&      ) cout<<(*ArtePointer<RTRA>(track->index)->mtra.begin())->topp->lund<<" ";
  //else                         cout<<"   ";
  for (vector<RRing>::iterator iring=init_ring_list.begin();
       iring!=init_ring_list.end(); ++iring) {
    cout<<iring->L<<"/"<<iring->Dh<<"/"<<iring->chi2<<"\t";
  }
  cout<<endl;
  ring->print(1);
  multi++;
}



void RRingFit::fit(vector<RTrack>::iterator newtrack) {

  globalFit=false;

  // ... STEP 0: fill initial condition/hypothesis
  static float mass[]={.0005109989, .105658357, .13957018, .493677, .93827};
  if (init_ring_list.size()) init_ring_list.clear();
  for (int i=0; i<5; i++) {
    RRing rring(newtrack);
    rring.mass = mass[i];
    float rad0 = RICH.radius0;
    float rad2 = (rad0*rad0) - (rring.mass*rring.mass)*(newtrack->ip*newtrack->ip);
    if (rad2<0) {
      continue;
    }
    float &delR=RAttribute::deltaR;
    rring.radiusH = rring.radius = sqrt(rad2);
    rring.r2min   = (rring.radiusH-delR)*(rring.radiusH-delR);
    if (rring.r2min<0) rring.r2min=0;
    rring.r2max   = (rring.radiusH+delR)*(rring.radiusH+delR);
    rring.setSArea();
    rring.setR2P2error2();
    rring.setShadow();
    init_ring_list.push_back(rring);
  }
  if (!init_ring_list.size()) {
    RErrorMsg err("None of hypothesis over threshold",0);
    throw err;
  }
  
  // ... STEP 1: loop over all hypothesis and do fitting
  int nfit=0;
  for (vector<RRing>::iterator iring=init_ring_list.begin();
       iring!=init_ring_list.end(); ++iring) {
    try  {
      // .. set current ring
      setRing(*iring);
      // .. fill photon ROI
      selectROI();
      // .. get initial likelihood, fill weights
      likelihood();
      //ring->print(0);
      // .. do fitting
      minimize();
      // .. evaluate quality
      fitQuality();
      // .. accept ring
      ring->accept=true;
      nfit++;
    }  catch (RErrorMsg &err) {
      if (err.code<0) throw;
    }
  }
  if (!nfit) {
    RErrorMsg err("None of hypothesis accepted",0);
    throw err;
  }

  // ... STEP 2: choose best ringfit, and store it in a global list
  for (vector<RRing>::iterator iring=init_ring_list.begin();
       iring!=init_ring_list.end(); ++iring) {
    if (!iring->accept) continue;
    else iring->accept=false;
    switch (RAttribute::probType) {
    case 0: if ( iring->Dr       < ring->Dr )       ring=iring; break; // best delta r
    case 1: if ( fabs(iring->Dh) < fabs(ring->Dh) ) ring=iring; break; // best delta r2p2
    case 2: if ( iring->L        > ring->L )        ring=iring; break; // best likelihood
    case 3: if ( iring->chi2     < ring->chi2 )     ring=iring; break; // best chi2
    case 4: RErrorMsg err("RAttribute::probType out of range",-1); throw err;
    }
  }
  ring->accept=true;
  try {
    probability();
    fillFitFlag();
  } catch(RErrorMsg &err) {
    if (err.code<0) throw;
  }
  RichObject<RRing>::list.push_back(*ring);

  // ... debug: print truth table
  if      (RAttribute::verb>1) printFitSummary();
  else if (RAttribute::verb>0) ring->print(1);
}




void RRingFit::globalfit() {

  globalFit=true;

  // .. loop over inital rings
  for (vector<RRing>::iterator iring=RichObject<RRing>::list.begin();
       iring!=RichObject<RRing>::list.end(); ++iring) {
    try {
      // .. set current ring
      setRing(*iring);
      // .. fill photon ROI
      selectROI();
      // .. eveluate likelihood, weights
      likelihood();
      // .. do fitting,
      minimize(); 
      // .. fit quality
      fitQuality();
      // .. calculate probability
      probability();
    } catch (RErrorMsg &err) {
      ring->accept=false;
      if (err.code<0) throw;
    }
    if (RAttribute::verb) {
      track->print();
      ring->print(1);
    }
  }

}
