//
//  Description:
//    Ring fitting
//
//
//  Author:  Denis Dujmic
//
//  Created: Apr 04 2001
//
//  Modifications:
//
//   2001/05/15 DD Initialization printout
//
#include "rise/Ring.h"
#include "rise/RErrorMsg.h"
#include "rise/RAttribute.h"

// .... ARTE specific ....
#ifdef USE_ARTE
void  RCopyHITB();
void  RCopyRTRA();
void  RFillARTE();
void  RSelectROIHITB();
float RShadow(float *track);
#endif
// .... Minuit minimization in ROOT ...
#ifdef USE_ROOT
void RMinuit();
#endif
// .... Custom minimization
void RGradient();



int Ring() {

  // ... initialize once per session
  static bool init=false;
  static RichObject<RHit>    hits;      // container for rich hits
  static RichObject<RTrack>  tracks;    // container for seeded tracks
  static RichObject<RRing>   rings;     // container for fitter rings
  static RRingFit            ringfit;   // fitting class
  if (!init) {
    RErrorMsg err("Initialization",0);
    err.idle();
#ifdef USE_ARTE
    hits.setFillFunction  (RCopyHITB);    // use hits from HITB
    tracks.setFillFunction(RCopyRTRA);    // user tracks from RTRA
    rings.setSaveFunction (RFillARTE);    // save results in ARTE tables
    ringfit.setSelectROI(RSelectROIHITB); // use HERA-B optical corrections
    RRing::setShadowFunction(RShadow);    // use HERA-B shadowing/acceptance
#endif
#ifdef USE_ROOT
    //ringfit.setMinFunction(RMinuit);      // use Minuit for fitting
#endif
    ringfit.setMinFunction(RGradient);    // use custom code for fitting
    ringfit.setN0T(11000.);               // set N0*T for HERA-B
    init=true;
    RAttribute::print();
  }

  // ... set up fitting for every event
  try {
    ringfit.setRadius0(ArtePointer<RITH>(1)->rad0);
    hits.fillList();  //hits.print();
    tracks.fillList();//tracks.print();
  } catch (RErrorMsg &err) {
    if (err.code<0) {
      err.print();
      return err.code;
    }
  }

  // ... fit rings
  rings.clear();
  if (RAttribute::verb) {
    cout<<"----------------------------------------------------------------------------------"<<endl;
    cout<<"----------------------------------------------------------------------------------"<<endl;
    cout<<"---------------> SINGLE FIT, total seeds="<<RichObject<RTrack>::list.size()<<endl;
    cout<<"----------------------------------------------------------------------------------"<<endl;
    cout<<"----------------------------------------------------------------------------------"<<endl;
  }
  for (vector<RTrack>::iterator itrack=RichObject<RTrack>::list.begin();
       itrack!=RichObject<RTrack>::list.end(); ++itrack) {
    try {
      ringfit.fit(itrack);
    } catch (RErrorMsg &err) {
      if (err.code<0) err.print();
    }
  }

  // ... make global fit
  if (RAttribute::globalFit) try {
    if (RAttribute::verb) {
    cout<<"----------------------------------------------------------------------------------"<<endl;
    cout<<"----------------------------------------------------------------------------------"<<endl;
    cout<<"---------------> GLOBAL FIT, total number of rings="<<RichObject<RRing>::list.size()<<endl;
    cout<<"----------------------------------------------------------------------------------"<<endl;
    cout<<"----------------------------------------------------------------------------------"<<endl;
    }
    ringfit.globalfit();
  } catch(RErrorMsg &err) {
    if (err.code<0) err.print();
  }

  // ... save rings
  rings.saveList();

  return 0;
}
