// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Track class
//
// changes:
// 21 Aug 2000 (TG) creation
// 24 Aug 2000 (TG) include arte.hh header file
// 06 Sep 2000 (TG) changed call to t_V.propagate()
// 24 Sep 2000 (TG) delete_mom() member function added temporarily for
//                  debugging. Not for public use!!
// 10 Okt 2000 (TG) fill member t_GM in constructors
// 26 Okt 2000 (TG) #ifndef NO_ARTE added (for integration into CLUE)
//
// *****************************************************************************
#include <iostream>
#include <iomanip>
#include "arte/arte.hh"
#include "vt++/VtTrack.hh"
#include "vt++/VtKalman.hh"
#include "vt++/VtRelation.hh"

using namespace MATRIX;

namespace VERTEX {

  //============================================================================
  // Track::Constructors
  //============================================================================
  Track::Track():
    t_arteptr(0),
    t_isrtra (false),
    t_Q      (0),
    t_rm     (0.),
    t_p      (VtVector(6))
  {}

#ifndef NO_ARTE
  Track::Track(const ArtePointer<RTRA>& rtra):
    t_arteptr(getIDX<RTRA>(rtra)),
    t_isrtra (true),
    t_Q      (sgn(rtra->pf)),
    t_rm     (0.),
    t_p      (MATRIX::VtVector(rtra->xf,
			       rtra->yf,
			       rtra->zf,
			       rtra->txf,
			       rtra->tyf,
			      (rtra->pf!=0)? fabs(1./rtra->pf) : 1.e50)),
    t_V      (rtra)
  {
    propagate(0.);           // propagate track to z=0.
    t_G = t_V.dsinv();
    t_GM.copy(t_V.dsinv(4)); // inverse without momentum
  }
  
  Track::Track(const ArtePointer<RSEG>& rseg):
    t_arteptr(getIDX<RSEG>(rseg)),
    t_isrtra (false),
    t_Q      (sgn(rseg->pf)),
    t_rm     (0.),
    t_p      (MATRIX::VtVector(rseg->xf,
			       rseg->yf,
			       rseg->zf,
			       rseg->txf,
			       rseg->tyf,
			      (rseg->pf!=0)? fabs(1./rseg->pf) : 1.e50)),
    t_V      (rseg)
  {
    propagate(0.);           // propagate track to z=0.
    t_G = t_V.dsinv();
    t_GM.copy(t_V.dsinv(4)); // inverse without momentum
  }
  
  Track::Track(const ArtePointer<MTRA>& mtra):
    t_arteptr(getIDX<MTRA>(mtra)),
    t_isrtra (false),
    t_Q      (0),
    t_rm     (0.),
    t_p      (MATRIX::VtVector(mtra->bver->x,
			       mtra->bver->y,
			       mtra->bver->z,
			       mtra->px/mtra->pz,
			       mtra->py/mtra->pz,
			       fabs(sqrt(mtra->px*mtra->px+mtra->py*mtra->py+
					 mtra->pz*mtra->pz)))),
    t_V      (mtra)
  {
    propagate(0.);           // propagate track to z=0.
    t_G = t_V.dsinv();
    t_GM.copy(t_V.dsinv(4)); // inverse without momentum
  }
#endif // NO_ARTE

  Track::Track(const Track* const track):
    t_arteptr(track->t_arteptr),
    t_isrtra (track->t_isrtra),
    t_Q      (track->t_Q),
    t_rm     (0.),
    t_p      (track->t_p),
    t_V      (track->t_V)
  {
    propagate(0.);           // propagate track to z=0.
    t_G = t_V.dsinv();
    t_GM.copy(t_V.dsinv(4)); // inverse without momentum
  }

  Track::Track(const Track& rhs):
    t_arteptr(rhs.t_arteptr),
    t_isrtra (rhs.t_isrtra),
    t_Q      (rhs.t_Q),
    t_rm     (rhs.t_rm),
    t_p      (rhs.t_p),
    t_V      (rhs.t_V)
  {
    propagate(0.);           // propagate track to z=0.
    t_G = t_V.dsinv();
    t_GM.copy(t_V.dsinv(4)); // inverse without momentum
  }

  Track::~Track() {}
  
  //============================================================================
  // Track::print
  //============================================================================
  void Track::print( std::ostream& os ) const {
    // os << "[" << t_arteptr << "," << this << "]";

    os  << " x: "  << x()
	<< " y: "  << y()
	<< " z: "  << z()
	<< " tx: " << tx()
	<< " ty: " << ty()
	<< " p: "  << p() 
	<< endl
	<< " Cov Matrix: " << V()
	<< " inverse: " << G()
	<< " inverse without momentum: " << GM();
  }

  //============================================================================
  // Track::propagate propagate Track to z 
  //============================================================================
  void Track::propagate(const double zz) {
    if( zz == z() ) return;

    double dz = zz - z();
    t_p[0] += dz * tx();
    t_p[1] += dz * ty();
    t_p[2]  = zz;
    t_V.propagate(dz); // propagate Track covariance matrix
    return;
  }

  //============================================================================
  // Track::delete_mom for debugging only
  //============================================================================
  void Track::delete_mom() {
    t_V.set_x_p(0.);
    t_V.set_y_p(0.);
    t_V.set_tx_p(0.);
    t_V.set_ty_p(0.);
    t_V.set_p(1e5);
    t_G = t_V;
    if(t_G.VtDsinv()==false)
      cout << " error inverting matrix!!" << endl;
	
  }
} // namespace VERTEX
