#ifndef __VTTRACK_HH
#define __VTTRACK_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Track class
//
// changes:
// 21 Aug 2000 (TG) creation
// 24 Aug 2000 (TG) added ArtePointer declaration
// 04 Sep 2000 (TG) added pt(), theta(), phi(), eta() member function
// 05 Sep 2000 (TG) declared some member functions inline
// 06 Sep 2000 (TG) added xvec(), tvec(), pvec(), E() member functions
// 06 Sep 2000 (TG) added comments
// 12 Sep 2000 (TG) added ex(), ey(), ez(), evec(), rm() member functions
// 10 Okt 2000 (TG) data member t_GM and member function GM() added
// 26 Okt 2000 (TG) added #include <cmath>
// 26 Okt 2000 (TG) #ifndef NO_ARTE added (for integration into CLUE)
//
// *****************************************************************************

#include <iosfwd>
#include <cmath>
#include "vt++/VtUtil.hh"
#include "vt++/CMatrix.hh"
#include "vt++/VtVector.hh"
#include "vt++/VtRelationList.hh"

template <class T> class ArtePointer;
#ifndef NO_ARTE
class RTRA;
class RSEG;
class MTRA;
#endif // NO_ARTE

namespace VERTEX {
  class Kalman;

  /** Track representation class
   */
  //============================================================================
  // Class Track
  //============================================================================
  class Track: public RelationList {
  public:
    /**@name Constructors */
    ///
    Track();
#ifndef NO_ARTE
    ///
    Track(const ArtePointer<RTRA>& rtra);
    ///
    Track(const ArtePointer<RSEG>& rseg);
    ///
    Track(const ArtePointer<MTRA>& mtra);
#endif // NO_ARTE
    ///
    Track(const Track* const track);
    ///
    Track(const Track& rhs);
    ///
    virtual ~Track();

    /**@name Access functions */
    /// x at z=0 (in Vt: p(3,..))
    inline double x()      const { return t_p[0]; }
    ///  y at z=0 (in Vt: p(4,..))
    inline double y()      const { return t_p[1]; }
    /// z = 0  in Vt
    inline double z()      const { return t_p[2]; }
    ///  slope    (in Vt: p(1,..))
    inline double tx()     const { return t_p[3]; }
    ///  slope    (in Vt: p(2,..))
    inline double ty()     const { return t_p[4]; }
    /// momentum (in Vt: p(5,..))
    inline double p()      const { return t_p[5]; }
    /// $p \cdot e_x$
    inline double px()     const { return p()  * ex(); }
    /// $p \cdot e_y$
    inline double py()     const { return p()  * ey(); }
    /// $p \cdot e_z$
    inline double pz()     const { return p()  * ez(); }
    /// $t_x \cdot e_z$
    inline double ex()     const { return tx() * ez(); }
    /// $t_y \cdot e_z$
    inline double ey()     const { return ty() * ez(); }
    /// $1/\sqrt{1+t_x^2+t_y^2}$
    inline double ez()     const { return 1./sqrt(1. + sqr(tx()) + sqr(ty())); }
    /// transv. momentum $\sqrt{p_x^2 + p_y^2}$
    inline double pt()     const { return sqrt(sqr(px()) + sqr(py())); }
    /// polar angle $\theta = \cos^{-1}(e_z)$
    inline double theta()  const { return acos(ez()); }
    /// azimuthal angle $\phi$
    inline double phi()    const { return (180. + atan2(ty(),tx())*57.29577957); }
    /// rapidity $-\log\tan(\theta/2.)$
    inline double eta()    const { return -log(tan(theta()/2.)); }
    /// charge
    inline short int Q()   const { return t_Q; }
    /// Energy with given rest-mass (in GeV) $E = \sqrt{p^2 + m^2}$
    inline double E(double rm = 0.1396) const { return sqrt(sqr(p()) + sqr(rm)); }

    /// $\vec{v} = (e_x,e_y,e_z)$ unit vector along track
    inline MATRIX::VtVector evec() const { return MATRIX::VtVector(ex(),ey(),ez()); }
    /// $\vec{v} = (x,y,z)$
    inline MATRIX::VtVector xvec() const { return MATRIX::VtVector(x(),y(),z()); }
    /// $\vec{v} = (t_x,t_y,1.)$ slope vector
    inline MATRIX::VtVector tvec() const { return MATRIX::VtVector(tx(),ty(),1.); }
    /// $\vec{v} = (p_x,p_y,p_z)$ mom. vector
    inline MATRIX::VtVector pvec() const { return MATRIX::VtVector(px(),py(),pz()); }

    /// is Track constructed from RTRA object?
    inline bool   IsRTRA()              const { return t_isrtra; }
    /// $\sqrt{\sigma_x^2(z+dz)}$
    inline double xerr(double dz = 0)   const { return sqrt(t_V.x_prop(dz)); }
    /// $\sqrt{\sigma_y^2(z+dz)}$
    inline double yerr(double dz = 0)   const { return sqrt(t_V.y_prop(dz)); }
    /// $\sqrt{\sigma^2_{tx}}$
    inline double txerr()               const { return sqrt(t_V(2,2)); }
    /// $\sqrt{\sigma^2_{ty}}$
    inline double tyerr()               const { return sqrt(t_V(3,3)); }
    /// $\sqrt{\sigma^2_p}$
    inline double perr()                const { return sqrt(t_V.p()); }

#ifndef NO_ARTE
    /// return RSEG pointer
    ArtePointer<RSEG> rseg() const { return getPTR<RSEG>(t_arteptr); }
    /// return RTRA pointer
    ArtePointer<RTRA> rtra() const { return getPTR<RTRA>(t_arteptr); }
    /// return MTRA pointer
    ArtePointer<MTRA> mtra() const { return getPTR<MTRA>(t_arteptr); }
#endif // NO_ARTE

    /// covariance matrix
    const MATRIX::CMatrix&     V()  const { return t_V; }
    /// inverse $5\times5$ covariance matrix
    const MATRIX::CMatrix&     G()  const { return t_G; }
    /// inverse $4\times4$ covariance matrix, without momentum
    const MATRIX::CMatrix&     GM() const { return t_GM; }

    /**@name Expert functions */
    /// propagate track by $dz$
    void propagate(const double dz); 
    /// called by cout
    void print( std::ostream& os ) const;
    /// rest-mass
    inline double rm() const { return t_rm; }
    /// set rest-mass (needed for mass constrained fits)
    inline void rm(const double mass) { t_rm = mass; }

    void delete_mom();

    bool operator< (const Track& rhs) { return t_arteptr <  rhs.t_arteptr; }
    int  operator==(const Track& rhs) { return t_arteptr == rhs.t_arteptr; }

  private:
    unsigned int        t_arteptr;
    bool                t_isrtra;
    short int           t_Q;      // charge
    double              t_rm;     // rest-mass (in Vt: xmass)
  protected:
    MATRIX::VtVector    t_p;      // track parameter vector (x,y,z,tx,ty,p)
			  	  // (in Vt: p(1,...,5) 
    MATRIX::CMatrix     t_V;      // 5x5 covariance matrix of Track
    MATRIX::CMatrix     t_G;      // inverse of t_cov (5x5)
    MATRIX::CMatrix     t_GM;     // inverse of t_cov (4x4, without momentum)
 }; // class Track


  //==============================================================================
  // operator <<
  //==============================================================================
  inline std::ostream& operator<< ( std::ostream& os, const Track& t ) {
    t.print(os);
    return os;
  }

} // end of namespace VERTEX
#endif
